/*-D************************************************************************* 
'*   This program source file is confidential and proprietary property of
'*    Welch Allyn, Inc. Unauthorized use or reproduction is prohibited.
'****************************************************************************
'****************************************************************************
'**
'****************************************************************************
'**  ////////////////////////////////////////////////////////////////////////
'**  '// CCommSvr.c     :                                                  //
'**  '// Language       :   C                                              //
'**  '// Application    :   WACP                                           //
'**  '// Sources        :   MOMP Standard Library                          //
'**  '///////////////////////////////////////////////////////////////////////
'*-E************************************************************************/
#include "CCommSvr.h"
#include "SysUtil.h"
#include "CWACPThread.h"

#if defined(STACK_SERIAL)
    #include "CSerialCommSvr.h"
#endif

#if defined(STACK_ETHERNET)
    #include "CEthCommSvr.h"
#endif

static CWACPThread CCommThread;

/*================================================================================
CCommSvr_Init:  Initializes function pointers and CWACPThread instance.
================================================================================*/
void CCommSvr_Init(CCommSvr* pObject, 
                   boolean (*RecvMsg)(char* pSourceID, uint8* pBuf, uint32 MsgSize),
                   void (*Terminate)(char* pSourceID, uint16 Port))
{
    /* Store the callback function to add a new message */
    pObject->pSetRecvMsg = RecvMsg;

    /* Store the callback function to terminate the connection */
    pObject->pTerminate = Terminate;
   
    /* Initialize the CCommThread instance */
    CWACPThread_Init(&CCommThread, WACP_CCOMSVR_THREAD);
}

/*================================================================================
CCommSvr_Destroy: Resets function pointers and CWACPThread instance.
================================================================================*/
void CCommSvr_Destroy(CCommSvr* pObject)
{
    /* Reset the callback */
    pObject->pSetRecvMsg = NULL;
    
    /* Reset the callback */
    pObject->pTerminate = NULL;

    /* Destroy the CCommThread instance */
    CWACPThread_Destroy(&CCommThread);
}

/*================================================================================
CCommSvr GetPort: Gets the current port from the passed in ConnArgs structure. 
================================================================================*/
uint16 CCommSvr_GetPort(const CCommSvr* pObject, const ConnArgs *pConnArgs)
{
    uint16 Port = 0;

    TOUCH(pObject);

    if (pConnArgs)
    {
#if defined(STACK_ETHERNET)
        if (pConnArgs->StackType == ETHERNET)
        {
            Port = ((CEthCommSvr*)(pConnArgs->pCommSvr))->Port;
        }
#endif

#if defined(STACK_SERIAL)
        if (pConnArgs->StackType == SERIAL)
        {
            Port = ((CSerialCommSvr*)(pConnArgs->pCommSvr))->Port;
        }
#endif
    }
    return (Port);
}

/*================================================================================
CCommSvr SetPort: Changes the current configuration to use the WACPPort. 
================================================================================*/
boolean CCommSvr_SetPort(const CCommSvr* pObject, const ConnArgs *pConnArgs)
{
    boolean Result = FALSE;

    TOUCH(pObject);

    if (pConnArgs)
    {
#if defined(STACK_ETHERNET)
        if (pConnArgs->StackType == ETHERNET)
        {
            CEthCommSvr_SetPort(pConnArgs->pCommSvr, pConnArgs->WACPPort);

            Result = TRUE;
        }
#endif

#if defined(STACK_SERIAL)
        if (pConnArgs->StackType == SERIAL)
        {
            CSerialCommSvr_SetPort(pConnArgs->pCommSvr, pConnArgs->WACPPort);

            Result = TRUE;
        }
#endif
    }

    return (Result);
}


/*================================================================================
CCommSvr Connect: Creates a socket that connects to the passed in SourceID and Port.
================================================================================*/
boolean CCommSvr_Connect(CCommSvr* pObject, ConnArgs *pConnArgs)
{
    boolean     Result = FALSE;

    if (pConnArgs)
    {
#if defined(STACK_ETHERNET)
        /* Initialize the appropriate CommSvr instance */
        if (pConnArgs->StackType == ETHERNET)
        {
            /* Allocate and initialize a new instance of the CEthCommSvr object */
            pConnArgs->pCommSvr = (CEthCommSvr*)SYS_MALLOC(sizeof(CEthCommSvr));
            if (pConnArgs->pCommSvr)
            {
                CEthCommSvr_Init(pConnArgs->pCommSvr, pObject->pSetRecvMsg, pObject->pTerminate);
                
                /* Update the configuration including SourceID for the CommSvr */
                ((EthAddr*)(pConnArgs->pConfig))->pSourceID = pConnArgs->SourceID;
                CEthCommSvr_SetConfig(pConnArgs->pCommSvr, (EthAddr*)((pConnArgs->pConfig)));

                /* Open up the communications channel */
                Result = CEthCommSvr_Connect(pConnArgs->pCommSvr, (uint8*)pConnArgs->SourceID, pConnArgs->RNDZPort);
                if (Result)
                {
                    /* Start up the receiving thread for the CommSvr. */
                    CWACPThread_StartThread(&CCommThread, CEthCommSvr_StartClient, pConnArgs->pCommSvr, pConnArgs->Priority);
                }
            }
        }
#endif

#if defined(STACK_SERIAL)
        if (pConnArgs->StackType == SERIAL)
        {
            /* Allocate and initialize a new instance of the CSerialCommSvr object */
            pConnArgs->pCommSvr = (CSerialCommSvr*)SYS_MALLOC(sizeof(CSerialCommSvr));
            if (pConnArgs->pCommSvr)
            {
                CSerialCommSvr_Init(pConnArgs->pCommSvr, pObject->pSetRecvMsg, pObject->pTerminate);
            
                /* Update the configuration including SourceID for the CommSvr */
                ((SerialAddr*)(pConnArgs->pConfig))->pSourceID = pConnArgs->SourceID;
                CSerialCommSvr_SetConfig(pConnArgs->pCommSvr, (SerialAddr*)((pConnArgs->pConfig)));

                /* Open up the communications channel */
                Result = CSerialCommSvr_Connect(pConnArgs->pCommSvr, (uint8*)pConnArgs->SourceID, pConnArgs->RNDZPort);
                if (Result)
                {
                    /* Start up the receiving thread for the CommSvr. */
                    CWACPThread_StartThread(&CCommThread, CSerialCommSvr_StartClient, (void*)pConnArgs->pCommSvr, pConnArgs->Priority);
                }
            }
        }
#endif
    }

    return Result;
}

/*================================================================================
CCommSvr Disconnect: terminates the current connection pointed to pConnArgs.
================================================================================*/
boolean CCommSvr_Disconnect(CCommSvr* pObject, ConnArgs *pConnArgs)
{
    boolean     Result = FALSE;

    TOUCH(pObject);

    if (pConnArgs)
    {
#if defined(STACK_ETHERNET)
        if (pConnArgs->StackType == ETHERNET)
        {
            /* Clean up the ethernet commsvr instance */
            CEthCommSvr_Destroy(pConnArgs->pCommSvr);
            SYS_FREE(pConnArgs->pCommSvr);
            pConnArgs->pCommSvr = NULL;
            Result = TRUE;
        }
#endif

#if defined(STACK_SERIAL)
        if (pConnArgs->StackType == SERIAL)
        {
            /* Clean up the serial commsvr instance */
            CSerialCommSvr_Destroy(pConnArgs->pCommSvr);
            SYS_FREE(pConnArgs->pCommSvr);
            pConnArgs->pCommSvr = NULL;
            Result = TRUE;
        }
#endif
    }

    return Result;
}

/*================================================================================
CCommSvr SendMessage:  Sends a message via the appropriate stack CommSvr.
================================================================================*/
boolean CCommSvr_SendMessage(const CCommSvr* pObject, const ConnArgs *pConnArgs, const uint8* pBuf, uint32 MsgSize)
{
    boolean Result = FALSE;

    TOUCH(pObject);
#if defined(STACK_ETHERNET)
    if (pConnArgs->StackType == ETHERNET)
    {
        if (pConnArgs->pCommSvr)
        {
            /* Send the message */
            Result = CEthCommSvr_SendMessage(pConnArgs->pCommSvr, pBuf, MsgSize);
        }
    }
#endif

#if defined(STACK_SERIAL)
    if (pConnArgs->StackType == SERIAL)
    {
        if (pConnArgs->pCommSvr)
        {
            /* Send the message */
            Result = CSerialCommSvr_SendMessage(pConnArgs->pCommSvr, pBuf, MsgSize);
        }
    }
#endif

    return(Result);
}




