/*-D************************************************************************* 
'*   This program source file is confidential and proprietary property of
'*    Welch Allyn, Inc. Unauthorized use or reproduction is prohibited.
'****************************************************************************
'****************************************************************************
'**
'****************************************************************************
'**  ////////////////////////////////////////////////////////////////////////
'**  '// CMsgSvr.c      :                                                  //
'**  '// Language       :   C                                              //
'**  '// Application    :   WACP                                           //
'**  '// Sources        :   MOMP Standard Library                          //
'**  '///////////////////////////////////////////////////////////////////////
'*-E************************************************************************/
#include "Target.h"
#include "CMsgSvr.h"
#include "DtData.h"
#include "CLinkTable.h"
#include "Wacp.h"

#if defined(ENCRYPTION_ON)
#include "AESCrypt.h"
#include "CEncryption.h"
#endif

void CMsgSvr_Init(CMsgSvr* pObject, uint16 Family, uint8 Genus, uint8 Species)
{
    if(pObject != NULL)
    {
        /*Initialize the structure*/
        memset(pObject, 0, sizeof(CMsgSvr));

        /* Set the ClassID */
        pObject->ClassID = MakeClassID(Family, Genus, Species);
    }
}

/*********************************************************************************
Returns the total number of bytes required to serialize the data.
*********************************************************************************/
uint32 CMsgSvr_SizeAll(const CMsgSvr* pObject)
{
    uint32 Size = 0;

    if(pObject != NULL)
    {
        Size  = sizeof(uint8)          /* Size of encryption/compression byte*/
                 + sizeof(uint32)         /*Size of the data array*/
                 + pObject->DataSize      /*data array size*/
                 + sizeof(uint16);        /* CRC bytes */

        /* ClassID + Payload Length(uint32 aggregate) + Payload */
        return sizeof(uint32) + sizeof(uint32) + Size;
    }
    else
        return Size;
}

/*********************************************************************************
Serializes the contents of the CMsgSvr object into the buffer passed in.
*********************************************************************************/
boolean CMsgSvr_Serialize(const CMsgSvr* pObject, const char* pSourceID, uint8* pBuf)
{    
    uint8*   pTemp = pBuf;
    boolean  Result = FALSE; 

    if((pObject != NULL) && (pTemp != NULL))
    {
        uint32 CRCSize;
        uint32 PayloadSize = (uint32)CMsgSvr_SizeAll(pObject) - MSGSVR_HEADER_SIZE;
        ConnArgs* pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceID);

        if (pConnArgs)
        {
            /* Serialize the DtData members */
            WriteInt32(&pTemp, pObject->ClassID);

            WriteInt32(&pTemp, PayloadSize);
            
            /* Serialize the CMsgSvr members */
            /* Set the Encryption byte */
            if (pObject->bEncryptionOn)
            {
                WriteInt8(&pTemp, pConnArgs->EncryptionType);
            }
            else
                WriteInt8(&pTemp, WACP_ENCRYPTION_NONE);

            /* Set the Data size */
            WriteInt32(&pTemp, pObject->DataSize);
            
            if (pObject->DataSize > 0)
            {
#if defined(ENCRYPTION_ON)
                /* Check to see if Encryption is enabled */
                if((pObject->bEncryptionOn) && (pConnArgs->EncryptionType))
                {   
                    /* Encrypt the byte stream after LnkTbl.EncryptionType up to (but not including) the CRC*/
                    ChunkAESEncrypt(pConnArgs->pKey, pObject->Data, pObject->Data, pObject->DataSize);
                }
#endif

                /* Set the Data */
                memcpy(pTemp, pObject->Data, pObject->DataSize);
                pTemp += pObject->DataSize;
            }

            /* Set the CRC */
            CRCSize = CMsgSvr_SizeAll(pObject) - sizeof(uint16);
            WriteInt16(&pTemp, CalculateCRC(CRC_SEED, pBuf, CRCSize));

            Result = TRUE;
        }/* if (pConnArgs) */
    }

    return Result; 
}
    
/*********************************************************************************
Used to deserialize the data elements of the object from a circular queue. This
excludes the class identifier.
*********************************************************************************/
boolean CMsgSvr_Deserialize(CMsgSvr* pObject, const char* pSourceID, uint8* pBuf)
{   
    uint8*  pTemp = pBuf;
    boolean Result = FALSE;
    byte    EncryptionType;

    TOUCH(pSourceID);

    if((pObject != NULL) && (pTemp != NULL))
    {
        uint32  PayloadSize;
        uint16  StoredCRC;
        uint16  CalcCRC;
        uint32  CRCSize;
        uint32  ClassID;

        /* Verify the Family and Genus match */
        ClassID = ReadInt32(&pTemp);
        if ((ClassID & 0xFFFFFF00) == (pObject->ClassID & 0xFFFFFF00))
        {
            /* Set the ClassID to update the species */
            pObject->ClassID = ClassID;

            /* Get the size of the payload */
            PayloadSize = ReadInt32(&pTemp);

            TOUCH(PayloadSize);
            
            /* Get the Encrypt/Compression Byte */
            EncryptionType = ReadInt8(&pTemp);
            
            /* Deserialize the CMsgSvr dynamic members */
            /* Get the DataSize */
            pObject->DataSize = ReadInt32(&pTemp);
            
            if (pObject->DataSize > 0)
            {
                pObject->Data = (uint8*)SYS_REALLOC(pObject->Data, pObject->DataSize);
                if (!pObject->Data)
                    return FALSE;

                /* Copy the data into Data buffer */
                memcpy(pObject->Data, pTemp, pObject->DataSize);
                pTemp += pObject->DataSize;

#if defined(ENCRYPTION_ON)
                /* Check the encryption type */
                if(EncryptionType)
                {
                    ConnArgs* pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceID);
                    if (pConnArgs)
                    {
                        /* Decrypt the byte stream after EncryptionType up to (but not including) the CRC*/
                        ChunkAESDecrypt(pConnArgs->pKey, pObject->Data, pObject->Data, pObject->DataSize);
                    }
                }
#endif
            }

            /* Calculate the CRC */
            CRCSize = CMsgSvr_SizeAll(pObject) - sizeof(uint16);
            CalcCRC = CalculateCRC(CRC_SEED, pBuf, CRCSize);

            /* Get the stored CRC */
            StoredCRC = ReadInt16(&pTemp);

            /* Check the CRC */
            if (CalcCRC == StoredCRC) 
                Result = TRUE;
        }
    }

    return Result;
}

void CMsgSvr_SetEncryption(CMsgSvr* pObject, boolean bEncryptFlag)
{
    if(pObject != NULL)
    {
        pObject->bEncryptionOn = bEncryptFlag;
    }
}

/*********************************************************************************
SetData: Adds a buffer to the CMsgSvr dynamic data buffer. If encryption is defined
    the size of the data is padded to a 16 byte boundary.
    If pSrcBuf is NULL, then clear the pDestBuf.
*********************************************************************************/
boolean CMsgSvr_SetData(CMsgSvr* pObject, const char* pSourceID, const uint8* pDataBuf, uint32 Size)
{
    boolean Result = FALSE;
    
    TOUCH(pSourceID);

    if(pObject != NULL)
    {
        if((pDataBuf) && (Size > 0))
        {
            uint32 PaddedSize = Size;

#if defined(ENCRYPTION_ON)
            ConnArgs* pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceID);
            if (pConnArgs)
            {
                /* Check the encryption type */
                if((pObject->bEncryptionOn) && (pConnArgs->EncryptionType))
                {
                    /* When using AES encryption, pad Data so that it lines up on 16 byte boundary */            
                    PaddedSize = Size + (16 - (Size % 16));
                }
            }
#endif

            pObject->Data = (uint8*)SYS_REALLOC(pObject->Data, PaddedSize);
            if(pObject->Data)
            {
                memcpy(pObject->Data, pDataBuf, PaddedSize);

                /* Set the DataSize Var */
                pObject->DataSize = PaddedSize;

                Result = TRUE;
            }
            else
                pObject->DataSize = 0;


        }
        else
        {
            SYS_FREE(pObject->Data);
            pObject->Data = NULL;
            pObject->DataSize = 0;
            Result = TRUE;
        }
    }

    return (Result);
}


