/*-D************************************************************************* 
'*   This program source file is confidential and proprietary property of
'*    Welch Allyn, Inc. Unauthorized use or reproduction is prohibited.
'*                         Copyright (c) 2004-2005 by
'*                              Welch Allyn Inc.
'*                            All rights reserved
'****************************************************************************
'****************************************************************************
'**
'****************************************************************************
'**  ////////////////////////////////////////////////////////////////////////
'**  '// CTrapObjSvr.c  :                                                  //
'**  '// Version        :   1.00                                           //
'**  '// Language       :   C                                              //
'**  '// Application    :   WACP                                           //
'**  '// Sources        :   MOMP Standard Library                          //
'**  '///////////////////////////////////////////////////////////////////////
'*-E************************************************************************/
#include "CTrapObjSvr.h"
#include "Wacp.h"
#include "CMOMPCallBack.h"

/*********************************************************************************
CTrapObjSvr Init: Start up routine
*********************************************************************************/
void CTrapObjSvr_Init(CTrapObjSvr *pObject)
{
    if (pObject != NULL)
    {
        /* Initialize the general error servers and error objects */
        CErrEStd_Init(&pObject->TrapErrObj);
        SetSpecies(&(pObject->TrapErrObj.ClassID), CERR_SpERRWACPSTACK);

        CErrSvr_Init(&pObject->TrapInErrSvr, GnERROR);
        CErrSvr_Init(&pObject->TrapOutErrSvr, GnERROR);
        
        /* Initialize the message servers */
        CMsgSvr_Init(&pObject->TrapMsgSvr, FmTRAP, GnRESPONSE, CMSG_SpNONE);

        /* Initialize the mutex for accessing the SendMsg function */
        CWACPMutex_Init(&pObject->ErrorMutex, NULL);
    }
}

/*********************************************************************************
CTrapObjSvr Destroy: Clean up routine.
*********************************************************************************/
void CTrapObjSvr_Destroy(CTrapObjSvr *pObject)
{
    if (pObject != NULL)
    {
        /* Destroy the ErrorMutex access mutex */
        CWACPMutex_Destroy(&pObject->ErrorMutex);
    }
}

/*================================================================================
CTrapObjSvr TxErrMsg : Helper routine that transmits the contents of a CErrSvr object
================================================================================*/
boolean CTrapObjSvr_TxErrMsg(const char *pSourceID, const CErrSvr *pCErrSvr)
{
    uint32  BufSz;
    boolean Result = FALSE;
    uint8   *pTempBuf = NULL;

    /* Get the Size of the serialized byte Stream */
    BufSz = CErrSvr_SizeAll(pCErrSvr);

    /* (Re)Allocate space for the serialized Buffer */
    pTempBuf = (uint8*)SYS_MALLOC(BufSz);
    if(pTempBuf != NULL)
    {
        /* Serialize the CErrSvr object */
        CErrSvr_Serialize(pCErrSvr, pTempBuf);

        /* Send the message */
        Result = Interchange_SendPacket(pSourceID, pTempBuf, BufSz);
      
        /* Free Outbuf */
        SYS_FREE(pTempBuf);
        pTempBuf = NULL;
    }

    return(Result);
}

/*================================================================================
CTrapObjSvr TxError : Sends a error message ACK/NACK

    The Species in the CErrSvr is a direct match with the species of the CErrEStd.

    CErrEStd                        CErrSvr
    CERRESTD_SpERRNONE = 0,         CES_SpNONE = 0,
    CERRESTD_SpERRWACPSTACK,        CES_SpWACPSTACK,
    CERR_SpERRAPPINTERNAL,          CES_SpERRAPPINTERNAL,
================================================================================*/
boolean CTrapObjSvr_TxError(CTrapObjSvr *pObject, const char *pSourceID, uint32 ClassID, 
                            uint8 ErrorSpecies, uint8* pBuf, uint32 Size)
{
    boolean Result = FALSE;

    if (pObject != NULL)
    {
        if ((pBuf != NULL) && (Size > 0))
        {
            CWACPMutex_Lock(&pObject->ErrorMutex, INFINITE);

            /* Load the last TrapErrObj sent out by deserializing the data 
            * The request for last error will return the data held by
            * pErrObj so we must update pObject each time an error message
            * is sent.
            */
            CErrEStd_Deserialize(&pObject->TrapErrObj, pBuf);

            CWACPMutex_Release(&pObject->ErrorMutex);
        }

        /* Set the ClassID */
        pObject->TrapOutErrSvr.ErrorClassID = ClassID;

        /* Set the Species to the right error type */
        SetSpecies(&pObject->TrapOutErrSvr.ClassID, ErrorSpecies);
        
        /* Set the data size */
        pObject->TrapOutErrSvr.DataSize = Size;
        
        /* Copy the data into the dynamic data buffer */
        Result = SetData(&pObject->TrapOutErrSvr.Data, pBuf, (uint16)Size);
        if (Result)
        {
#if defined(TRACE_SERVER_ON_STATUS)
            CTraceSvr_tprintf(&TraceServer, TRACE_Type_SEND,
                "CTrapObjSvr", "TxError", (uint16)(ClassID >> 16), (uint8)(ClassID >> 8), (uint8)ClassID, 
                "ErrorSpecies = %d\n", ErrorSpecies);
#endif
            /* Send the message */
            Result = CTrapObjSvr_TxErrMsg(pSourceID, &pObject->TrapOutErrSvr);

            /* Free pObject->TrapOutErrSvr.Data */
            SYS_FREE(pObject->TrapOutErrSvr.Data);
            pObject->TrapOutErrSvr.Data = NULL;
        }
    }
    return Result;
}

/*================================================================================
TRAP SendError: Internal call for sending a CERRESTD_SpWACPSTACK error to the 
    application. It takes the ClassID and ErrorCode to create a CErrEStd object 
    byte stream and sends it to the application.
================================================================================*/
boolean CTrapObjSvr_SendError(CTrapObjSvr *pObject, const char *pSourceID, uint32 ClassID, 
                              uint16 ErrorCode, boolean bNotify)
{
    uint32 Result = 0;

    if (pObject && bNotify)
    {
        uint32 MsgSize = 0;
        uint8  *pTempBuf = NULL;

        CWACPMutex_Lock(&pObject->ErrorMutex, INFINITE);

        /* Create a new instance of the error object */
        CErrEStd_Init(&pObject->TrapErrObj);
        SetSpecies(&pObject->TrapErrObj.ClassID, CERR_SpERRWACPSTACK);

        /* Set the ErrorCode */
        pObject->TrapErrObj.m_ErrorCode = ErrorCode;

        /* Set the ClassID */
        pObject->TrapErrObj.m_ErrClassID = ClassID;

        /* Set the Error Description */
        memcpy(pObject->TrapErrObj.m_Description, (char*)CErrEStd_TypStackErrorTbl[ErrorCode], 
            sizeof(pObject->TrapErrObj.m_Description));

        /* Get the Size of the error object */
        MsgSize = CErrEStd_SizeAll(&pObject->TrapErrObj);

        /* (Re)Allocate space for the serialized Buffer */
        pTempBuf = (uint8*)SYS_MALLOC(MsgSize); 
        if(pTempBuf != NULL)
        {
            /* Serialize the error object */
            Result = CErrEStd_Serialize(&pObject->TrapErrObj, pTempBuf);
            if (Result)
            {
                RecvError(pSourceID, CERR_SpERRWACPSTACK, ClassID, pTempBuf);
            }
            SYS_FREE(pTempBuf);
            pTempBuf = NULL;
        }
        
        CWACPMutex_Release(&pObject->ErrorMutex);
    }

    return (Result ? TRUE : FALSE);
}

/*================================================================================
CTrapObjSvr RecvMsgHandler: Decodes the Family, Genus, and Species tag and deserializes
     the messages with the appropriate object.
================================================================================*/
boolean CTrapObjSvr_RecvMsgHandler (CTrapObjSvr *pObject, const char *pSourceID, uint8* pBuf)
{
    uint8   Genus;
    uint8   Species;
    boolean RetVal=TRUE;
#if defined(TRACE_SERVER_ON_STATUS)
    uint16   Family = GetFamily(pBuf);;
#endif

    if (pObject != NULL)
    {
        /* Extract the Genus */
        Genus = GetGenus(pBuf);

        /* Extract the Species */
        Species = GetSpecies(pBuf);

        /* Genus Router */
        switch (Genus)
        {           
            case GnERROR:
            {
                /* Clear the data field */
                SetData(&pObject->TrapInErrSvr.Data, NULL, 0);

                /* Deserialize the data and check for errors */
                if (CErrSvr_Deserialize(&pObject->TrapInErrSvr, pBuf))
                {
                    uint8   *pErrBuf = pObject->TrapInErrSvr.Data;

                    /* Setting pErrBuf to NULL means pObject is a simple ACK/NACK message */
                    if ( pObject->TrapInErrSvr.DataSize == 0 )
                        pErrBuf = NULL;

#if defined(TRACE_SERVER_ON_STATUS)
                    if (Species == CES_SpNONE)
                    {
                        CTraceSvr_tprintf(&TraceServer, TRACE_Type_STATUS,
                                "CTrapObjSvr", "RecvMsgHandler", Family, Genus, Species, 
                                "Received ACK Message\n");
                    }
                    else if (Species == CES_SpWACPSTACK)
                    {
                        CTraceSvr_tprintf(&TraceServer, TRACE_Type_STATUS,
                                "CTrapObjSvr", "RecvMsgHandler", Family, Genus, Species, 
                                "Received Stack NACK Message\n");
                    }
                    else if (Species == CES_SpERRMSGPROC)
                    {
                        CTraceSvr_tprintf(&TraceServer, TRACE_Type_STATUS,
                                "CTrapObjSvr", "RecvMsgHandler", Family, Genus, Species, 
                                "Received Message Processing NACK Message\n");
                    }
                    else if (Species == CES_SpERRAPPINTERNAL)
                    {
                        CTraceSvr_tprintf(&TraceServer, TRACE_Type_STATUS,
                                "CTrapObjSvr", "RecvMsgHandler", Family, Genus, Species, 
                                "Received Application NACK Message\n");
                    }
#endif
                    RecvError(pSourceID, Species, pObject->TrapInErrSvr.ErrorClassID, pErrBuf);

                    /* Free pObject->TrapInErrSvr.Data */
                    SYS_FREE(pObject->TrapInErrSvr.Data);
                    pObject->TrapInErrSvr.Data = NULL;
                }
                
                break;
            }/* case GnERROR */

            case GnREQUEST:
            {
                CMsgSvr_Init(&pObject->TrapMsgSvr, FmTRAP, GnREQUEST, CMSG_SpNONE);

                /* Clear the data field */
                SetData(&pObject->TrapMsgSvr.Data, NULL, 0);

                /* Deserialize the data and check for errors */
                if (CMsgSvr_Deserialize(&pObject->TrapMsgSvr, pSourceID, pBuf))
                {
                    if (Species == TRAP_CRST_SpGET_LASTERROR)
                    {
                        uint8 *pTempBuf = NULL;

                        /* Get the Size of the error object */
                        uint32 MsgSize = 0;
                        
                        CWACPMutex_Lock(&pObject->ErrorMutex, INFINITE);

                        MsgSize = CErrEStd_SizeAll(&pObject->TrapErrObj);

                        /* (Re)Allocate space for the serialized Buffer */
                        pTempBuf = (uint8*)SYS_MALLOC(MsgSize);
                        if(pTempBuf != NULL)
                        {
                            /* Serialize the CErrSvr object */
                            CErrEStd_Serialize(&pObject->TrapErrObj, pTempBuf);

                            /* Resend the last error message */
                            SEND_RESPONSE(pSourceID, FmTRAP, TRAP_CRSP_SpPUT_LASTERROR, pTempBuf, MsgSize, FALSE);

                            /* Free Outbuf */
                            SYS_FREE(pTempBuf);
                            pTempBuf = NULL;
                        }

                        CWACPMutex_Release(&pObject->ErrorMutex);

                    }/* if (Species == TRAP_CRSP_SpGET_LASTERROR) */
                }/* if (CMsgSvr_Deserialize(&pObject->TrapRequest, pSourceID, pBuf)) */

                /* Free pObject->TrapMsgSvr.Data */
                SYS_FREE(pObject->TrapMsgSvr.Data);
                pObject->TrapMsgSvr.Data = NULL;
                
                break;
            }/* case GnREQUEST */

            case GnRESPONSE:
            {
                CMsgSvr_Init(&pObject->TrapMsgSvr, FmTRAP, GnRESPONSE, CMSG_SpNONE);

                /* Clear the data field */
                SetData(&pObject->TrapMsgSvr.Data, NULL, 0);

                /* Deserialize the data and check for errors */
                if (CMsgSvr_Deserialize(&pObject->TrapMsgSvr, pSourceID, pBuf))
                {
                    if (Species == TRAP_CRSP_SpPUT_LASTERROR)
                    {
                        uint8   *pErrBuf = pObject->TrapMsgSvr.Data;

                        RecvError(pSourceID, GetSpecies(pErrBuf), MakeClassID(FmTRAP, GnRESPONSE, CMSG_SpNONE), pErrBuf);
                    }
                }/* if (CMsgSvr_Deserialize(&pObject->TrapResponse, pSourceID, pBuf)) */

                /* Free pObject->TrapMsgSvr.Data */
                SYS_FREE(pObject->TrapMsgSvr.Data);
                pObject->TrapMsgSvr.Data = NULL;
                
                break;
            }/* case GnRESPONSE */

            default:
                break;

        }/* switch (Genus)*/
    }
    return RetVal;
}
    




