/*-D************************************************************************* 
'*   This program source file is confidential and proprietary property of
'*    Welch Allyn, Inc. Unauthorized use or reproduction is prohibited.
'****************************************************************************
'****************************************************************************
'**
'****************************************************************************
'**  ////////////////////////////////////////////////////////////////////////
'**  '// Interchange.c  :                                                  //
'**  '// Language       :   C                                              //
'**  '// Application    :   WACP                                           //
'**  '// Sources        :   MOMP Standard Library                          //
'**  '///////////////////////////////////////////////////////////////////////
'*-E************************************************************************/
#include "Interchange.h"
#include "CWACPThread.h"
#include "CWACPSemaphore.h"
#include "CWACPQueue.h"
#include "CConObjSvr.h"
#include "Wacp.h"

#if defined(TRACE_SERVER_ON)
char TraceString[] = "SourceID %s\n\nMessage Data:\n%A";
char TraceAuthString[] = "Unauthenticated Message SourceID %s\n\nMessage Data:\n%A";
char TraceAuthString2[] = "Unauthenticated Message SourceID %s Port %d\n\nMessage Data:\n%A";
char CRCTraceString[] = "\nCRC Error. PacketLen - %d TimeStamp = %d\nSourceID %s\n\nMessage Data:\n%A";
#endif

/* Callback pointer */
typedef void (*FuncPtr)(char* pSourceID, uint8* pBuf, uint32 BufSize);
static FuncPtr pCallback;

/* Interchange globals */
static uint8* pOutBuffer;
static CWACPSemaphore  DataSem;
static CWACPQueue      MsgQueue;

/* Trace Server globals */
#if defined(TRACE_SERVER_ON)
CTraceSvr TraceSvr;
uint32 StartTime;
#endif

/* Connection globals */
static CCommSvr CommSvr;

/* Thread objects */
static CWACPThread RecvThread;
static boolean bThreadStatus;

/********************************************************************************************
Interchange_Init - Entry point for starting up the Interchange layer. All threads created by
    Interchange will be set to value passed into Priority. This starts the receiving thread
    and initializes the data semaphore, WACPQueue, and CommSvr.
********************************************************************************************/
void Interchange_Init(int32 Priority)
{
#if defined(TRACE_SERVER_ON)
    StartTime = GetTickCount();
#endif

    bThreadStatus = TRUE;
    
    /* Initialize the CCommSvr instance. All connections will go through this object */
    CCommSvr_Init(&CommSvr, Interchange_RecvMsg, Interchange_Terminate);

    /* Initialize the Recv thread*/
    CWACPThread_Init(&RecvThread, EXCG_RECV_THREAD);

    /* Initialize the data semaphore that is used to indicate a new message is available for
     * the application
     */
    CWACPSemaphore_Init(&DataSem, NULL, 0, MAX_QUEUE_SIZE);
    
    /* Initialize the message queue */
    CWACPQueue_Init(&MsgQueue, MAX_QUEUE_SIZE);

    /* Start up the Receive thread */
    CWACPThread_StartThread(&RecvThread, Interchange_RecvThread, NULL, Priority);

#if defined(TRACE_SERVER_ON)
    /* Create the trace server */
    CTraceSvr_Init(&TraceSvr);

    /* Open the Trace file here so we can start writing to it */
    CTraceSvr_OpenFile(&TraceSvr, TRACE_EXCG_LOG, "Interchange", NULL);
    CTraceSvr_GenerateXSLFile(&TraceSvr, NULL);
    CTraceSvr_EnableLogging(&TraceSvr, TRUE, TRUE);
#endif
}

/********************************************************************************************
Interchange_Destroy - Cleanup routine for destroying the Interchange layer. All object 
    instantiated by Interchange are destroyed and a fake semaphore is raised after setting 
    the ThreadStatus to FALSE to the receive thread to terminate the thread.
********************************************************************************************/
void Interchange_Destroy(void)
{
    bThreadStatus = FALSE;

    /* Raise a semaphore for the receiving thread to terminate */
    CWACPSemaphore_Raise(&DataSem);

    /* Pause to allow the receiving thread a chance to terminate */
    SYS_SLEEP(1);

    /* Free pOutBuffer */
    SYS_FREE(pOutBuffer);
    pOutBuffer = NULL;

    /* Destroy the CCommSvr instance */
    CCommSvr_Destroy(&CommSvr);

    /* Destroy the queue buffer */
    CWACPQueue_Destroy(&MsgQueue);

    /* Destroy Recv thread */
    CWACPThread_Destroy(&RecvThread);

#if defined(TRACE_SERVER_ON)
    CTraceSvr_CloseFile(&TraceSvr);
#endif 
    
}

/********************************************************************************************
InterChange_ProcessPackets - Dequeues a message from the CWACPQueue instance. The message
    SourceID is used to verify the authentication. If authenticated or a RNDZ message, then 
    the callback into the application layer is called.
********************************************************************************************/
static uint32 InterChange_ProcessPackets(void)
{
    uint32 Result = FALSE;
    QPACKET* pData = NULL;

    /* Dequeue the new message */
    pData = CWACPQueue_Dequeue(&MsgQueue);
    if (pData)
    {
        /* Use the SourceID to get the LinkTable entry */
        ConnArgs* pConnArgs = CLinkTable_GetEntry(&LinkTable, pData->pSourceID);
        if (pConnArgs)
        {
            pCallback(pData->pSourceID, pData->pDataBuf, pData->DataSize);
        }

        SYS_FREE(pData->pDataBuf);
        SYS_FREE(pData);
    }
#if defined(TRACE_SERVER_ON)
    else
    {
        CTraceSvr_tprintf(&TraceSvr, TRACE_Type_ERROR,
                      "Interchange", "ProcessPackets", 0, 0, 0, 
                      "Queue is empty\n");
    }
#endif
    return (Result);
}

/********************************************************************************************
Interchange_RecvThread - Main processing thread for incoming messages. This method uses
    InterChange_ProcessPackets to process the message once the data semaphore is fired.
********************************************************************************************/
uint32 WACPAPI Interchange_RecvThread(void* pData)
{
    uint32 Result;
    TOUCH(pData);

    /* Wait for the DataSem to fire for processing a message or until we time out.*/
    Result = CWACPSemaphore_Wait(&DataSem, INFINITE);

    while(bThreadStatus == TRUE)
    {
        if (Result == WAIT_SUCCESS)
        {
            /* Check for new messages */
            InterChange_ProcessPackets();
        }

        /* Wait for the DataSem to fire for processing a message or until we time out.*/
        Result = CWACPSemaphore_Wait(&DataSem, INFINITE);
    }

#if defined(TRACE_SERVER_ON_STATUS)
    CTraceSvr_tprintf(&TraceSvr, TRACE_Type_STATUS,
                      "Interchange", "RecvThread", 0, 0, 0, 
                      "RecvThread terminating\n");
#endif

    /* Clean up the Exit Event object */
    CWACPSemaphore_Destroy(&DataSem);

    return TRUE;
}

/********************************************************************************************
Interchange_RecvMsg - Callback used by the CommSvr instances to queue up an incoming message.
********************************************************************************************/
boolean Interchange_RecvMsg(char* pSourceID, uint8* pBuf, uint32 MsgSize)
{
    boolean Result = TRUE;
    PACKET  Pack;
    uint8*  pTemp;
    uint8*  pDataBuf = NULL;
    uint32  DataLength;

    /* Set the Temp Pointer beyond the PREAMBLE of the byte stream */
    pTemp = (uint8*)pBuf + PREAMBLE_SIZE;

    /* Read the Packet Length */
    Pack.PacketLen = ReadInt32(&pTemp);

    /* Read the Port */
    Pack.Port = ReadInt16(&pTemp);

    /* Packet Data Length */
    DataLength = (Pack.PacketLen - PACKET_LITE_HEADER_SIZE) - CRCSIZE;

    /* Check for a valid buffer size */
    if (DataLength < RCVBUFSIZE)
    {
        pDataBuf = SYS_MALLOC(DataLength);
        if (pDataBuf)
        {
            memcpy(pDataBuf, pTemp, DataLength);
        }
    }
    else
    {
#if defined(TRACE_SERVER_ON)
        CTraceSvr_tprintf(&TraceSvr, TRACE_Type_ERROR,
                      "Interchange", "RecvMsg", 0, 0, 0, 
                      "Max Data size Error Length = %d", DataLength);
#endif
        return (FALSE);
    }

    /* Move the pointer past the data */
    pTemp += DataLength;

    /* Get the CRC */
    Pack.CRC = ReadInt16(&pTemp);

    /* Verify the CRC */
    if (Pack.CRC == CalculateCRC(CRC_SEED, (uint8*)pBuf, Pack.PacketLen - CRCSIZE))
    {       
#if defined(TRACE_SERVER_ON) && defined(TRACE_SERVER_ON_STATUS)
        Interchange_DebugPacket(pBuf, Pack.PacketLen);
#endif
    }
    else
    {
#if defined(TRACE_SERVER_ON)
        uint32 TimeStamp = GetTickCount();
        CTraceSvr_tprintf(&TraceSvr, TRACE_Type_ERROR,
                      "Interchange", "RecvMsg", 0, 0, 0, 
                      CRCTraceString, Pack.PacketLen, TimeStamp - StartTime, pSourceID, pBuf, MsgSize);
        Interchange_DebugPacket(pBuf, Pack.PacketLen);
#endif
        /* Set the length to 0 to specify a CRC error to the application */
        DataLength = 0;
    }

    if (pDataBuf)
    {
        boolean bRejectFlag = FALSE;

        /* Only allow WACP messages if RNDZ is disabled, RNDZStatus is TRUE, or if this is a RNDZ message */
        ConnArgs* pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceID);

        /* Test that we have a valid Port number and, if RNDZ is enabled,
         * that RNDZ is complete before allowing any WACP messages.
         */
        if (pConnArgs->bRNDZDisabled == FALSE)
        {
            bRejectFlag = (boolean)(!(Pack.Port == pConnArgs->RNDZPort || Pack.Port == pConnArgs->WACPPort) ||
                       (((Pack.Port == pConnArgs->WACPPort) && (pConnArgs->bRNDZStatus == FALSE))));
        }

        if (bRejectFlag == FALSE)
        {
            /* Queue up the new message */
            if (CWACPQueue_Enqueue(&MsgQueue, pSourceID, pDataBuf, DataLength) == TRUE)
            {
                /* Raise a semaphore to notify the receiving thread of the new message */
                CWACPSemaphore_Raise(&DataSem);
            }
            else
            {
                SYS_FREE(pDataBuf);
            }
        }
        else
        {
            SYS_FREE(pDataBuf);
#if defined(TRACE_SERVER_ON)
            CTraceSvr_tprintf(&TraceSvr, TRACE_Type_ERROR,
                      "Interchange", "RecvMsg", 0, 0, 0, 
                      TraceAuthString2, pSourceID, Pack.Port, 
                      pBuf, Pack.PacketLen);
#endif
        }
    }

#if defined(TRACE_SERVER_ON)
    else
    {
        CTraceSvr_tprintf(&TraceSvr, TRACE_Type_ERROR,
                      "Interchange", "RecvMsg", 0, 0, 0, 
                      "Unable to allocate memory for Message SourceID %s\n\nMessage Data:\n%A",
                      pSourceID, pBuf, Pack.PacketLen);
    }
#endif

#if !defined(TRACE_SERVER_ON)
    TOUCH(MsgSize);
#endif

    return(Result);
}

/********************************************************************************************
Interchange_Connect - Calls CommSvr_Connect to open a connection to the communications channel
    as specified in pConfig. This will skip any SourceIDs that are already in a connected 
    state. This only establishes the physical connection and does not start RNDZ 
    authentication, which is done by the application.
********************************************************************************************/
boolean Interchange_Connect(ConnArgs** pConfig, uint8 NumConnections)
{
    boolean Result = FALSE;
    uint16 i = 0;

    /* Cycle through all connections */
    for (i=0; i < NumConnections; i++)
    {
        if (pConfig[i]->bConnected == FALSE)
        {
            /* Start the connection */
            Result = CCommSvr_Connect(&CommSvr, pConfig[i]);
            if (Result)
            {
                /* Update the connection status */
                pConfig[i]->bConnected = TRUE;
            }
#if defined(TRACE_SERVER_ON)
            else
            {
                CTraceSvr_tprintf(&TraceSvr, TRACE_Type_ERROR,
                            "Interchange", "Connect", 0, 0, 0, 
                            "CCommSvr_Connect failed to connect to %s\n", pConfig[i]->SourceID);
            }
#endif
        }
#if defined(TRACE_SERVER_ON)
        else
        {
            CTraceSvr_tprintf(&TraceSvr, TRACE_Type_ERROR,
                        "Interchange", "Connect", 0, 0, 0, 
                        "%s is already connected\n", pConfig[i]->SourceID);
        }
#endif
    }

    return(Result);
}

/********************************************************************************************
Interchange_Disconnect - Closes the communication channel that corresponds to the SourceID
    passed in.
********************************************************************************************/
boolean Interchange_Disconnect(const char* pSourceID)
{
    boolean Result = FALSE;
    ConnArgs* pConnArgs = NULL;

    /* Find the entry in the linktable */
    pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceID);
    if ((pConnArgs) && (pConnArgs->bConnected == TRUE))
    {
        Result = CCommSvr_Disconnect(&CommSvr, pConnArgs);
        if (Result == TRUE)
        {
            /* Reset the RNDZ authentication */
            pConnArgs->bRNDZStatus = FALSE;

            /* Reset the connection status */
            pConnArgs->bConnected = FALSE;

            /* Destroy the queue buffer to remove any remnant messages */
            CWACPQueue_Destroy(&MsgQueue);

            /* Reinitialize the message queue for the next connection */
            CWACPQueue_Init(&MsgQueue, MAX_QUEUE_SIZE);

            /* Free up the encryption key */
            SYS_FREE(pConnArgs->pKey);
            pConnArgs->pKey = NULL;

            /* Free up the GUID object */
            SYS_FREE(pConnArgs->pGUIDStd);
            pConnArgs->pGUIDStd = NULL;
        }
#if defined(TRACE_SERVER_ON)
        else
        {
            CTraceSvr_tprintf(&TraceSvr, TRACE_Type_ERROR,
                    "Interchange", "Disconnect", 0, 0, 0, 
                    "CCommSvr_Disconnect failed to disconnect %s\n", pSourceID);
        }
#endif
    }
    else
    {
        Result = FALSE;
#if defined(TRACE_SERVER_ON_STATUS)
        CTraceSvr_tprintf(&TraceSvr, TRACE_Type_ERROR,
                    "Interchange", "Disconnect", 0, 0, 0, 
                    "Unable to find %s\n", pSourceID);
#endif
    }

    return (Result);
}

/********************************************************************************************
Interchange_UpdateStatus - This updates the Port setting for the CommSvr which switches
    all communications from using the RNDZ Port to the WACP Port.
********************************************************************************************/
boolean Interchange_UpdateStatus(const char* pSourceID)
{
    boolean Result=TRUE;
    ConnArgs* pConnArgs = NULL;
    
    /* Find the entry in the linktable */
    pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceID);
    if (pConnArgs)
    {
        /* Update the CommSvr port to use the WACPPort */
        CCommSvr_SetPort(&CommSvr, pConnArgs);
    }

    return (Result);
}

/********************************************************************************************
Interchange_SendPacket - Packetizes the data buffer and uses the pSourceID to send the data 
    through the CommSvr layer.
********************************************************************************************/
boolean Interchange_SendPacket(const char* pSourceID, const uint8* pBuf, uint32 BufSize)
{
    boolean Result=TRUE;
    uint32 PacketLen;

    /* Set the Size of the entire packet */
    PacketLen = PACKET_LITE_HEADER_SIZE + BufSize + CRCSIZE;

    /* Allocate room for the serialize packet Stream freed in EvaluateSend*/
    pOutBuffer = (uint8*)SYS_REALLOC(pOutBuffer, PacketLen);
    if (pOutBuffer)
    {
        ConnArgs* pConnArgs = NULL;
        uint8* pTemp = pOutBuffer; 

        pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceID);
        if ((pConnArgs) && (pConnArgs->bConnected == TRUE))
        {
            /* Serialize the packet header struct */
            /* Set the PREAMBLE */    
            WriteInt8(&pTemp, CTL_W);
            WriteInt8(&pTemp, CTL_A);
            WriteInt8(&pTemp, CTL_L);

            /* Write the Packet Length */ 
            WriteInt32(&pTemp, PacketLen);

            /* Write the Port */
            WriteInt16(&pTemp, CCommSvr_GetPort(&CommSvr, pConnArgs));

            /* Copy the data Stream */
            memcpy((char*)pTemp, pBuf, BufSize);
            pTemp += BufSize;
            
            /* Write the CRC */
            WriteInt16(&pTemp, CalculateCRC(CRC_SEED, pOutBuffer, PacketLen - CRCSIZE));

            /* Send the data */
            Result = CCommSvr_SendMessage(&CommSvr, pConnArgs, pOutBuffer, PacketLen);

#if defined(TRACE_SERVER_ON)      
            if (!Result)
            {
                CTraceSvr_tprintf(&TraceSvr, TRACE_Type_ERROR,
                      "Interchange", "Send Packet", 0, 0, 0, 
                      "CCommSvr_SendMessage failed for %s", pConnArgs->SourceID);
            }
#endif

#if defined(TRACE_SERVER_ON) && defined(TRACE_SERVER_ON_STATUS)      
            CTraceSvr_tprintf(&TraceSvr, TRACE_Type_SEND,
                      "Interchange", "Send Packet", 0, 0, 0, 
                      TraceString, pSourceID, (char*)pOutBuffer, PacketLen);
#endif
        }

    }/* if (pOutBuffer) */
    else
    {
#if defined(TRACE_SERVER_ON)
        CTraceSvr_tprintf(&TraceSvr, TRACE_Type_ERROR,
                      "Interchange", "Send Packet", 0, 0, 0, 
                      "Send Message Failed");
#endif
    }

    return (Result);
}


/********************************************************************************************
Interchange_RegisterCallback - Registers the callback for routing incoming messages to the 
    application layer.
********************************************************************************************/
boolean Interchange_RegisterCallback(void (*DataReadyEvent)(char* pSourceID, uint8* pBuf, uint32 BufSize))
{
    boolean Result=TRUE;
    
    pCallback = DataReadyEvent;
    
    return (Result);
}

#if defined(TRACE_SERVER_ON)
/********************************************************************************************
Interchange_DebugPacket - Debug method for tracing the data packet in a more readable layout.
********************************************************************************************/
void Interchange_DebugPacket(uint8* pBuf, uint32 PacketLength)
{
    uint32 i = 0;

    SYS_PRINTF("\n\nPacket:\n");

    SYS_PRINTF("Preamble: \t%.2X %.2X %.2X\n", pBuf[0], pBuf[1], pBuf[2]);
    SYS_PRINTF("PacketLength: \t%.2X %.2X %.2X %.2X\n", pBuf[3], pBuf[4], pBuf[5], pBuf[6]);
    SYS_PRINTF("Port: \t\t%.2X %.2X\n", pBuf[7], pBuf[8]);
    if ((ReadInt16Ext(pBuf + 9) < FmMAX) && (pBuf[11] < GnMAX))
        SYS_PRINTF("ClassID: \t%s %s %d\n\n", FmStringTable[ReadInt16Ext(pBuf + 9)], GnStringTable[pBuf[11]], pBuf[12]);
    else
        SYS_PRINTF("ClassID: \tF:%d G:%d S:%d\n\n", ReadInt16Ext(pBuf + 9), pBuf[11], pBuf[12]);
    for (i=9; i < PacketLength - 2;i++)
    {
        if (i % 25 == 0)
            SYS_PRINTF("\n");
        SYS_PRINTF("%.2X ", (uint8)pBuf[i]);
    }

    SYS_PRINTF("\n\nStoredCRC: \t%.4X\n", ReadInt16Ext(pBuf + (PacketLength - 2)));
    SYS_PRINTF("CalcCRC: \t%.4X\n", CalculateCRC(CRC_SEED, (uint8*)pBuf, PacketLength - CRCSIZE));

    SYS_PRINTF("\n\n");
}
#endif

/********************************************************************************************
Interchange_Terminate - Terminates the connection based on the passed in SourceID.
********************************************************************************************/
void Interchange_Terminate(char* pSourceID, uint16 Port)
{
    uint8*  pDataBuf = NULL;
    uint16  DataSize = 0;

    /* Faked Termination message */
    const uint8 WacpTerminate[] = 
    {   
        0x00,0x1E,0x01,0x00,0x00,
        0x00,0x00,0x07,0x00,0x00,
        0x00,0x00,0x00,0x61,0xAD
    };


    /* Only terminate if this is a WACP Port message type 
     * or special 0 type 
     */
    ConnArgs* pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceID);
    if (pConnArgs)
    {
        if (Port != pConnArgs->RNDZPort)
        {
            DataSize = sizeof(WacpTerminate);
            pDataBuf = SYS_MALLOC(DataSize);
            if (pDataBuf)
            {
                QPACKET* pData;

                /* Remove all existing messages */
                do
                {
                    /* Dequeue the new message */
                    pData = CWACPQueue_Dequeue(&MsgQueue);
                    if (pData)
                    {
                        SYS_FREE(pData->pDataBuf);
                        SYS_FREE(pData);
                    }                
                }while (pData != NULL);

                memcpy(pDataBuf, WacpTerminate, DataSize);

                /* Queue up the new message */
                if (CWACPQueue_Enqueue(&MsgQueue, pSourceID, pDataBuf, DataSize) == TRUE)
                {
                    /* Raise a semaphore to notify the receiving thread of the new message */
                    CWACPSemaphore_Raise(&DataSem);
                }
                else
                {
                    SYS_FREE(pDataBuf);
                }
            }
        }
    }
}

/*================================================================================
GeneratePacket: Wraps the passed in Buffer into a PACKET structure and    
adds header information based on parameters. Returns a newly malloced     
uint8 buffer filled with the new packet.
================================================================================*/
uint8* GeneratePacket(uint32* pOutSize, int16 Port, uint8* pBuf, uint32 Size)
{    
    uint8* OutBuffer = NULL;    
    
    /* Set the Size of the entire packet */    
    *pOutSize = PACKET_LITE_HEADER_SIZE + Size + CRCSIZE;    
    
    /* Allocate room for the serialize packet Stream freed in EvaluateSend */    
    OutBuffer = (uint8*)SYS_MALLOC(*pOutSize);        
    if (OutBuffer)    
    {        
        uint8* pTemp = OutBuffer;         
        
        /* Serialize the packet header struct */        
        /* Set the PREAMBLE */            
        *pTemp++ = CTL_W;        
        *pTemp++ = CTL_A;        
        *pTemp++ = CTL_L;        
        
        /* Write the Packet Length */        
        WriteInt32(&pTemp, *pOutSize);        
        
        /* Write the Port */        
        WriteInt16(&pTemp, Port);        
        
        /* Copy the data Stream */        
        memcpy((char*)pTemp, pBuf, Size);        
        pTemp += Size;                
        
        /* Write the CRC */        
        WriteInt16(&pTemp, CalculateCRC(CRC_SEED, OutBuffer, *pOutSize - CRCSIZE));   

    }/* if (OutBuffer) */    
    
    return(OutBuffer);
}
