/*-D************************************************************************* 
'*   This program source file is confidential and proprietary property of
'*    Welch Allyn, Inc. Unauthorized use or reproduction is prohibited.
'****************************************************************************
'****************************************************************************
'**
'****************************************************************************
'**  ////////////////////////////////////////////////////////////////////////
'**  '// WACP.c                                                            //
'**  '// Language       :   C                                              //
'**  '// Application    :   WACP                                           //
'**  '// Sources        :   MOMP Standard Library                          //
'**  '///////////////////////////////////////////////////////////////////////
'*-E************************************************************************/

#include "wacp.h"

/* Globals */
CLinkTable   LinkTable;

/* Locals */
static uint8*       pMsgBuffer;
static CConObjSvr   Connection;
static CTrapObjSvr  TrapServer;
static CMsgSvr      InMsgSvr;
static CMsgSvr      OutMsgSvr;
static CWACPMutex   SendMutex;


#if defined(TRACE_SERVER_ON)
CTraceSvr       TraceServer;
char WacpString[] = "SourceID = %s\nMessage Data:\n%A";
#endif

/*================================================================================
DataReadyEvent: Callback for messages received from the Interchange layer.
================================================================================*/
static void DataReadyEvent(char* pSourceID, uint8* pBuf, uint32 Size)
{
    uint16 Family = GetFamily(pBuf);
    uint8  Genus = GetGenus(pBuf);
    uint8  Species = GetSpecies(pBuf);

#if defined(TRACE_SERVER_ON_STATUS)
    CTraceSvr_tprintf(&TraceServer, TRACE_Type_RECEIVE,
                      "WACP", "DataReadyEvent", Family, Genus, Species, 
                      WacpString, pSourceID, pBuf, Size);
#endif

    /* Update the Connection Object server thread */
    CConObjSvr_StopKeepAlive(&Connection, pSourceID);

    if (Size)
    {
        switch(Family)
        {
            case FmTRAP:
            {
                CTrapObjSvr_RecvMsgHandler(&TrapServer, pSourceID, pBuf);
                break;
            }
            case FmCONNECTION:
            {
                CConObjSvr_RecvMsgHandler(&Connection, pSourceID, pBuf);
                break;
            }
            default:
            {
                uint8* pDataBuf = pBuf;
                uint32 Result = TRUE;

                if (Genus != GnDATA)
                {
                    /* Initialize the InMsgSvr structure */
                    CMsgSvr_Init(&InMsgSvr, Family, Genus, Species);

                    /* Serialize the data */
                    Result  = CMsgSvr_Deserialize(&InMsgSvr, pSourceID, pBuf);

                    /* Set the databuf pointer to the contents of the InMsgSvr object */
                    pDataBuf = InMsgSvr.Data;
                }

                if (Result)
                {
                    /* Route the message to the Wacp application */
                    Result = WACP_RecvMsg(pSourceID, Family, Genus, Species, pDataBuf);
                }
                else
                {
                    /* Send CRC Error message */
                    CTrapObjSvr_SendError(&TrapServer, pSourceID, MakeClassID(FmTRAP, GnERROR, CERR_SpERRWACPSTACK),
                                          CERR_CRC_MISMATCH, TRUE);
                }

                if (Genus != GnDATA)
                {
                    /* Free InMsgSvr.Data */
                    SYS_FREE(InMsgSvr.Data);
                    InMsgSvr.Data = NULL;
                }
                break;
            }
        };
    }
    else
    {
        /* Send CRC Error message */
        CTrapObjSvr_SendError(&TrapServer, pSourceID, MakeClassID(FmTRAP, GnERROR, CERR_SpERRWACPSTACK),
                              CERR_CRC_MISMATCH, TRUE);
    }
}

/********************************************************************************************
WACP_RecvMsg: Main interface method for receiving a WACP message.
********************************************************************************************/
boolean WACP_RecvMsg (char* pSourceID, uint16 Family, uint8 Genus, uint8 Species, 
                      uint8* pBuffer)
{
    boolean Result = FALSE;

    switch(Family)
    {
        case FmRENDEZVOUS:
        {
            Result = CRNDZApp_RxMsg(pSourceID, Genus, Species, pBuffer);
            break;
        }
        case FmCONNECTION:
        {
            ConnArgs  *pConnArgs = NULL;

            Result = RecvConnection(pSourceID, Species, pBuffer);
    
            pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceID);
            if (pConnArgs) 
            {
                if (pConnArgs->bRNDZStatus == FALSE)
                {
                    CRNDZApp_Terminate(pSourceID, FALSE);
                }
            }
            break;
        }
        default:
        {
            uint32 ClassID = MakeClassID(Family, Genus, Species);
            
            RecvMessage(pSourceID, ClassID, pBuffer);
            break;
        }
    };

    return(Result);
}

/********************************************************************************************
WACP_SendMsg: Main interface method for sending a WACP message.
********************************************************************************************/
boolean WACP_SendMsg (const char* pSourceID, uint16 Family, uint8 Genus, uint8 Species, 
                      const uint8* Buffer, uint32 Size, boolean bEncryptionOn)
{
    uint32 MsgSize = 0;
    boolean Result = TRUE;
    
    CWACPMutex_Lock(&SendMutex, INFINITE);

#if defined(TRACE_SERVER_ON_STATUS)
    CTraceSvr_tprintf(&TraceServer, TRACE_Type_SEND,
                      "WACP", "SendMsg", Family, Genus, Species, 
                      WacpString, pSourceID, Buffer, Size);
#endif

    switch(Genus)
    {
        case GnDATA:
        {
            /* Pass the data straight through */
            MsgSize = Size;
            pMsgBuffer = (uint8*)SYS_REALLOC(pMsgBuffer, MsgSize);
            if(pMsgBuffer != NULL)
            {
                memcpy(pMsgBuffer, Buffer, MsgSize);
            }
            break;
        }
        default:
        {
            /* Initialize the OutMsgSvr structure */
            CMsgSvr_Init(&OutMsgSvr, Family, Genus, Species);

            if(bEncryptionOn)
            {
                CMsgSvr_SetEncryption(&OutMsgSvr, bEncryptionOn);
            }

            if(CMsgSvr_SetData(&OutMsgSvr, pSourceID, Buffer, Size))
            {
                /* Get the full message size */
                MsgSize = CMsgSvr_SizeAll(&OutMsgSvr);

                pMsgBuffer = (uint8*)SYS_REALLOC(pMsgBuffer, MsgSize);
                if(pMsgBuffer != NULL)
                {
                    /* Serialize the data */
                    Result = CMsgSvr_Serialize(&OutMsgSvr, pSourceID, pMsgBuffer);
                }

                /* Free Data */
                SYS_FREE(OutMsgSvr.Data);
                OutMsgSvr.Data = NULL;
            }
        }
    }

    if (Result)
    {
        /* Send the message */
        Result = Interchange_SendPacket(pSourceID, pMsgBuffer, MsgSize);
    }

    /* Free pMsgBuffer */
    SYS_FREE(pMsgBuffer);
    pMsgBuffer = NULL;

    CWACPMutex_Release(&SendMutex);

    return(Result);
}

/********************************************************************************************
WACP_Init - WACP entry point for starting up the WACP stack. 
********************************************************************************************/
boolean WACP_Init(int32 ThreadPriority)
{
    /* Initialize the Linktable */
    CLinkTable_Init(&LinkTable);

    /* Initialize the Trap Object Server */
    CTrapObjSvr_Init(&TrapServer);

    /* Initialize the Connection Object Server */
    CConObjSvr_Init(&Connection, &LinkTable, ThreadPriority);

    /* Initialize the Interchange layer */
    Interchange_Init(ThreadPriority);
    Interchange_RegisterCallback(DataReadyEvent);

#if defined(TRACE_SERVER_ON)
    CTraceSvr_Init(&TraceServer);
    /* Open the Trace file here so we can start writing to it */
    CTraceSvr_OpenFile(&TraceServer, TRACE_WACP_LOG, "WACP", NULL);
    CTraceSvr_GenerateXSLFile(&TraceServer, NULL);
    CTraceSvr_EnableLogging(&TraceServer, TRUE, TRUE);
#endif

    /* Initialize the mutex for accessing the SendMsg function */
    CWACPMutex_Init(&SendMutex, NULL);

    return TRUE;
}

/********************************************************************************************
WACP_Destroy - WACP entry point for terminating the WACP stack. 
********************************************************************************************/
void WACP_Destroy(void)
{
    uint16 i;

    /* Get a list of all entries in linktable and shut down all of them individually */
    char* pSourceList[MAX_CONNECTIONS];
    ConnArgs *pConnArgs;

    CLinkTable_GetSourceList(&LinkTable, pSourceList);
    for (i=0; pSourceList[i] != NULL; i++)
    {
        /* Look for active connections only */
        pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceList[i]);
        if ((pConnArgs) && (pConnArgs->bConnected))
            WACP_Close(pSourceList[i]);
    }

    /* Stop the interchange layer */
    Interchange_Destroy();

    /* Stop the Connection object server thread */
    CConObjSvr_Destroy(&Connection);

    /* Clean up the Trap Object Server */
    CTrapObjSvr_Destroy(&TrapServer);


#if defined(TRACE_SERVER_ON)
    CTraceSvr_CloseFile(&TraceServer);
#endif

    /* Destroy the Linktable */
    CLinkTable_Destroy(&LinkTable);

    /* Destroy the SendMsg function access mutex */
    CWACPMutex_Destroy(&SendMutex);
}

/********************************************************************************************
WACP_SetConnOptions - WACP entry point for setting the options for the connection object server.
    This sets what type of keep alive status will be enabled, the keepalive interval, and the
    keepalive timeout value.

    typedef  enum CONNCFGtypSPECIES {
        CONN_CFG_SpKEEPALIVEON = 0,
        CONN_CFG_SpKEEPALIVEOFF,
        CONN_CFG_SpKEEPALIVECHECK,
        CONN_CFG_SpMAX
    }CONN_CFG_SPECIES;

********************************************************************************************/
boolean WACP_SetConnOptions(const char* pSourceID, uint16 Option, time_t Interval, time_t TimeOut)
{
    boolean Result = FALSE;

    if (CConObjSvr_SetTxOptions(&Connection, pSourceID, (uint8)Option))
    {
        if (CConObjSvr_SetKeepAliveInterval(&Connection, pSourceID, Interval))
        {
            Result = CConObjSvr_SetTimeoutInterval(&Connection, pSourceID, TimeOut);
        }
    }
    return (Result);
}

/********************************************************************************************
WACP_GetConnOptions - WACP entry point for getting the options for the connection object server. 
    This gets what type of keep alive status has be enabled, the keepalive interval, and the 
    keepalive timeout value.

    typedef  enum CONNCFGtypSPECIES {
        CONN_CFG_SpKEEPALIVEON = 0,
        CONN_CFG_SpKEEPALIVEOFF,
        CONN_CFG_SpKEEPALIVECHECK,
        CONN_CFG_SpMAX
    }CONN_CFG_SPECIES;

********************************************************************************************/
boolean WACP_GetConnOptions(const char* pSourceID, uint16 *pOption, time_t *pInterval, time_t *pTimeOut)
{
    boolean Result = FALSE;

    if ((pSourceID) && (pOption) && (pInterval) && (pTimeOut))
    {
        *pOption = CConObjSvr_GetTxOptions(&Connection, pSourceID);
        *pInterval = CConObjSvr_GetKeepAliveInterval(&Connection, pSourceID);
        *pTimeOut = CConObjSvr_GetTimeoutInterval(&Connection, pSourceID);
        Result = TRUE;
    }
    return (Result);
}

/********************************************************************************************
WACP_SendTrap - WACP entry point for sending an application ACK/NACK message
********************************************************************************************/
boolean WACP_SendTrap(const char* pSourceID, uint32 MsgClassID, uint32 ErrClassID, uint8 ErrorSpecies,
                      uint16 ErrorCode, const char* pErrMsg, uint16 ErrBufSize, const uint8* pData, 
                      uint16 DataSize)
{
    boolean Result =  FALSE;
    uint32 ObjSize;
    uint8* ObjBuffer = NULL;
    CErrEStd AppErrObj;

    CWACPMutex_Lock(&SendMutex, INFINITE);

    /* Initailize the error object */
    CErrEStd_Init(&AppErrObj);
    SetSpecies(&AppErrObj.ClassID, ErrorSpecies);

    /* Set the ErrClassID, ErrorCode and description */
    AppErrObj.m_ErrClassID = ErrClassID;          /* Message or object ClassID */
    AppErrObj.m_ErrorCode = ErrorCode;

    /* Copy the error description */
    if (pErrMsg)
    {
        if (ErrBufSize > sizeof(AppErrObj.m_Description))
            ErrBufSize = sizeof(AppErrObj.m_Description);

        memcpy(AppErrObj.m_Description, pErrMsg, ErrBufSize);
    }

    /* Copy the new data pointer */
    SetData(&AppErrObj.m_Data, pData, DataSize);
    AppErrObj.m_DataSize = DataSize;

    /* Get object size */
    ObjSize = CErrEStd_SizeAll(&AppErrObj);

    /* Allocate memory */
    if((ObjBuffer = (uint8*)SYS_REALLOC(ObjBuffer,ObjSize)) != NULL)
    {
        /* Serialize object */
        Result = CErrEStd_Serialize(&AppErrObj,ObjBuffer);
        if (Result)
        {
            Result = CTrapObjSvr_TxError(&TrapServer, pSourceID, MsgClassID, ErrorSpecies, 
                                  ObjBuffer, ObjSize);
        }
        SYS_FREE(ObjBuffer);
    }

    /* Clear up the memory allocated by SetData */
    SYS_FREE(AppErrObj.m_Data);
    AppErrObj.m_Data = NULL;

    CWACPMutex_Release(&SendMutex);

    return (Result);
}

/********************************************************************************************
WACP_SendMsgProcTrap - WACP entry point for sending a message processing NACK message
********************************************************************************************/
boolean WACP_SendMsgProcTrap(const char* pSourceID, uint32 ClassID, uint16 ErrorCode)
{
    boolean Result = FALSE;

    if (ErrorCode < CERR_PROC_MAX)
    {
        Result = WACP_SendTrap(pSourceID, ClassID, ClassID, CERR_SpERRMSGPROC, ErrorCode,
                      (char*)CErrEStd_TypProcErrorTbl[ErrorCode],
                      (uint16)strlen(CErrEStd_TypProcErrorTbl[ErrorCode]), 
                      NULL, 0);
    }

    return (Result);
}

/********************************************************************************************
WACP_GetLastError - WACP entry point for requesting the last error generated by the connected
    device.
********************************************************************************************/
boolean WACP_GetLastError(const char* pSourceID)
{
    boolean Result = SEND_REQUEST(pSourceID, FmTRAP, TRAP_CRST_SpGET_LASTERROR, NULL, 0, FALSE);
    return (Result);
}

/*================================================================================
WACP_Open: WACP entry point for opening all communications channels. 
    Returns FALSE if any of the connections are unsuccessful. 
    Returns TRUE if all connections are opened successfully. 

    This will overwrite any existing entries in the LinkTable so long as the
    connection is not already open.
================================================================================*/
boolean WACP_Open(ConnArgs** pConnArgs, uint8 NumConnections)
{
    boolean Result = TRUE;
    uint8 i;
    ConnArgs *pTempArgs;

    /* Check for existing connections that are open */
    for (i=0; i < NumConnections; i++)
    {
        pTempArgs = CLinkTable_GetEntry(&LinkTable, pConnArgs[i]->SourceID);
        if (pTempArgs)
        {
            if (pTempArgs->bConnected)
            {
                Result = FALSE;
                break;
            }
        }
    }

    /* Open the channels only if none of the new channels are open */
    if (Result)
    {
        for (i=0; i < NumConnections; i++)
        {
            /* Update Application Linktable. */
            Result = CLinkTable_AddEntry(&LinkTable, pConnArgs[i]);
            CConObjSvr_AddConnection(&Connection, pConnArgs[i]->SourceID);

            /* If RNDZ is disabled, all communications will go through the RNDZ Port */
            if (pConnArgs[i]->bRNDZDisabled == TRUE)
            {
                pConnArgs[i]->WACPPort = pConnArgs[i]->RNDZPort;
            }
        }

        if (Result)
        {
            /* Call Interchange to open the connections */
            Result = Interchange_Connect(pConnArgs, NumConnections);
        }
    }

    return (Result);
}

/********************************************************************************************
WACP_Connect - WACP entry point for starting up a connection.
********************************************************************************************/
boolean WACP_Connect(const char* pSourceID)
{
    ConnArgs    *pConnArgs = NULL;
    boolean     Result = FALSE;

    /* Look up the linktable for the SourceID. If a match is found, use
     * the information in the ConnArgs structure to start the RNDZ connection
     * sequence.
     */
    pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceID);
    if (pConnArgs)
    {
        /* Check for a valid timeout. A timeout of 0 indicates a host that does not want
         * to block.
         */
        if (pConnArgs->Timeout != 0)
        {
            /* Start the Client RNDZ authentication */
            Result = CRNDZApp_AppStart(pConnArgs->bSvrFlag, pConnArgs->Timeout, pConnArgs);
            if ((Result) && (!pConnArgs->bSvrFlag))
            {
                CConObjSvr_SetTxOptions(&Connection, pConnArgs->SourceID, DEFAULT_CONN_SETTING);
            }
        }
        else
            Result = FALSE;
    }
    return (Result);
}

/********************************************************************************************
WACP_Close - WACP entry point for closing a connection.
********************************************************************************************/
boolean WACP_Close(const char* pSourceID)
{
    boolean Result = FALSE;

    /* Send Shutdown command */
    CConObjSvr_Disconnect(&Connection, pSourceID);

    /* Disconnect the connection */
    Result = Interchange_Disconnect(pSourceID);

    return (Result);
}

/********************************************************************************************
WACP_GetGUIDObj - WACP entry point for retrieving the CRNDZDStd object for a connection.
********************************************************************************************/
CRNDZDStd* WACP_GetGUIDObj(const char* pSourceID)
{
    ConnArgs*   pConnArgs = NULL;
    CRNDZDStd*   pGUIDObj = NULL;

    /* Look up the linktable for the SourceID. */
    pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceID);
    if (pConnArgs)
    {
        pGUIDObj = pConnArgs->pGUIDStd;
    }
    return (pGUIDObj);
}

/********************************************************************************************
WACP_GetGUID - WACP entry point for retrieving the GUID for a connection.
********************************************************************************************/
uint8* WACP_GetGUID(const char* pSourceID)
{
    ConnArgs*   pConnArgs = NULL;
    uint8*      pGUID = NULL;


    /* Look up the linktable for the SourceID. */
    pConnArgs = CLinkTable_GetEntry(&LinkTable, pSourceID);
    if (pConnArgs)
    {
        pGUID = ((CRNDZDStd*)(pConnArgs->pGUIDStd))->m_DeviceGUID;
        
    }
    return (pGUID);
}