/*-D************************************************************************* 
'*   This program source file is confidential and proprietary property of
'*    Welch Allyn, Inc. Unauthorized use or reproduction is prohibited.
'*                         Copyright (c) 2005-2006 by
'*                              Welch Allyn Inc.
'*                            All rights reserved
'****************************************************************************
'****************************************************************************
'**
'****************************************************************************
'**  ////////////////////////////////////////////////////////////////////////
'**  '// CDMOIBStringTable.cpp                                             //
'**  '// Version        :   1.00                                           //
'**  '// Language       :   C++                                            //
'**  '// Application    :   WACP                                           //
'**  '///////////////////////////////////////////////////////////////////////
'*-E************************************************************************/

#include "CDMOIBStringTable.h"
#include "MiscUtil.h"

/*
===============================================================================
    Table Manager Implementation
===============================================================================
    Class Constructor/Destructor
===============================================================================
*/
CDMOIBTableManager::CDMOIBTableManager(TableEntry_t* pDefault) :
m_pHead     (pDefault),
m_pDefault  (pDefault),
m_pSelected (pDefault)
{
}
CDMOIBTableManager::~CDMOIBTableManager()
{
    // Iterate through the added dynamic tables
    while(m_pHead != m_pDefault)
    {
        m_pSelected = m_pHead;
        m_pHead = m_pHead->m_pNext;
        delete [] ((byte*)m_pSelected);
    }
}

/*
===============================================================================
    Method to add a generic table entry to the Table Manager
===============================================================================
*/
byte* CDMOIBTableManager::AddTable(byte* pTable)
{
    TableEntry_t* l_pTableEntry         = NULL;
    TableEntry_t* l_pTableEntryAnchor   = NULL;
    
    if(pTable != NULL)
    {        
        uint32 l_TableSize  = TableSize(pTable);
        uint32 l_TableID    = TableID(pTable);

        l_pTableEntry       = m_pHead;
        l_pTableEntryAnchor = m_pHead;

        while(l_pTableEntry != m_pDefault)
        {   
            // If we found an existing Entry with the same ID, try to
            // reuse it. We are only searching entries that are dynamic, so
            // it is safe to assume we can delete it if necessary.         
            if(TableID(l_pTableEntry->m_pTable) == l_TableID)
            {
                l_pTableEntryAnchor->m_pNext = l_pTableEntry->m_pNext;

                if(TableSize(l_pTableEntry->m_pTable) < l_TableSize)
                {
                    if(m_pHead == l_pTableEntry)
                        m_pHead = m_pHead->m_pNext;

                    if(m_pSelected == l_pTableEntry)
                        m_pSelected = NULL;

                    delete [] l_pTableEntry;
                    l_pTableEntry = m_pDefault;
                }
                break;
            }

            l_pTableEntryAnchor = l_pTableEntry;
            l_pTableEntry = l_pTableEntry->m_pNext;
        }        

        if(l_pTableEntry == m_pDefault)
            l_pTableEntry = (TableEntry_t*)new byte[sizeof(TableEntry_t) + l_TableSize];

        if(l_pTableEntry != NULL)
        {   
            if(m_pSelected == NULL)
                m_pSelected = l_pTableEntry;

            l_pTableEntry->m_pNext  = m_pHead;
            m_pHead                 = l_pTableEntry;
            l_pTableEntry->m_pTable = ((byte*)l_pTableEntry)
                                      + sizeof(TableEntry_t);

            memcpy(l_pTableEntry->m_pTable, pTable, l_TableSize);

            return l_pTableEntry->m_pTable;
        }
    }

    return NULL;
}

/*
===============================================================================
    Method to get a generic table entry from the Table Manager
===============================================================================
*/
byte* CDMOIBTableManager::GetTable(uint32 Identifier)
{
    TableEntry_t* l_pTableEntry = m_pHead;

    while(l_pTableEntry != NULL)
    {
        if(TableID(l_pTableEntry->m_pTable) == Identifier)
            return l_pTableEntry->m_pTable;

        l_pTableEntry = l_pTableEntry->m_pNext;
    }

    return NULL;
}
/*
===============================================================================
    Method to set a generic table entry from the Table Manager as the 
    Selected table.
===============================================================================
*/
byte* CDMOIBTableManager::SetTable(uint32 Identifier)
{
    byte* l_pTable = NULL;

    if(m_pSelected != NULL && TableID(m_pSelected->m_pTable) == Identifier)
    {
        l_pTable = m_pSelected->m_pTable;
    }
    else
    {
        // The requested Identifier doesn't match the currently selected table
        // or m_pSelected is NULL. In either case, start at the Head of our
        // TableManager and iterate through it to find a match
        TableEntry_t* l_pIter = m_pHead;

        while(l_pIter != NULL)
        {
            if(TableID(l_pIter->m_pTable) == Identifier)
            {
                // We found a match, so change our Selected entry and 
                // break out from the loop.
                m_pSelected = l_pIter;
                l_pTable = m_pSelected->m_pTable;
                break;
            }

            // No match, so keep looking...
            l_pIter = l_pIter->m_pNext;
        }
    }

    return l_pTable;
}

/*
===============================================================================
    Link Table Manager Implementation
===============================================================================
    Class Constructor/Destructor
===============================================================================
*/
CDMOIBLinkTableManager::CDMOIBLinkTableManager(TableEntry_t* pDefaultTable) :
CDMOIBTableManager(pDefaultTable)
{
}
uint32 CDMOIBLinkTableManager::TableID(byte* pTable)
{
    return LinkTableID(pTable);
}
uint32 CDMOIBLinkTableManager::TableSize(byte* pTable)
{
    return LinkTableSize(pTable);
}

/*
===============================================================================
    Language Table Manager Implementation
===============================================================================
    Class Constructor/Destructor
===============================================================================
*/
CDMOIBLanguageTableManager::CDMOIBLanguageTableManager(TableEntry_t* pDefaultTable) :
CDMOIBTableManager(pDefaultTable)
{
}
uint32 CDMOIBLanguageTableManager::TableID(byte* pTable)
{
    return LanguageTableID(pTable);
}
uint32 CDMOIBLanguageTableManager::TableSize(byte* pTable)
{
    return LanguageTableSize(pTable);
}

/*
===============================================================================
    CDMOIBStringTable class implementation
===============================================================================
    Class Constructor/Destructor
===============================================================================
*/
CDMOIBStringTable::CDMOIBStringTable(void* pDefaultStringTable,
                                     void* pDefaultLinkTable)
:
m_LanguageTables((TableEntry_t*)pDefaultStringTable),
m_LinkTables    ((TableEntry_t*)pDefaultLinkTable)
{
}
CDMOIBStringTable::~CDMOIBStringTable()
{
    // The destructors for the table managers will do all the work 
    // automatically, since they are declared statically.
}
/*
===============================================================================
    Method to get the language ID of the currently selected language table.
===============================================================================
*/
uint8 CDMOIBStringTable::GetLanguageID()
{
    uint8 l_RetVal = LANGUAGE_DEFAULT;
    
    if(m_LanguageTables.m_pSelected != NULL)
    {
        l_RetVal = (uint8)LanguageTableID(m_LanguageTables.m_pSelected->m_pTable);
    }

    return l_RetVal;
}
/*
===============================================================================
    Method to change the selected language table.
===============================================================================
*/
boolean CDMOIBStringTable::SetLanguageID(uint8 LanguageID)
{
    return (m_LanguageTables.SetTable(LanguageID) != NULL);
}

/*
===============================================================================
    Method to add a new link table dynamically
===============================================================================
*/
void* CDMOIBStringTable::AddLinkTable(void* pLinkTable)
{
    void*   l_pRetVal = NULL;
    uint32  l_Offset = LinkTableSize(pLinkTable) - sizeof(uint16);
    uint16  l_CRC = CalculateCRC((uint8*)pLinkTable, l_Offset);

    if(LINE_CHECK((uint8*)pLinkTable + l_Offset, &l_CRC, sizeof(uint16)))
    {
        l_pRetVal = m_LinkTables.AddTable((byte*)pLinkTable);
    }

    return l_pRetVal;
}
/*
===============================================================================
    Method to retrieve a reference to an existing link table managed by the
    String table.
===============================================================================
*/
void* CDMOIBStringTable::GetLinkTable(uint32 ClassID)
{
    return m_LinkTables.SetTable(ClassID);
}
/*
===============================================================================
    Method to add a new language table dynamically
===============================================================================
*/
void* CDMOIBStringTable::AddLanguageTable(void* pLanguageTable)
{   
    void*   l_pRetVal = NULL;
    uint32  l_Offset = LanguageTableSize(pLanguageTable) - sizeof(uint16);
    uint16  l_CRC = CalculateCRC((uint8*)pLanguageTable, l_Offset);

    if(LINE_CHECK((uint8*)pLanguageTable + l_Offset, &l_CRC, sizeof(uint16)))
    {
        l_pRetVal = m_LanguageTables.AddTable((byte*)pLanguageTable);
    }

    return l_pRetVal;
}
/*
===============================================================================
    Method to retrieve a reference to an existing language table managed by the
    String table.
===============================================================================
*/
void* CDMOIBStringTable::GetLanguageTable(uint8 LanguageID)
{
    return m_LanguageTables.GetTable(LanguageID);
}

/*
===============================================================================
    Method to retrieve a string from the specified language table, matching
    the specified symbolic link.
===============================================================================
*/
const ansichar* CDMOIBStringTable::GetString(uint16 SymbolicLink, 
                                             uint8 LanguageID)
{
    byte* l_pLanguageTable = NULL;

    if(LanguageID != LANGUAGE_DEFAULT)
        l_pLanguageTable = (byte*)GetLanguageTable(LanguageID);
    else if(m_LanguageTables.m_pSelected != NULL)
        l_pLanguageTable = m_LanguageTables.m_pSelected->m_pTable;

    if(l_pLanguageTable != NULL)
    {
        byte*   l_pLanguageTableIter = l_pLanguageTable;
        uint16  l_SymbolicLinkCount;
        uint32  l_SymbolicLinkOffset;

        // Skip the LanguageID at the head
        l_pLanguageTableIter++;  

        // Grab the SymbolicLink Count
        LINE_COPY(&l_SymbolicLinkCount, l_pLanguageTableIter, sizeof(uint16));
        l_pLanguageTableIter += sizeof(uint16);

        // Check to make sure the SymbolicLink requested is within our boundaries
        if(SymbolicLink < l_SymbolicLinkCount)
        {
            // Move forward into the lookup table using the SymbolicLink
            l_pLanguageTableIter += (SymbolicLink * sizeof(uint32));
            LINE_COPY(&l_SymbolicLinkOffset, l_pLanguageTableIter, sizeof(uint32));

            // Use the value at the SymbolicLink position to index into the
            // string block at the end of the table.
            return (const ansichar*)(l_pLanguageTable + l_SymbolicLinkOffset);
        }
    }

    return NULL;
}
/*
===============================================================================
    Method to retrieve the textual description for a class.
===============================================================================
*/
const ansichar* CDMOIBStringTable::GetClassName(uint32 ClassID, uint8 LanguageID)
{
    // The ClassName is actually the enumeration name for the Species
    return GetEnumerationString(ClassID, SPECIES_ENUMERATION_ID, LanguageID);
}
/*
===============================================================================
    Method to retrieve the textual description for a species.
===============================================================================
*/
const ansichar* CDMOIBStringTable::GetSpeciesString(uint32 ClassID, 
                                                    uint8 Species,
                                                    uint8 LanguageID)
{
    return GetEnumeratorString(ClassID, SPECIES_ENUMERATION_ID, Species, LanguageID);
}
/*
===============================================================================
    Method to retrieve the textual description for a member of a definition.
===============================================================================
*/
const ansichar* CDMOIBStringTable::GetMemberString(uint32 ClassID, 
                                                   uint16 MemberID,
                                                   uint8 LanguageID)
{
    return GetEnumerationString(ClassID, MemberID, LanguageID);
}
/*
===============================================================================
    Method to retrieve the textual description for an enumeration.
===============================================================================
*/
const ansichar* CDMOIBStringTable::GetEnumerationString(uint32 ClassID, 
                                                        uint16 Enumeration,
                                                        uint8 LanguageID)
{
    // Get a reference to the link table associated with the specified ClassID,
    // if one exists.
    byte* l_pLinkTableEntry = (byte*)GetLinkTable(ClassID);

    if(l_pLinkTableEntry != NULL)
    {
        byte*   l_pTableHead = l_pLinkTableEntry;
        uint16* l_pTableIter;            
        uint16  l_Value;

        // Move our iterator past some of the header information, since we are
        // not concerned about it for our lookup process.
        l_pTableIter = (uint16*)(l_pTableHead 
                                + sizeof(uint32)    // Skip ClassID
                                + sizeof(uint16));  // Skip Entry Size                                            

        // The next value in the header is the number of members available in
        // this lookup segment, so copy it out into l_Value.
        LINE_COPY(&l_Value, l_pTableIter, sizeof(uint16));
        l_pTableIter++;

        // Check to see if the requested Enumeration value is greater than our
        // member count. If it is, then we must be looking for an actual 
        // enumeration. It it is not, then we are looking for a regular member,
        // so continue.
        if(Enumeration > l_Value)
        {
            // We are looking for an enumeration, whose numerical value is 
            // represented by a reverse-order uint16 value - meaning that
            // enumerations start at UINT16_MAX, and work downward 
            // sequentially as more are added to a definition (0xFFFF, 0xFFFE,
            // 0xFFFD, etc.). To extract the actual offset value for the 
            // enumeration, we subtract it from UINT16_MAX, then add it to
            // the member count. 
            Enumeration = UINT16_MAX - Enumeration + l_Value;
            LINE_COPY(&l_Value, l_pTableIter, sizeof(uint16));
        }

        // l_Value is either member count or enumeration count at this point,
        // so make sure that our Enumeration is still less that it whether 
        // we're dealing with a member or enumeration.
        if(Enumeration < l_Value)
        {
            // Each entry in the lookup portion contains 2 uint16 values, so
            // we move ahead by the Enumeration value * 2. Also, we move ahead
            // 1 uint16 value to skip over the EnumerationCount in the header,
            // and 1 more uint16 value to offset into the pair of uint16 values
            // at the requested enumeration position. Altogether, it is 
            // Enumeration * 2 + 1 + 1 = Enumeration * 2 + 2 = the following...
            l_pTableIter += (Enumeration + 1) * 2;
            LINE_COPY(&l_Value, l_pTableIter, sizeof(uint16));
            return GetString(l_Value, LanguageID);
        }
    }

    return NULL;
}
/*
===============================================================================
    Method to get the text for an enumerator value, from a specified 
    enumeration set.
===============================================================================
*/
const ansichar* CDMOIBStringTable::GetEnumeratorString(uint32 ClassID, 
                                                       uint16 Enumeration, 
                                                       uint16 Enumerator, 
                                                       uint8 LanguageID)
{
    byte* l_pLinkTableEntry = (byte*)GetLinkTable(ClassID);

    if(l_pLinkTableEntry != NULL)
    {
        byte*   l_pTableHead = l_pLinkTableEntry;
        uint16* l_pTableIter;
        uint16* l_pTableEnd;
        uint16  l_Value;

        // Move our iterator past some of the header information, since we are
        // not concerned about it for our lookup process.
        l_pTableIter = (uint16*)(l_pTableHead 
                                + sizeof(uint32)    // Skip ClassID
                                + sizeof(uint16));  // Skip Entry Size
                                            
        // The next value in the header is the number of members available in
        // this lookup segment, so copy it out into l_Value.
        LINE_COPY(&l_Value, l_pTableIter, sizeof(uint16));
        l_pTableIter++;

        // Check to see if the requested Enumeration value is greater than our
        // member count. If it is, then we must be looking for an actual 
        // enumeration. It it is not, then we are looking for a regular member,
        // so continue.
        if(Enumeration > l_Value)
        {
            // We are looking for an enumeration, whose numerical value is 
            // represented by a reverse-order uint16 value - meaning that
            // enumerations start at UINT16_MAX, and work downward 
            // sequentially as more are added to a definition (0xFFFF, 0xFFFE,
            // 0xFFFD, etc.). To extract the actual offset value for the 
            // enumeration, we subtract it from UINT16_MAX, then add it to
            // the member count. 
            Enumeration = UINT16_MAX - Enumeration + l_Value;
            LINE_COPY(&l_Value, l_pTableIter, sizeof(uint16));
        }

        // l_Value is either member count or enumeration count at this point,
        // so make sure that our Enumeration is still less that it whether 
        // we're dealing with a member or enumeration.
        if(Enumeration < l_Value)
        {
            // Each entry in the lookup portion contains 2 uint16 values, so
            // we move ahead by the Enumeration value * 2. Also, we move ahead
            // 1 uint16 value to skip over the EnumerationCount in the header,
            // and that will put us at the position of the stored offset value
            // to the related enumerator set.
            l_pTableIter += (Enumeration * 2) + 1;
            LINE_COPY(&l_Value, l_pTableIter, sizeof(uint16));

            if(l_Value == UINT16_MAX)
                return NULL;

            l_pTableIter = (uint16*)(l_pTableHead + l_Value);

            // The first value at the start of the enumerator set is the offset
            // to the end of the set. So grab it and set our TableEnd to the
            // appropriate position for our bounds.
            LINE_COPY(&l_Value, l_pTableIter, sizeof(uint16));
            l_pTableEnd = (uint16*)(l_pTableHead + l_Value);
            l_pTableIter++;

            // Iterate through the enumerators...
            while(l_pTableIter < l_pTableEnd)
            {
                // Check to see if the current enumerator is the one we are
                // looking for.
                if(LINE_CHECK(l_pTableIter++, &Enumerator, sizeof(uint16)) == TRUE)
                {
                    LINE_COPY(&l_Value, l_pTableIter, sizeof(uint16));
                    return GetString(l_Value, LanguageID);
                }
                l_pTableIter++;
            }
        }
    }

    return NULL;
}


/*
===============================================================================
    Global Helper functions
===============================================================================
    Function to get the size of a LinkTable
===============================================================================
*/
uint32 LinkTableSize(void* pLinkTable)
{
    uint32 l_RetValue = 0;
    
    if(pLinkTable != NULL)
    {
        LINE_COPY(&l_RetValue, (byte*)pLinkTable + sizeof(uint32), sizeof(uint16));
    }
    return l_RetValue;
}

/*
===============================================================================
    Function to get the ClassID of a LinkTable
===============================================================================
*/
uint32 LinkTableID(void* pLinkTable)
{
    uint32 l_RetValue = (uint32)ANY_CLASS::CLASSID;
    
    if(pLinkTable != NULL)
    {
        LINE_COPY(&l_RetValue, pLinkTable, sizeof(uint32));
    }
    return l_RetValue;
}

/*
===============================================================================
    Function to get the size of a LanguageTable
===============================================================================
*/
uint32 LanguageTableSize(void* pLanguageTable)
{       
    uint32 l_RetValue = 0;
    
    if(pLanguageTable != NULL)
    {
        // Get the Link Count value first
        LINE_COPY(&l_RetValue, ((byte*)pLanguageTable)
                              + sizeof(uint8),       // Skip Language ID Byte
                                sizeof(uint16));
                                
        // Then get the end of the string table
        LINE_COPY(&l_RetValue, ((byte*)pLanguageTable)
                              + sizeof(uint8)               // Skip Language ID Byte
                              + sizeof(uint16)              // Skip Link Count
                              + l_RetValue * sizeof(uint32) // Skip Past all Links
                               ,sizeof(uint32));                                        
    }
    return l_RetValue;
}

/*
===============================================================================
    Function to get the LanguageID of a LanguageTable
===============================================================================
*/
uint32 LanguageTableID(void* pLanguageTable)
{
    uint8 l_RetValue = LANGUAGE_UNKNOWN;
    
    if(pLanguageTable != NULL)
    {
        l_RetValue = *(byte*)pLanguageTable;
    }
    return l_RetValue;
}


