/*-D************************************************************************* 
'*   This program source file is confidential and proprietary property of
'*    Welch Allyn, Inc. Unauthorized use or reproduction is prohibited.
'****************************************************************************
'****************************************************************************
'**
'****************************************************************************
'**  ////////////////////////////////////////////////////////////////////////
'**  '// Mem.cpp                                                           //
'**  '// Language       :   C++                                            ///
'**  '// Application    :   WACP                                           //
'**  '// Sources        :   MOIB Standard Library                          //
'**  '///////////////////////////////////////////////////////////////////////
'****************************************************************************
'**
'**
'****************************************************************************
'****************************************************************************
'** 
'*-E************************************************************************/

#include "Mem.h"
#include <string.h>

/***************************************************************************
Constructor. Passes the parameters to the QManager constructor.
***************************************************************************/
uint8QManager::uint8QManager(uint8* pStart, uint32 Elements, boolean bFull)
{
    pBufStart = pStart;
    pBufEnd = pStart + Elements * sizeof(uint8);
    NumElements = Elements;
    bFull ? ElementsUsed = Elements : ElementsUsed = 0;
    pHead = pStart;
    pTail = pStart;
} 

/***************************************************************************
Used to insert an object at the head of the queue. Returns false if there
is no space left. Otherwise, the object is copied to the head of the queue
and true is returned.
***************************************************************************/
boolean uint8QManager::enqueue(const uint8 Object)
{
    if (ElementsUsed != NumElements)
    {
        *pHead = Object;
        ++ElementsUsed;
        ++pHead;
        if (pHead >= pBufEnd) 
        { 
            pHead = pBufStart; 
        }
        return true;
    }
    return false;
}

/***************************************************************************
Used to get an object from the tail of the queue. Returns false if there
are no elements in the queue. Otherwise, the object at the tail is copied
into the object pointed to by the input parameter and true is returned.
***************************************************************************/
boolean uint8QManager::dequeue(uint8* pObject)
{
    if (ElementsUsed)
    {
        *pObject = *pTail;
        --ElementsUsed;
        ++pTail;
        if (pTail >= pBufEnd) 
        { 
            pTail = pBufStart; 
        }
        return true;
    }
    return false;
}

boolean uint8QManager::enqueue_array(uint8* pArray, uint32 Elements)
{
    uint32 NumFree = NumElements - ElementsUsed;

    if (NumFree >= Elements)
    {
        if (pHead + Elements < pBufEnd)
        {
            memcpy(pHead, pArray, Elements);
            pHead +=Elements;
        }
        else
        {
            uint32 ElementHead = (uint32)(pBufEnd - pHead);
            uint32 ElementTail = Elements - ElementHead;

            memcpy(pHead, pArray, ElementHead);
            memcpy(pBufStart, pArray + ElementHead, ElementTail);
            pHead = pBufStart + ElementTail;
        }
        
        ElementsUsed += Elements;
        return true;
    }
    return false;
}

boolean uint8QManager::dequeue_array(uint8* pArray, uint32 Elements)
{
    if (ElementsUsed >= Elements)
    {
        if (pTail + Elements < pBufEnd)
        {
            memcpy(pArray, pTail, Elements);
            pTail +=Elements;
        }
        else
        {
            uint32 ElementHead = (uint32)(pBufEnd - pTail);
            uint32 ElementTail = Elements - ElementHead;

            memcpy(pArray, pTail, ElementHead);
            memcpy(pArray + ElementHead, pBufStart, ElementTail);
            pTail = pBufStart + ElementTail;
        }
        
        ElementsUsed -= Elements;
        return true;
    }
    return false;
}
/***************************************************************************
Method to increment the tail pointer by elements. 
***************************************************************************/
boolean uint8QManager::dequeue_array(uint32 Elements)
{
    if (get_num_used() >= Elements)
    {
        pTail += Elements;
        ElementsUsed -= Elements;

        return true;
    }
    return false;
}
/***************************************************************************
Returns the number of free slots remaining in the queue.
***************************************************************************/
uint32 uint8QManager::get_num_free(void)
{
    return NumElements - ElementsUsed;
}

/***************************************************************************
Returns the number of elements used in the queue.
***************************************************************************/
uint32 uint8QManager::get_num_used(void)
{
    return ElementsUsed;
}

/***************************************************************************
Returns the number of elements used in the queue.
***************************************************************************/
uint32 uint8QManager::get_num_elements(void)
{
    return NumElements;
}

/***************************************************************************
Method to place an unsigned 8 bit integer into the queue. Returns true if
the queue has enough room to hold the value. If the queue does not have
enough room, returns false and places nothing into the queue.
***************************************************************************/
boolean uint8QManager::put (uint8 Val)
{
    return enqueue(Val);
}

/***************************************************************************
Method to get an unsigned 8 bit integer from the queue. Returns true if
there are any elements in the queue to get.
***************************************************************************/
boolean uint8QManager::get(uint8* pVal)
{
    return dequeue(pVal);
}

/***************************************************************************
Method to place a signed 8 bit integer into the queue. Returns true if the
queue has enough room to hold the value. If the queue does not have enough
room, returns false and places nothing into the queue.
***************************************************************************/
boolean uint8QManager::put(int8 Val)
{
    return enqueue((uint8)Val);
}

/***************************************************************************
Method to get a signed 8 bit integer into the queue. Returns true if there
are any elements in the queue.
***************************************************************************/
boolean uint8QManager::get(int8* pVal)
{
    return dequeue((uint8*)pVal);
}
/**************************************************************************
widechar support.  Just passes through to uint16 handlers
***************************************************************************/
#if defined(WIDE_CHAR_SUPPORT)
boolean uint8QManager::put(widechar Val)
{
    return put((uint16)Val);
}

boolean uint8QManager::get(widechar* pVal)
{
    return get((uint16*)pVal);
}
#endif

/***************************************************************************
Method to place an unsigned 16 bit integer into the queue. Returns true if
the queue has enough room to hold the Val. If the queue does not have
enough room, returns false and places nothing in the queue.
***************************************************************************/
boolean uint8QManager::put(uint16 Val)
{
    if (get_num_free() >= 2)
    {
#if (MEDIA_ENDIANNES == BIG_ENDIAN)
        enqueue((uint8)(Val>>8));
        enqueue((uint8)Val);
#else
        enqueue((uint8)Val);
        enqueue((uint8)(Val>>8));
#endif
        return true;
    }
    return false;
}

/***************************************************************************
Method to get an unsigned 16 bit integer from the queue. If there are not
2 bytes in the queue, then nothing is gotten and false is returned. Else,
the 2 bytes are dequeued and formed into a 16 bit value and true is returned.
***************************************************************************/
boolean uint8QManager::get(uint16* pVal)
{
    uint8 Temp;
    if (get_num_used() >= 2)
    {
#if (MEDIA_ENDIANNES == BIG_ENDIAN)
        dequeue(&Temp);
        *pVal = ((uint16)Temp) << 8;
        dequeue(&Temp);
        *pVal |= Temp;
#else
        dequeue(&Temp);
        *pVal = Temp;
        dequeue(&Temp);
        *pVal |= ((uint16)Temp) << 8;
#endif
        return true;
    }
    return false;
}

/***************************************************************************
Method to place a signed 16bit integer into the queue. See return values
for put (uint16).
***************************************************************************/
boolean uint8QManager::put(int16 Val)
{
    return put((uint16)Val);
}

/***************************************************************************
Method to get a signed 16bit integer from the queue. See return values for
get (uint16).
***************************************************************************/
boolean uint8QManager::get(int16* pVal)
{
    return get((uint16*)pVal);
}

/***************************************************************************
Method to place an unsigned 32bit integer into the queue. Returns true if
the queue has enough space to hold the Val. If the queue does not have
enough room, returns false and places nothing into the queue.
***************************************************************************/
boolean uint8QManager::put(uint32 Val)
{
    if (get_num_free() >= 4 )
    {
#if (MEDIA_ENDIANNES == BIG_ENDIAN)
        enqueue((uint8)(Val>>24));
        enqueue((uint8)(Val>>16));
        enqueue((uint8)(Val>>8));
        enqueue((uint8)Val);
#else
        enqueue((uint8)Val);
        enqueue((uint8)(Val>>8));
        enqueue((uint8)(Val>>16));
        enqueue((uint8)(Val>>24));
#endif
        return true;
    }
    return false;
}

/***************************************************************************
Method to get an unsigned 32 bit integer from the queue. If there are not
enough bytes in the queue, no bytes are dequeued and false is returned. If
there are 4 bytes, the bytes are gotten and formed into a uint32 for return.
true is also returned.
***************************************************************************/
boolean uint8QManager::get(uint32* pVal)
{
    uint8 temp;
    if (get_num_used() >= 4 )
    {
#if (MEDIA_ENDIANNES == BIG_ENDIAN)
        dequeue(&temp);
        *pVal  = (uint32)temp << 24;
        dequeue(&temp);
        *pVal |= (uint32)temp << 16;
        dequeue(&temp);
        *pVal |= (uint32)temp << 8;
        dequeue(&temp);
        *pVal |= (uint32)temp;
#else
        dequeue(&temp);
        *pVal  = (uint32)temp;
        dequeue(&temp);
        *pVal |= ((uint32)temp) << 8;
        dequeue(&temp);
        *pVal |= ((uint32)temp) << 16;
        dequeue(&temp);
        *pVal |= ((uint32)temp) << 24;
#endif
        return true;
    }
    return false;
}

/***************************************************************************
Method to place a signed 32bit integer into the queue. See return values
for put (uint32).
***************************************************************************/
boolean uint8QManager::put(int32 Val)
{
    return put((uint32)Val);
}

/***************************************************************************
Method to get a signed 32 bit integer from the queue. See return values
for get (uint32).
***************************************************************************/
boolean uint8QManager::get(int32* pVal)
{
    return get((uint32 *)pVal);
}

/* -------------------------------------------------------------------- */
/* -------------------------------------------------------------------- */
#define SIZEOF_FLOAT   (sizeof(float))
#if (TARGET_ENDIANNESS == MEDIA_ENDIANNES)
    #define F32_START_INDEX (0)
    #define F32_INC (1)
#else
    #define F32_START_INDEX (SIZEOF_FLOAT-1)
    #define F32_INC (-1)
#endif

/***************************************************************************
Method to put a 4 byte IEEE format floating point value into the queue. 
***************************************************************************/
boolean uint8QManager::put(float Val)
{
    union {
        float fval;
        uint8 uval[SIZEOF_FLOAT];
    } ftou;
    uint8*  pByte;
    int8    Inc;
    int8    Cnt;

    if (get_num_free() >= SIZEOF_FLOAT)
    {
        ftou.fval = Val;
        pByte = &ftou.uval[F32_START_INDEX];
        Inc = F32_INC;
        for (Cnt = 0; Cnt < SIZEOF_FLOAT; Cnt++)
        {
            enqueue((uint8)*pByte);
            pByte += Inc;
        }
        return TRUE;
    }
    return FALSE;
}

/***************************************************************************
Method to get a 4 byte IEEE format floating point value from the queue. 
***************************************************************************/
boolean uint8QManager::get(float* pVal)
{
    union {
        float fval;
        uint8 uval[SIZEOF_FLOAT];
    } ftou;
    uint8*  pByte;
    int8    Inc;
    int8    Cnt;

    pByte = &ftou.uval[F32_START_INDEX];
    Inc = F32_INC;

    if (get_num_used() >= SIZEOF_FLOAT)
    {
        for (Cnt = 0; Cnt < SIZEOF_FLOAT; Cnt++)
        {
            dequeue(pByte);
            pByte += Inc;
        }

        *pVal = ftou.fval;
        return TRUE;
    }
    return FALSE;
}

/* -------------------------------------------------------------------- */
/* -------------------------------------------------------------------- */
#define SIZEOF_DOUBLE  (sizeof(double))
#if (TARGET_ENDIANNESS == MEDIA_ENDIANNES)
    #define F64_START_INDEX (0)
    #define F64_INC (1)
#else
    #define F64_START_INDEX (SIZEOF_DOUBLE-1)
    #define F64_INC (-1)
#endif

/***************************************************************************
Method to put a 8 byte IEEE format double value into the queue. 
***************************************************************************/
boolean uint8QManager::put(double Val)
{
    union {
        double dval;
        uint8 uval[SIZEOF_DOUBLE];
    } ftou;
    uint8*  pByte;
    int8    Inc;
    int8    Cnt;

    ftou.dval = Val;
    pByte = &ftou.uval[F64_START_INDEX];
    Inc = F64_INC;

    if (get_num_free() >= SIZEOF_DOUBLE)
    {
        for (Cnt = 0; Cnt < SIZEOF_DOUBLE; Cnt++)
        {
            enqueue((uint8)*pByte);
            pByte += Inc;
        }
        return TRUE;
    }
    return FALSE;
}

/***************************************************************************
Method to get a 8 byte IEEE format double value from the queue. 
***************************************************************************/
boolean uint8QManager::get(double* pVal)
{
    union {
        double dval;
        uint8 uval[SIZEOF_DOUBLE];
    } ftou;
    uint8*  pByte;
    int8    Inc;
    int8    Cnt;

    pByte = &ftou.uval[F64_START_INDEX];
    Inc = F64_INC;

    if (get_num_used() >= SIZEOF_DOUBLE)
    {
        for (Cnt = 0; Cnt < SIZEOF_DOUBLE; Cnt++)
        {
            dequeue(pByte);
            pByte += Inc;
        }
        *pVal = ftou.dval;
        return TRUE;
    }
    return FALSE;
}

void *uint8QManager::GetHeadPtr()
{
    return pHead;
}

void *uint8QManager::GetTailPtr()
{
    return pTail;
}

/*================================================================================
ReadInt16 Endian-proof function to read a 16 bit integer from a buffer.
================================================================================*/
uint16 ReadInt16(uint8* pBuf)
{
#if (MEDIA_ENDIANNES == BIG_ENDIAN)
    return ((((uint16)*(pBuf)) << 8) | ((uint16)*(pBuf+1))); 
#else
    return ((((uint16)*(pBuf+1)) << 8) | ((uint16)*(pBuf)));
#endif
}

/*================================================================================
ReadInt32 Endian-proof function to read an 32 bit integer from a buffer.
================================================================================*/
uint32 ReadInt32(uint8* pBuf)
{
#if (MEDIA_ENDIANNES == BIG_ENDIAN)
    return ((((uint32)*(pBuf))   << 24) | 
            (((uint32)*(pBuf+1)) << 16) |
            (((uint32)*(pBuf+2)) <<  8) |
              (uint32)*(pBuf+3)); 
#else
    return ((((uint32)*(pBuf+3)) << 24) | 
            (((uint32)*(pBuf+2)) << 16) |
            (((uint32)*(pBuf+1)) <<  8) |
              (uint32)*(pBuf)); 
#endif
}

/*================================================================================
WriteInt16 Endian-proof function to write a 16 bit integer into a buffer.
================================================================================*/
void WriteInt16(uint8* pBuf, uint16 Val)
{
#if (MEDIA_ENDIANNES == BIG_ENDIAN)
    *(pBuf) = (uint8)(Val >> 8);
    *(pBuf+1) = (uint8)(Val);
#else
    *(pBuf+1) = (uint8)(Val >> 8);
    *(pBuf) = (uint8)(Val);
#endif
}

/*================================================================================
WriteInt32 Endian-proof function to write a 32 bit integer into a buffer.
================================================================================*/
void WriteInt32(uint8* pBuf, uint32 Val)
{
#if (MEDIA_ENDIANNES == BIG_ENDIAN)
    *(pBuf)   = (uint8)(Val >> 24);
    *(pBuf+1) = (uint8)(Val >> 16);
    *(pBuf+2) = (uint8)(Val >> 8);
    *(pBuf+3) = (uint8)(Val);
#else
    *(pBuf+3) = (uint8)(Val >> 24);
    *(pBuf+2) = (uint8)(Val >> 16);
    *(pBuf+1) = (uint8)(Val >> 8);
    *(pBuf)   = (uint8)(Val);
#endif
}


/* End of File */





