/*-D************************************************************************* 
'*   This program source file is confidential and proprietary property of
'*    Welch Allyn, Inc. Unauthorized use or reproduction is prohibited.
'*                         Copyright (c) 2005-2006 by
'*                              Welch Allyn Inc.
'*                            All rights reserved
'****************************************************************************
'****************************************************************************
'**
'****************************************************************************
'**  ////////////////////////////////////////////////////////////////////////
'**  '// CDMOIBDictionary.cpp                                              //
'**  '// Version        :   1.00                                           //
'**  '// Language       :   C++                                            //
'**  '// Application    :   WACP                                           //
'**  '///////////////////////////////////////////////////////////////////////
'*-E************************************************************************/

#include "CDMOIBDictionary.h"
#include "MiscUtil.h"

/*
===============================================================================
    CDMOIBDictionary class implementation
===============================================================================
    Class Constructor/Destructor
===============================================================================
*/
CDMOIBDictionary::CDMOIBDictionary(void* pDefaultDictionary)
{
    m_pEntryDefault = (TableEntry_t*)pDefaultDictionary;
    m_pEntryHead    = m_pEntryDefault;
    m_SearchMode    = ALL_DEFINITIONS;    
}
CDMOIBDictionary::~CDMOIBDictionary()
{
    TableEntry_t* l_pEntry = m_pEntryHead;

    // Iterate through the added dynamic definitions
    while(l_pEntry != m_pEntryDefault)
    {
        m_pEntryHead = l_pEntry->m_pNext;
        delete [] ((byte*)l_pEntry);
        l_pEntry = m_pEntryHead;        
    }
}

/*
===============================================================================
    Method to get the current search mode of the dictionary (All defs, dynamic
    only, or static only).
===============================================================================
*/
SearchMode_t CDMOIBDictionary::GetSearchMode()
{
    return m_SearchMode;
}
/*
===============================================================================
    Method to set the current search mode of the dictionary.
===============================================================================
*/
void CDMOIBDictionary::SetSearchMode(SearchMode_t SearchMode)
{
    if(SearchMode < UKNOWN_SEARCH_MODE)
        m_SearchMode = SearchMode;
}

/*
===============================================================================
    Method to get the number of definitions currently managed by this 
    dictionary. This is dependent on the search mode setting, and will only 
    include definitions that match the search mode in the reported count.
===============================================================================
*/
uint16 CDMOIBDictionary::GetDefinitionCount()
{
    uint16          l_RetVal = 0;
    TableEntry_t*   l_pEntry = m_pEntryHead;
    TableEntry_t*   l_pEntryEnd = NULL;

    switch(m_SearchMode)
    {
        case STATIC_DEFINITIONS_ONLY:
        {
            l_pEntry = m_pEntryDefault;
            break;
        }
        case DYNAMIC_DEFINITIONS_ONLY:
        {
            l_pEntryEnd = m_pEntryDefault; 
            break;
        }
    }

    while(l_pEntry != l_pEntryEnd)
    {
        l_pEntry = l_pEntry->m_pNext;
        l_RetVal++;
    }

    return l_RetVal;
}
/*
===============================================================================
    Method to find a DMOIB def based on it's Index. This is dependent on the 
    search mode setting, and will only include definitions that match the 
    search mode when indexing into the list of definitions.
===============================================================================
*/
void* CDMOIBDictionary::GetDefinitionAt(uint16 Index)
{
    TableEntry_t*  l_pEntry = m_pEntryHead;
    TableEntry_t*  l_pEntryEnd = NULL;
        
    switch(m_SearchMode)
    {
        case STATIC_DEFINITIONS_ONLY:
        {
            l_pEntry = m_pEntryDefault;
            break;
        }
        case DYNAMIC_DEFINITIONS_ONLY:
        {
            l_pEntryEnd = m_pEntryDefault; 
            break;
        }
    }

    while(l_pEntry != l_pEntryEnd && Index-- > 0)
    {
        l_pEntry = l_pEntry->m_pNext;
    }
    
    return (l_pEntry == l_pEntryEnd ? NULL : l_pEntry->m_pTable);
}

/*
===============================================================================
    Methods to retrieve a definition based on it's ClassID/F|G|S combo. This
    performs a linear search of the managed definitions to find a definition
    with a matching ID.
===============================================================================
*/
void* CDMOIBDictionary::GetDefinition(uint16 Family, uint8 Genus, uint8 Species)
{ 
    TableEntry_t* l_pEntry = FindEntry(MAKE_CLASSID(Family, Genus, Species), m_SearchMode);
    return (l_pEntry != NULL ? l_pEntry->m_pTable : NULL);
}
void* CDMOIBDictionary::GetDefinition(uint32 ClassID)
{
    TableEntry_t* l_pEntry = FindEntry(ClassID, m_SearchMode);
    return (l_pEntry != NULL ? l_pEntry->m_pTable : NULL);
}
/*
===============================================================================
    Method to add a new definition to the managed list.
===============================================================================
*/
void* CDMOIBDictionary::AddDefinition(void* pDefinition)
{   
    if(pDefinition != NULL)
    {
        uint32  l_Offset = DefinitionFullSize(pDefinition) - sizeof(uint16);
        uint16  l_CRC = CalculateCRC((uint8*)pDefinition, l_Offset);

        if(LINE_CHECK((uint8*)pDefinition + l_Offset, &l_CRC, sizeof(uint16)))
        {
            uint32          l_ClassID = DefinitionClassID(pDefinition);
            uint16          l_Version = DefinitionVersion(pDefinition);
            uint16          l_DefSize = DefinitionFullSize(pDefinition);
            TableEntry_t*   l_pEntry = FindEntry(l_ClassID, ALL_DEFINITIONS);

            // Check to see if we have a definition available in the added defs.
            if(l_pEntry == NULL || l_Version > DefinitionVersion(l_pEntry->m_pTable))
            {
                // Create a new buffer to hold our Entry.
                l_pEntry = (TableEntry_t*)new byte[sizeof(TableEntry_t)+l_DefSize];

                if(l_pEntry != NULL)
                {
                    l_pEntry->m_pTable = (byte*)(l_pEntry + 1);                
                    memcpy(l_pEntry->m_pTable, pDefinition, l_DefSize);

                    l_pEntry->m_pNext = m_pEntryHead;
                    m_pEntryHead = l_pEntry;

                    return l_pEntry->m_pTable;
                }
            }
        }
    }

    return NULL;
}

/*
===============================================================================
    Method to add multiple definitions in a packed buffer
===============================================================================
*/
uint16 CDMOIBDictionary::AddDefinitions(void* pDefinitions, uint32 Size)
{
    uint16  l_Result = NO_DEFINITIONS_ADDED; // = 0
    uint32  l_SizeIter = 0;
    
    if(pDefinitions != NULL && Size > 0)
    {
        while(l_SizeIter < Size)
        {
            void*   l_pDefinition = (((byte*)pDefinitions) + l_SizeIter);
            uint32  l_Offset = DefinitionFullSize(l_pDefinition) - sizeof(uint16);
            uint16  l_CRC = CalculateCRC((uint8*)l_pDefinition, l_Offset);

            // Do an initial CRC check to see if the current definition is
            // valid.
            if(LINE_CHECK((uint8*)l_pDefinition + l_Offset, &l_CRC, sizeof(uint16)))
            {
                // If the CRC check passed, then the definitions is valid, so
                // call the AddDefinition() method to handle it. This method
                // will return NULL for 2 cases: 1) If the definition CRC
                // fails, or 2) The definition already exists in the 
                // dictionary. Since we already passed the CRC check above,
                // the only way this can return NULL is if the definition
                // already exists. If it already exists, it is not a failure
                // case for this method, but we don't want to count it towards
                // the total definitions added for this method call. So only
                // increment l_Result if it returns a non-NULL value.
                if(AddDefinition(l_pDefinition) != NULL)
                {
                    l_Result++;
                }
                
                // Move on to the next definition, regardless of whether the
                // definition was added or not.
                l_SizeIter += DefinitionFullSize(((byte*)pDefinitions) + l_SizeIter);                
            }
            else
            {
                // Otherwise the CRC has failed, so break out of the loop.
                break;                      
            }
        }

        if(l_SizeIter == Size)
        {
            l_Result = ALL_DEFINITIONS_ADDED;
        }
    }

    return l_Result;
}

/*
===============================================================================
    Helper method to retrieve a definition's entry based on the provided 
    ClassID. 

    [private]
===============================================================================
*/
TableEntry_t* CDMOIBDictionary::FindEntry(uint32 ClassID, SearchMode_t SearchMode)
{
    TableEntry_t*  l_pEntry = m_pEntryHead;
    TableEntry_t*  l_pEntryEnd = NULL;
        
    switch(SearchMode)
    {
        case STATIC_DEFINITIONS_ONLY:
        {
            l_pEntry = m_pEntryDefault;
            break;
        }
        case DYNAMIC_DEFINITIONS_ONLY:
        {
            l_pEntryEnd = m_pEntryDefault; 
            break;
        }
    }

    while(l_pEntry != l_pEntryEnd)
    {
        // We use LINE_CHECKN to determine how many bytes of the ClassIDs 
        // match, starting at the Family, working towards the Species. 
        // LINE_CHECKN will return how many bytes were left over that 
        // didn't match between the 2 ClassIDs.
        switch(LINE_CHECKN(((byte*)l_pEntry->m_pTable) + 1, &ClassID, sizeof(uint32)))
        {
            case 1:
                // If only 1 byte didn't match then that implies the 
                // Family and Genus DID match, only the Species didn't. So
                // check to see if there are any other Species available in
                // this definition that match the ClassID's Species.
                if(IsSpeciesValid(l_pEntry->m_pTable, GET_SPECIES(ClassID)) == TRUE)
            case 0:
                // If 0 bytes DIDN'T match, then that means the whole ClassID
                // DID match, so we can just return this entry. Also, if the
                // conditional check from the "case 1:" section was successful,
                // it will fall through to this case and also return the Entry.
                return l_pEntry;
            default:
                // Otherwise more than just the Species was different, so keep
                // looking.
                break;
        }
        l_pEntry = l_pEntry->m_pNext;
    }
    
    return NULL;
}

/*
===============================================================================
    Global Helper functions
===============================================================================
    Function to return the ClassID of a Definition
===============================================================================
*/
uint32 DefinitionClassID(void* pDefinition)
{
    uint32 l_RetValue = (uint32)ANY_CLASS::CLASSID;
    
    if(pDefinition != NULL)
    {
        LINE_COPY(&l_RetValue, ((const byte*)pDefinition)
                              + sizeof(uint8),              // Species Count
                                sizeof(uint32));
    }

    return l_RetValue;    
}
/*
===============================================================================
    Function to return the Version of a Definition
===============================================================================
*/
uint16 DefinitionVersion(void* pDefinition)
{
    uint16 l_RetValue = 0;
    
    if(pDefinition != NULL)
    {
        LINE_COPY(&l_RetValue, ((const byte*)pDefinition)
                              + sizeof(uint8)               // Species Count
                              + sizeof(uint16)              // Family
                              + sizeof(uint8)               // Genus
                              + *((const byte*)pDefinition), // All Species
                                sizeof(uint16));
    }

    return l_RetValue;    
}

/*
===============================================================================
    Function to return the byte-size of a Definition
===============================================================================
*/
uint16 DefinitionSize(void* pDefinition)
{
    uint16 l_RetValue = 0;
    
    if(pDefinition != NULL)
    {
        LINE_COPY(&l_RetValue, ((const byte*)pDefinition)  
                              + sizeof(uint8)               // Species Count
                              + sizeof(uint16)              // Family
                              + sizeof(uint8)               // Genus
                              + *((const byte*)pDefinition)  // All Species
                              + sizeof(uint16),             // Version
                                sizeof(uint16));
    }

    return l_RetValue;    
}
/*
===============================================================================
    Function to return the full byte-size of a Definition
===============================================================================
*/
uint16 DefinitionFullSize(void* pDefinition)
{
    uint16 l_RetValue = 0;
    
    if(pDefinition != NULL)
    {
        LINE_COPY(&l_RetValue, ((const byte*)pDefinition)  
                              + sizeof(uint8)               // Species Count
                              + sizeof(uint16)              // Family
                              + sizeof(uint8)               // Genus
                              + *((const byte*)pDefinition) // All Species
                              + sizeof(uint16)              // Version
                              + sizeof(uint16),             // DefSize
                                sizeof(uint16));
    }

    return l_RetValue;    
}
/*
===============================================================================
    Function to return the size of a Definition's base set of XML strings.
    This will be 0 if there are no XML strings, and implies XML 
    (De)Serialization is not supported with this definition.
===============================================================================
*/
uint16 DefinitionXMLSize(void* pDefinition)
{
    uint16 l_RetValue = 0;
    
    if(pDefinition != NULL)
    {
        LINE_COPY(&l_RetValue, ((const byte*)pDefinition)  
                              + sizeof(uint8)               // Species Count
                              + sizeof(uint16)              // Family
                              + sizeof(uint8)               // Genus
                              + *((const byte*)pDefinition) // All Species
                              + sizeof(uint16)              // Version
                              + sizeof(uint16)              // DefSize
                              + sizeof(uint16),             // Full DefSize
                                sizeof(uint16));
    }

    return l_RetValue;    
}
/*
===============================================================================
    Function to return the total number of members in a definition
===============================================================================
*/
uint16 DefinitionMemberCount(void* pDefinition)
{
    uint16 l_RetValue = 0;
    
    if(pDefinition != NULL)
    {
        LINE_COPY(&l_RetValue, ((const byte*)pDefinition) 
                              + sizeof(uint8)               // Species Count
                              + sizeof(uint16)              // Family
                              + sizeof(uint8)               // Genus
                              + *((const byte*)pDefinition) // All Species
                              + sizeof(uint16)              // Version
                              + sizeof(uint16)              // DefSize
                              + sizeof(uint16)              // Full DefSize
                              + sizeof(uint16),             // XMLBaseSize
                                sizeof(uint16));
    }

    return l_RetValue;    
}

/*
===============================================================================
    Function to return the total number of members in a definition
===============================================================================
*/
uint16 DefinitionMemberCount(void* pDefinition, uint16 Version)
{
    uint16 l_RetValue = 0;
    
    // If the Definition doesn't exist, we can't parse it for our structure, 
    // so just skip it.
    if(pDefinition != NULL)
    {
        uint32  l_Offset = DefinitionFullSize(pDefinition) - SIZEOF_CRC;
        uint16  l_CRC = CalculateCRC((uint8*)pDefinition, l_Offset);
        uint16  l_MinVersion = 0;
        uint16  l_MaxVersion = 0;
        uint16  l_MemberCount = 0;
        byte*   l_pDefPtr = (byte*)pDefinition;

        if(LINE_CHECK(((uint8*)pDefinition) + l_Offset, &l_CRC, SIZEOF_CRC))
        {
            l_pDefPtr += sizeof(uint8)     // Skip Species Count
                       + sizeof(uint16)    // Skip Family
                       + sizeof(uint8)     // Skip Genus
                       + (*l_pDefPtr);     // Skip all the Species

            LINE_COPY(&l_MaxVersion, l_pDefPtr, SIZEOF_VERSION);
            l_pDefPtr += SIZEOF_VERSION;

            // Check to make sure the requested version is less than or
            // equal to the current definition's version.
            if(Version <= l_MaxVersion)
            {
                l_pDefPtr += sizeof(uint16) // Skip the DefSize
                           + sizeof(uint16) // Skip the Full DefSize
                           + sizeof(uint16);// Skip the Base XML Size
                
                // Grab the total member count
                LINE_COPY(&l_MemberCount, l_pDefPtr, SIZEOF_MEMBER_COUNT);
                l_pDefPtr += SIZEOF_MEMBER_COUNT;

                // We have to parse the full definition once to see what data members
                // we will be using in this version of the object...
                while(l_MemberCount-- > 0)
                {            
                    // Check to see if we are greater than the Min Version
                    LINE_COPY(&l_MinVersion, l_pDefPtr, SIZEOF_VERSION);
                    l_pDefPtr += SIZEOF_VERSION;
                    // Check to see if we are less than the Max Version
                    LINE_COPY(&l_MaxVersion, l_pDefPtr, SIZEOF_VERSION);
                    l_pDefPtr += SIZEOF_VERSION;

                    // This member is compatible, so increase our count.
                    if(l_MinVersion <= Version && Version <= l_MaxVersion)
                    {
                        l_RetValue++;
                    }

                    // Skip the ClassID and array elements if they are there.
                    l_pDefPtr += (((*l_pDefPtr & DMOIB_TYPE_MASK_BW) == DMOIB_TYPE_FLAG_OBJ) ? SIZEOF_CLASSID : 0) + 
                                  ((*l_pDefPtr & DMOIB_TYPE_FLAG_ARR) ? SIZEOF_ARRAY_SIZE : 0);

                    // Skip the actual member type information.
                    l_pDefPtr++;
                }
            }
        }
    }

    return l_RetValue;    
}

/*
===============================================================================
    Function to return the member type for a specified member
===============================================================================
*/
uint8 DefinitionMemberAtType(void* pDefinition, uint16 Offset, uint32* pClassID, uint16* pSize)
{
    uint8 l_MemberType = DMOIB_TYPE_UNKNOWN;

    if(pDefinition != NULL)
    {   
        byte*   l_pDefinition = (byte*)pDefinition;
        uint16  l_MemberCount = 0;     
        uint16  l_MemberIter = 0;
        uint8   l_CurrentMemberType = DMOIB_TYPE_UNKNOWN;
    
        // Skip past all the stuff at the front of the definition
        l_pDefinition += sizeof(uint8)              // Species Count
                       + sizeof(uint16)             // Family
                       + sizeof(uint8)              // Genus
                       + *l_pDefinition             // All Species
                       + sizeof(uint16)             // Version
                       + sizeof(uint16)             // DefSize
                       + sizeof(uint16)             // Full DefSize
                       + sizeof(uint16);            // XMLBaseSize
                       
        // Grab the member count
        LINE_COPY(&l_MemberCount, l_pDefinition, SIZEOF_MEMBER_COUNT);
        l_pDefinition += SIZEOF_MEMBER_COUNT;
                       
        // We have to parse the full definition up to the point of our offset
        // member since each member section is variable size.
        for(l_MemberIter = 0; l_MemberIter < l_MemberCount; l_MemberIter++)
        {            
            // Skip min/max versions
            l_pDefinition += (SIZEOF_VERSION + SIZEOF_VERSION);
            
            // Grab the current member type and move past it
            l_CurrentMemberType = *l_pDefinition;
            l_pDefinition += SIZEOF_TYPE_INFO;

            // Check to see if this member is an object type
            if((l_CurrentMemberType & DMOIB_TYPE_MASK_BW) == DMOIB_TYPE_FLAG_OBJ)
            {
                // If the pClassID variable is not NULL and our current member 
                // matches the requested offset, we need to fill it with the 
                // ClassID information for this member
                if(pClassID != NULL && Offset == l_MemberIter)
                {
                    LINE_COPY(pClassID, l_pDefinition, SIZEOF_CLASSID);
                }
                
                l_pDefinition += SIZEOF_CLASSID;
            }
                
            if(l_CurrentMemberType & DMOIB_TYPE_FLAG_ARR)
            {
                // If the pSize variable is not NULL and our current member 
                // matches the requested offset, we need to fill it with the 
                // array size information for this member
                if(pSize != NULL && Offset == l_MemberIter)
                {
                    LINE_COPY(pSize, l_pDefinition, SIZEOF_ARRAY_SIZE);
                }
                
                l_pDefinition += SIZEOF_ARRAY_SIZE;
            }
            
            if(Offset == l_MemberIter)
            {
                l_MemberType = l_CurrentMemberType;
                break;
            }
        }
    }
    
    return l_MemberType;
}

/*
===============================================================================
    Function to check if the specified Species is a valid entry in the 
    specified definition.
===============================================================================
*/
boolean IsSpeciesValid(void* pDefinition, uint8 Species)
{
    boolean l_bRetVal = FALSE;

    if(pDefinition != NULL)
    {
        uint8*  l_pSpeciesIter = ((uint8*)pDefinition)  
                                + sizeof(uint8)               // Species Count
                                + sizeof(uint16)              // Family
                                + sizeof(uint8);              // Genus

        uint8*  l_pSpeciesEnd = l_pSpeciesIter + *((uint8*)pDefinition);      

        while(l_bRetVal == FALSE && l_pSpeciesIter != l_pSpeciesEnd)
        {
            l_bRetVal = (*l_pSpeciesIter++ == Species);
        }
    }
    
    return l_bRetVal;
}
