/*-D************************************************************************* 
'*   This program source file is confidential and proprietary property of
'*    Welch Allyn, Inc. Unauthorized use or reproduction is prohibited.
'****************************************************************************
'****************************************************************************
'**
'****************************************************************************
'**  ////////////////////////////////////////////////////////////////////////
'**  '// CBaseCommSvr.cpp:                                                 //
'**  '// Language       :   C++                                            //
'**  '// Application    :   WACP                                           //
'**  '// Sources        :   MOMP Standard Library                          //
'**  '///////////////////////////////////////////////////////////////////////
'*-E************************************************************************/

#include "CBaseCommSvr.h"
#include "CCommSvr.h"
#include "Mem.h"
#include "SysUtil.h"


#ifndef WACP_MAX
#define WACP_MAX(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifndef WACP_MIN
#define WACP_MIN(a,b)            (((a) < (b)) ? (a) : (b))
#endif

/*================================================================================
    CONSTRUCTORS / DESTRUCTORS
================================================================================*/
/*================================================================================
  Default constructor
================================================================================*/
CBaseCommSvr::CBaseCommSvr()
{
    pSetRecvMsg = NULL;
    pDeviceArrival = NULL;
    pDeviceRemoval = NULL;
    pSemName = NULL;

    /* Zero out structure */
    memset(SourceID, 0, sizeof(SourceID));         
    
    /* Initialize the Port */
    Port = 0;

    /* Set initial parsing buffer */
    pParseBuf = NULL;
    ParseBufSize = 0;
    ParseState = 0;

    /* Initialize the stack type */
    StackType = STACK_UNDEFINED;
}

/*================================================================================
  Default destructor
================================================================================*/
CBaseCommSvr::~CBaseCommSvr()
{
    if (pParseBuf)
    {
        SYS_FREE(pParseBuf);
    }
}

/*================================================================================
CBaseCommSvr Init: Initialization method for setting up callbacks and receive 
                   Semaphore name.
Parameters:
    RecvMsgFunc:         Pointer to a function that processes new WACP messages.
    DeviceArrivalFunc:   Pointer to a function that creates a connection.
    DeviceRemovalFunc:   Pointer to a function that terminates a connection.
    pAppName:        Semaphore name to be sent in the calls to RecvMsg and 
                     TerminateClient callbacks.
================================================================================*/
void CBaseCommSvr::Init(SetRecvBufPtr RecvMsgFunc,
                        DevArrPtr DeviceArrivalFunc,
                        DevRemPtr DeviceRemovalFunc,
                        char* pAppName)
                  
{
    /* Store the callback function to add a new message */
    pSetRecvMsg = RecvMsgFunc;

    /* Store the callback function to create a connection */
    pDeviceArrival = DeviceArrivalFunc;
    
    /* Store the callback function to terminate a connection */
    pDeviceRemoval = DeviceRemovalFunc;
    
    /* Save the application name that is created from the Semaphore name.
     * This is used to differentiate between separate dll instances of the stack.
     */
    pSemName = pAppName;
}

/*================================================================================
CBaseCommSvr GetSourceID: Returns the SourceID of the object

Returns: 
    Pointer to the SourceID
================================================================================*/
char* CBaseCommSvr::GetSourceID()
{
    return((char*)SourceID);
}
/*================================================================================
CBaseCommSvr SetSourceID: Sets the SourceID

Parameters:
    pSourceID:      SourceID of the connection.

Returns: 
    TRUE, if the pSourceID is valid and not the same as the current SourceID.
    FALSE, if the pSourceID is NULL or the same as the current SourceID. 
================================================================================*/
boolean CBaseCommSvr::SetSourceID(char* pSourceID)
{   
    boolean Result = FALSE;
    if ((pSourceID != NULL) && (pSourceID != SourceID))
    {
        uint16 Size;

        memset(SourceID, 0, sizeof(SourceID));
        Size = (uint16)strlen(pSourceID);
        if (Size >= sizeof(SourceID))
        {
            /* Always NULL terminate the string */
            Size = sizeof(SourceID) - 1;
        }

        memcpy(SourceID, pSourceID, Size);
        Result = TRUE;
    }

    return (Result);
}
/*================================================================================
CBaseCommSvr GetPortNum: Returns the Port used by the CBaseCommSvr

Returns: 
    Port value.
================================================================================*/
uint16 CBaseCommSvr::GetPortNum()
{
    return(Port);
}

/*================================================================================
CBaseCommSvr SetPortNum: Sets the Port used by the CBaseCommSvr

Parameters:
    NewPort:      New port value to be stored.

Returns: 
    Always returns TRUE 
================================================================================*/
boolean CBaseCommSvr::SetPortNum(uint16 NewPort)
{
    boolean Result = TRUE;

    Port = NewPort;

    return (Result);
}

/*================================================================================
CBaseCommSvr GetStackType: Returns the type of stack that is being used 
                           ( i.e. Ethernet, RS232, Bluetooth, USB, etc )
Returns: 
    The stacktype value.
================================================================================*/
STACK_TYPE CBaseCommSvr::GetStackType()
{
    return(StackType);
}

/*================================================================================
CBaseCommSvr::StartReader - Pass through function that calls on CWACPThread to 
    start the CommSvr reader thread. This allows the CBaseCommSvr instance to 
    hold onto a CWACPThread instance for the life of the thread.

Parameters:
    StartAddress - Address of the thread process function. WACPAPI is specified
                   in Target.h

    pArgList - Pointer to data that is to be passed into the thread process 
               function as a parameter. 

    ThreadPriority - Priority value of the thread based on OS specific settings.
                     For Win32 THREAD_PRIORITY_NORMAL is 0 with higher priority
                     settings having positive values and lower priority having
                     negative values. 
    pName - Pointer to the thread name. Primarily used for identifying threads 
            for debug. 

Returns: 
    TRUE when the Thread is created and priority is set
    FALSE, if the an error occurred in creating the thread or setting the 
        priority. 
================================================================================*/
boolean CBaseCommSvr::StartReader(uint32 (WACPAPI* StartAddress)(void*), void* pArgList, 
                                 int32 ThreadPriority, char* pName)
{
    boolean Result = FALSE;

    Result = RThread.StartThread(StartAddress, pArgList, ThreadPriority, pName);

    return(Result);
}

/*================================================================================
CBaseCommSvr Arrival: Starts the RNDZ connection sequence by creating an 
                      autodiscovery message and passing it up to the Interchange 
                      layer.
Parameters:
    pSourceID:      SourceID of the connection.
    StackType:      Communication stack type.
    pConfig:        Addr structure specific to stack type.

Returns: TRUE or FALSE
    TRUE:  If the auto discovery message was processed.
    FALSE: If GeneratePacket fails to generate the AutoDiscovery message or if
           recvmsg callback fails to process the AutoDiscovery message.

================================================================================*/
boolean CBaseCommSvr::Arrival(char* pSourceID, STACK_TYPE StackType, 
                              void* pConfig, uint16 ConfigSize)
{
    boolean Result = FALSE;
    uint32  PacketLen;    
    uint8*  pPacketBuf = NULL;
    char*   pNewSourceID = pSourceID;

    ConnArgs* pConnArgs = (ConnArgs*)SYS_CALLOC(sizeof(ConnArgs), 1);
    
    /* Check for a valid pointer */
    if (pConnArgs && pSourceID)
    {
        if (pConfig)
        {
            pConnArgs->StackType = StackType;
            pConnArgs->RNDZPort = RNDZ_PORT_DEFAULT;
            pConnArgs->WACPPort = WACP_PORT_DEFAULT;
            pConnArgs->bSvrFlag = FALSE;
            pConnArgs->EncryptionType = WACP_ENCRYPTION_NONE;
            pConnArgs->Timeout = DEFAULT_RNDZ_TIMEOUT;
            pConnArgs->ConfigSize = ConfigSize; 
            pConnArgs->pConfig = pConfig; 
            memcpy(pConnArgs->SourceID, pSourceID, strlen(pSourceID)); 

            /* Call the Device Arrival callback to setup a new client connection */
            pNewSourceID = pDeviceArrival(pSemName, pSourceID, pConnArgs);
            SYS_FREE(pConnArgs);
        }

        if (pNewSourceID)
        {
            static const uint8 RndzConnect[] = /* RNDZ_CRST_SpSTART_RNDZ */
            { 
                0x00, 0x1D, 0x0B, 0x01, 0x00, 0x00, 0x00, 0x12,
                0x00, 0x00, 0x00, 0x00, 0x0B, 0x52, 0x4E, 0x44,
                0x5A, 0x43, 0x4F, 0x4E, 0x4E, 0x45, 0x43, 0x54,
                0x57, 0xF3
            };

            /* Generate the AutoDiscovery packet */
            pPacketBuf = GeneratePacket(&PacketLen,
                                        Port, 0, 0, 
                                        (uint8*)RndzConnect,
                                        sizeof(RndzConnect));
            if (pPacketBuf)
            {
#ifdef DEBUG_PRINT
                SYS_PRINTF("CBaseCommSvr::Arrival: %s AutoDiscovery Packet\n", pNewSourceID);
#endif
                /* Call the receive message callback to process the new message */
                Result = pSetRecvMsg(pSemName, pNewSourceID, (uint8*)pPacketBuf, PacketLen);

                /* Free up the memory allocated by GeneratePacket */
                SYS_FREE(pPacketBuf);
            }
        }/* if (pNewSourceID) */
    }/* if (pConnArgs && pSourceID) */

    return (Result);
}


/*================================================================================
CBaseCommSvr Removal: Calls the terminate callback to terminate the connection.
================================================================================*/
boolean CBaseCommSvr::Removal()
{
#ifdef DEBUG_PRINT
   SYS_PRINTF("CBaseCommSvr::Removal: %s\n", SourceID);
#endif

    boolean Result = pDeviceRemoval(pSemName, SourceID);

    return (Result);
}

/*================================================================================
CBaseCommSvr WACPParser: Parses a data stream for a WACP packets. This will process
                         all WACP packets contained in the data stream.
Parameters:
    pBuf:           Pointer to the data stream
    BufSize:        Size of the data stream.
    pPacketLength:  Pointer to a value that will return the packet length
                    for the current WACP packet.

Returns: The number of bytes processed. 
    0, if a partial WACP packet was not found.
    BufSize, if a WACP packet was not found.

    If the WACP header is found *pPacketLength is set to the current packet length.
    Otherwise, the Packetlength is set to 0.
================================================================================*/
uint32 CBaseCommSvr::WACPParser(uint8* pBuf, uint32 BufSize, uint32 *pPacketLength)
{
    uint32      BytesProcessed = 0;
    uint32      PacketLen;
    uint8*      pTemp = pBuf;

    /* Initialize the PacketLength pointer to 0 */
    if (pPacketLength != NULL)
    {
        *pPacketLength = 0;
    }

    for (;;)
    {
        /* Only process messages that contain a full message header */
        if (BufSize > MSG_HEADER)
        {
            /* Verify the Preamble */
            if( (pBuf[0] == CTL_W) && 
                (pBuf[1] == CTL_A) && 
                (pBuf[2] == CTL_L) )
            {
                /* Get the packet length */
                pTemp = (uint8*)&pBuf[PREAMBLE_SIZE];
                PacketLen = (int)ReadInt32(pTemp);

                /* Check to see if the buffer size is greater than or equal to the PacketLength 
                 * and that we don't exceed the max buffer size.
                 */
                if ((PacketLen <= BufSize) && (PacketLen < RCVBUFSIZE))
                {
#ifdef DEBUG_PRINT
                    uint32 i;
                    SYS_PRINTF("\n\nNew Msg:\n");
                    for (i=0; i < PacketLen;i++)
                    {
                        if (i % 25 == 0)
                            SYS_PRINTF("\n");
                        SYS_PRINTF("%.2x ", (uint8)pBuf[i]);
                    }
                    SYS_PRINTF("\n\n");
#endif
                    /* Update the Recv IN_queue with the new message */
                    pSetRecvMsg(pSemName, SourceID, (uint8*)pBuf, PacketLen);

                    /* Increment the pBuf pointer to the next byte after the WACP packet */
                    pBuf += PacketLen;

                    /* Update the size of the remaining data stream */
                    BufSize -= PacketLen;

                    /* Update the number of bytes parsed */
                    BytesProcessed += PacketLen;
                    continue;
                }
                else
                {
                    if (pPacketLength != NULL)
                    {
                        /* Update the number of bytes of the current packet */
                        *pPacketLength = PacketLen;
                    }
#ifdef DEBUG_PRINT
                    SYS_PRINTF("Partial WACP Packet found: Current Packet length = %d\n", PacketLen);
                    SYS_PRINTF("BufSize = %d, BytesProcessed = %d\n", BufSize, BytesProcessed);
#endif
                }
                break;
            }
            else
            {
                if (BufSize > PREAMBLE_SIZE)
                {
#ifdef DEBUG_PRINT
                    SYS_PRINTF("Found invalid preamble, checking next byte\n");
                    SYS_PRINTF("BufSize = %d, BytesProcessed = %d\n", BufSize, BytesProcessed);
#endif
                    /* Update the number of bytes parsed */
                    BytesProcessed++;

                    /* Update the number of bytes left to parsed */
                    BufSize--;

                    /* increment the pointer */
                    pBuf++;
                }
                else
                {
#ifdef DEBUG_PRINT_STATUS
                    SYS_PRINTF("No valid PREAMBLE found\n");
#endif
                    break;
                }
            }

        }/* if (MsgSize > MSG_HEADER) */
        else
        {
#ifdef DEBUG_PRINT
            SYS_PRINTF("BufSize = %d, BytesProcessed = %d\n", BufSize, BytesProcessed);
#endif
            break;
        }

    }/* for (;;) */


    return BytesProcessed;
}

/*================================================================================
CBaseCommSvr WACPParserEx: Parses a data stream for a WACP packets. This will process
                         all WACP packets contained in the data stream, and allocate
                         a buffer only if a partial packet is found.  Full packets are
                         sent directly to RecvMsg queue.
Parameters:
    pBuf:           Pointer to the data stream
    BufSize:        Size of the data stream.
    pPacketLengthHint: Optional pointer to a value that will return the bytes required
                    for remainder of the current WACP packet, or -1 if an incomplete
                    WACP header.

Returns: The number of bytes processed. Since this implementation parses through
         invalid WACP data, the return value should normally equal the BufSize.
================================================================================*/
uint32 CBaseCommSvr::WACPParserEx(uint8* pBuf, uint32 BufSize, uint32* pPacketLengthHint)
{
    uint32 PacketLen, MinSize = 0, BytesProcessed = 0, Bytes;
    uint8* pTmpBuf;

    /* Initialize the PacketLength pointer to 0 */
    if (pPacketLengthHint != NULL)
    {
        *pPacketLengthHint = 0;
    }
    for (;BufSize;)
    {
        if (ParseState < PREAMBLE_SIZE)
        {
            if (0 != memcmp(WACP_PREAMBLE + ParseState, pBuf, WACP_MIN(PREAMBLE_SIZE - ParseState, BufSize)))
            {
                if (pParseBuf)
                {
                    SYS_FREE(pParseBuf);
                    pParseBuf = NULL;
                    ParseBufSize = 0;
                }
                ParseState = 0;
                if (--BufSize)
                {
                    pBuf++;
                }
                BytesProcessed++;
                continue;
            }
        }
        if (ParseState <= MSG_HEADER)
        {
            if (NULL == pParseBuf)
            {
                ParseBufSize = PARSEREX_MIN_BUFSIZE;
                pParseBuf = (uint8*)SYS_REALLOC(pParseBuf, ParseBufSize); //allocate some reasonable space
                if (NULL == pParseBuf)
                {
                    break;
                }
            }
            MinSize = WACP_MIN(MSG_HEADER - ParseState, BufSize);
            memcpy(pParseBuf + ParseState, pBuf, MinSize);
            if (ParseState + BufSize <= MSG_HEADER)
            {
                if (NULL != pPacketLengthHint)
                {	//send as much as you can next time
                    *pPacketLengthHint = (uint32)-1;
                }
                BytesProcessed = BufSize;
                ParseState += MinSize;
                break;
            }
        }
        pTmpBuf = pParseBuf ? pParseBuf : pBuf;
        PacketLen = ReadInt32(&pTmpBuf[PREAMBLE_SIZE]);
        if (PacketLen > RCVBUFSIZE)
        {
            if (pParseBuf)
            {
                SYS_FREE(pParseBuf);
                pParseBuf = NULL;
                ParseBufSize = 0;
            }
            ParseState = 0;
            MinSize = WACP_MIN(MSG_HEADER, BufSize);
            BufSize -= MinSize;
            pBuf += MinSize;
            BytesProcessed += MinSize;
            continue;
        }
        if (pParseBuf)
        {
            if (ParseBufSize != PacketLen)
            {
                ParseBufSize = PacketLen;
                pTmpBuf = pParseBuf = (uint8*)SYS_REALLOC(pTmpBuf, ParseBufSize);
                if (NULL == pParseBuf)
                {
                    break;
                }
            }
            MinSize = WACP_MIN(ParseBufSize - ParseState, BufSize);
            memcpy(pTmpBuf + ParseState, pBuf, MinSize);
        }
        if (PacketLen <= BufSize + ParseState && PacketLen < RCVBUFSIZE)
        {
            pSetRecvMsg(pSemName, SourceID, pTmpBuf, PacketLen);
            MinSize = WACP_MIN(BufSize, PacketLen);
            Bytes = ParseState ? PacketLen - ParseState : MinSize;
            BufSize -= Bytes;
            SYS_FREE(pParseBuf);
            pParseBuf = NULL;
            ParseBufSize = 0;
            if (BufSize)
            {
                pBuf += Bytes;
            }
            BytesProcessed += Bytes;
            ParseState = 0;
        }
        else if (PacketLen > BufSize + ParseState)
        {
            uint32 Remainder = 0;
            if (pPacketLengthHint != NULL)
            {
                *pPacketLengthHint = PacketLen - BufSize + ParseState;
            }
            if (NULL == pParseBuf)
            {
                ParseBufSize = PacketLen;
                pParseBuf = (uint8*)SYS_MALLOC(ParseBufSize);
                MinSize = WACP_MIN(BufSize, ParseBufSize);
                memcpy(pParseBuf, pTmpBuf, MinSize);
                BytesProcessed = MinSize;
                BufSize = 0;
            }
            else
            {
                BytesProcessed += MinSize;
                if (BufSize >= BytesProcessed)
                {
                    BufSize -= BytesProcessed;
                    pBuf += MinSize;
                }
                else
                {
                    Remainder = WACP_MAX(BufSize, BytesProcessed) - WACP_MIN(BufSize, BytesProcessed);
                    BufSize = 0;
                }
            }
            ParseState += BytesProcessed - Remainder;
        }
    }

    return BytesProcessed;
}
/* End of File */
