/*-D************************************************************************* 
'*   This program source file is confidential and proprietary property of
'*    Welch Allyn, Inc. Unauthorized use or reproduction is prohibited.
'****************************************************************************
'****************************************************************************
'**
'****************************************************************************
'**  ////////////////////////////////////////////////////////////////////////
'**  '// CWACPApp.cpp                                                      //
'**  '// Language       :   C++                                            //
'**  '// Application    :   WACP                                           //
'**  '// Sources        :   MOMP Standard Library                          //
'**  '///////////////////////////////////////////////////////////////////////
'*-E************************************************************************/

#include "CWACPApp.h"
#include "CMOMPbaseclass.h"
#include "SysUtil.h"

#if defined(TRACE_SERVER_ON)
    #include "CTraceSvr.h"
#endif

/*================================================================================
    CONSTRUCTORS / DESTRUCTORS
================================================================================*/
/*********************************************************************************
    Default constructor
    CWACPApp Start up: All available servers are instantiated here. 
    i.e Trap, Connection, Data and Rendezvous object servers.
    Starts up the Application Reader thread.
*********************************************************************************/
CWACPApp::CWACPApp(boolean (*SendPacket)(char*, char*, uint8*, uint32),
                   boolean (*MsgFunc)(char*, uint16, void*, uint16),
                   CLinkTable* pNewCLinkTable,
                   LinkStruct* pLinkStruct)
{
    /* Set the pointer to the link table and the CLinkTable instance */
    pLnkStruct = pLinkStruct;
    pCLinkTable = pNewCLinkTable;

    bAppInit = FALSE;

    bActive = TRUE;

    pCallBackObj = (CMOMPBaseClass*)pLnkStruct->pCallBackObj;
    pCMOMPBaseClass = NULL;

    /* Set the Link Layer callback function for sending a packet */
    pSendPacket = SendPacket;

    /* Initialize the RNDZStub */
    RNDZStub.Init(this);

    /* Initialize the Object Servers */
    TrapObjSvr.Init(SendPacket, RxErrorWrapper, pLnkStruct);
    FmRouter.Init(&TrapObjSvr, this);
    ConObjSvr.Init(SendPacket, MsgFunc, RxConnWrapper, this, pLnkStruct);
    RendezvousSvr.Init(&TrapObjSvr, SendPacket, RxRNDZMessageWrapper, pLnkStruct);

    /* This message server is used for all Family types registered by the application */
    DataObjSvr.Init(&TrapObjSvr, SendPacket, RxMessageWrapper, pLnkStruct);

    /* Add the Object server to the FmRouter table */
    FmRouter.RegisterServer(FmTRAP, &TrapObjSvr);
    FmRouter.RegisterServer(FmRENDEZVOUS, &RendezvousSvr);
    FmRouter.RegisterServer(FmCONNECTION, &ConObjSvr);
    FmRouter.RegisterServer(FmMAX, &DataObjSvr);
}

CWACPApp::~CWACPApp()
{
    bActive = FALSE;

    /* Only call AppEnd if AppInit has been previously called */
    if (bAppInit == TRUE)
    {
        /* Callback into the calling object notifying the application is coming to a close */
        pCallBackObj->AppEndInt(this);
    }
}

/*================================================================================
Static Callback function wrapper for handling RNDZ messages
================================================================================*/
uint32 CWACPApp::RxRNDZMessageWrapper(char* pSourceID, uint32 ClassID, uint8* pBuf, void* pWACPApp)
{
    uint32  Result = TRUE;

    /* Call back into the RNDZStub class to handle a RNDZ Status message */
    if (pWACPApp)
    {
        Result = ((CWACPApp*)pWACPApp)->RNDZStub.RecvMessage(pSourceID, ClassID, pBuf);
    }
    return (Result);
}

/*================================================================================
Static Callback function wrapper for handling Command, Request, Response, 
Status, Event, and Data Object messages
================================================================================*/
uint32 CWACPApp::RxMessageWrapper(char* pSourceID, uint32 ClassID, uint8* pBuf, void* pWACPApp)
{
    uint32 Result = 0;
    ConnArgs* pConnArgs;

    /* Check for valid linktable entry */
    pConnArgs = ((CWACPApp*)pWACPApp)->pCLinkTable->GetEntry(pSourceID);
    if (pConnArgs)
    {
        /* Check to see if we passed RNDZ authentication or if it's disabled */
        if ((pConnArgs->bRNDZStatus) || (pConnArgs->bRNDZDisabled == TRUE))
        {
            /* Call back into the current MOMPBaseclass if it exists to notify the
             * application stub class that a data message has arrived
             */
            if (((CWACPApp*)pWACPApp)->pCMOMPBaseClass)
            {
                Result = ((CWACPApp*)pWACPApp)->pCMOMPBaseClass->RecvMessage(pSourceID, ClassID, pBuf);
            }
            else
            {
                /* Otherwise callback into the original mompbaseclass passed into the stack */
                Result = ((CWACPApp*)pWACPApp)->pCallBackObj->RecvMessage(pSourceID, ClassID, pBuf);
            }
        }
#if defined(TRACE_SERVER_ON)
        else
        {
            SYS_TRACE(TRACE_Type_ERROR, "CWACPApp", "RxMessageWrapper",
                      0, 0, 0, "Unauthenticated Message: RNDZ Authentication incomplete");
        }
#endif
    }

    return (Result);
}
/*================================================================================
Static Callback function wrapper for handling Error messages
================================================================================*/
uint32 CWACPApp::RxErrorWrapper(char* pSourceID, uint8 ErrorType, uint32 ClassID, 
                                uint8* pBuf, void* pWACPApp)
{
    uint32 Result = 0;

    /* Call back into the current MOMPBaseclass if it exists to notify the
     * application stub class that an error message has arrived
     */
    if (((CWACPApp*)pWACPApp)->pCMOMPBaseClass)
    {
        Result = ((CWACPApp*)pWACPApp)->pCMOMPBaseClass->RecvError(pSourceID, ErrorType, ClassID, pBuf);
    }
    else
    {
        /* Otherwise callback into the original mompbaseclass passed into the stack */
        Result = ((CWACPApp*)pWACPApp)->pCallBackObj->RecvError(pSourceID, ErrorType, ClassID, pBuf);
    }

    return (Result);
}

/*================================================================================
Static Callback function wrapper for handling Connection messages
================================================================================*/
uint32 CWACPApp::RxConnWrapper(char* pSourceID, uint32 ClassID, uint8* pBuf, void* pWACPApp)
{
    uint32 Result = 0;
    ConnArgs* pConnArgs;
    uint8 Genus = GetGenus(ClassID);
    uint8 Species = GetSpecies(ClassID);

    /* Check for termination sequence */
    if ((Genus == GnCOMMAND) && (Species == CONN_CMD_SpSHUTDOWN))
    {
        /* Grab the connection information to check for an incomplete RNDZ authentication */
        pConnArgs = ((CWACPApp*)pWACPApp)->pCLinkTable->GetEntry(pSourceID);
        if (pConnArgs)
        {
            if (pConnArgs->pByPassSrc)
            {
                ConnArgs* pByConnArgs = ((CWACPApp*)pWACPApp)->pCLinkTable->GetEntry(pConnArgs->pByPassSrc);
                if (pByConnArgs)
                {
                    ((CWACPApp*)pWACPApp)->SendMsg(pByConnArgs->SourceID, FmCONNECTION, GnCOMMAND, 
                                                   CONN_CMD_SpBYPASSTERMINATED, NULL, 0, FALSE);
                    ((CWACPApp*)pWACPApp)->SetByPass(pByConnArgs->SourceID, NULL);
                }
            }
            /* Terminate RNDZ if the RNDZ authentication has not passed */
            if((pConnArgs->bRNDZStatus == FALSE) && (pConnArgs->bRNDZDisabled == FALSE))
                ((CWACPApp*)pWACPApp)->RNDZTerminate(pSourceID, FALSE);
       }
    }

    /* Call back into the current MOMPBaseclass if it exists to notify the
     * application stub class that a connection status message has arrived
     */
    if (((CWACPApp*)pWACPApp)->pCMOMPBaseClass)
    {
        Result = ((CWACPApp*)pWACPApp)->pCMOMPBaseClass->RecvMessage(pSourceID, ClassID, pBuf);
    }
    else
    {
        /* Otherwise callback into the original mompbaseclass passed into the stack */
        Result = ((CWACPApp*)pWACPApp)->pCallBackObj->RecvMessage(pSourceID, ClassID, pBuf);
    }

    return (Result);
}
/*================================================================================
RNDZTerminate : Terminates the current RNDZ connection with status being set
    to TRUE or FALSE. If TRUE, then we push the connection object server settings.
================================================================================*/
void CWACPApp::RNDZTerminate(char* pSourceID, boolean bStatus)
{
    ConnArgs* pConnArgs;

#if defined(TRACE_SERVER_ON)
    SYS_TRACE(TRACE_Type_SEND, "CWACPApp", "RNDZTerminate", 0,0,0, 
        "%s: \nTerminating RNDZ with bStatus = %d", pSourceID, bStatus);
#endif
    pConnArgs = pCLinkTable->GetEntry(pSourceID);
    if (pConnArgs)
    {
        pConnArgs->bRNDZStatus = bStatus;

        if (bStatus)
        {
            /* Close the RNDZ Connection */
            ConObjSvr.UpdateStatus(pConnArgs->SourceID, bStatus);

            /* Return the ConObjSvr option to the default for the WACP application */
            SetConnOptions(pConnArgs->SourceID, DEFAULT_CONN_SETTING, FALSE, 
                           DEFAULT_KEEPALIVE_INTERVAL, DEFAULT_TIMEOUT_INTERVAL);

            /* Force the TxOptions to be synced after a connect. This is to ensure that
            * after a UDP connection has been made, some data is sent to the server
            * to let it know a client is available.
            */
            SetConnOptions(pConnArgs->SourceID, DEFAULT_CONN_SETTING, TRUE, 
                           DEFAULT_KEEPALIVE_INTERVAL, DEFAULT_TIMEOUT_INTERVAL);
        }

        /* Notify the application of the connection status */
        if (bStatus)
            RxConnWrapper(pSourceID, MakeClassID(FmCONNECTION, GnEVENT, CONN_CEVNT_SpCONNECTED), NULL, this);
    }
}

/*================================================================================
RNDZConnect: Sends a command to initiate a Rendezvous session
================================================================================*/
boolean CWACPApp::RNDZConnect(char* pSourceID)
{
    boolean Result =  FALSE;
    ConnArgs* pConnArgs = NULL;

    if (pSourceID)
    {
        pConnArgs = pCLinkTable->GetEntry(pSourceID);
        if (pConnArgs)
        {
            const char* RNDZConnect="RNDZCONNECT";
            uint32 Size = (uint32)strlen(RNDZConnect); 

            /* Don't connect if already connected */
            if (!pConnArgs->bRNDZStatus)
            {
                if (pConnArgs->bSvrFlag == FALSE)
                {
                    /* Start the timeout check */
                    SetConnOptions(pSourceID, CONN_CFG_SpKEEPALIVECHECK, FALSE, 
                        DEFAULT_KEEPALIVE_INTERVAL, pConnArgs->Timeout);
#if defined(TRACE_SERVER_ON)
                    SYS_TRACE(TRACE_Type_SEND, "CWACPApp", "RNDZConnect", FmRENDEZVOUS, GnREQUEST, RNDZ_CRST_SpSTART_RNDZ, 
                        "%s: \nSending Request RNDZ_CRST_SpSTART_RNDZ", pConnArgs->SourceID);
#endif
                    Result = SendMsg(pSourceID, FmRENDEZVOUS, GnREQUEST, RNDZ_CRST_SpSTART_RNDZ, (uint8*)RNDZConnect, Size, FALSE);
                }
            }/* if (!pConnArgs->bConnected) */
            else
            {
#if defined(TRACE_SERVER_ON)
                SYS_TRACE(TRACE_Type_STATUS, "CWACPApp", "RNDZConnect", 0, 0, 0, 
                    "%s: \nAlready Connected", pConnArgs->SourceID);
#endif
                Result = TRUE;
            }

        }/*if (pConnArgs)*/

    }/* if (pSourceID) */

    return (Result);
}

/*================================================================================
CWACPApp::Open - Creates the communication pipe for the passed in arguments if 
    the LinkTable entry does not already exist.
================================================================================*/
boolean CWACPApp::Open(ConnArgs** ppConnArgs, uint8 NumConnections, boolean bSvrFlag)
{
    boolean Result = TRUE;
    uint8 i;
    ConnArgs* pTempArgs = NULL;

    if (ppConnArgs && NumConnections)
    {
        /* Check for existing connections that are open */
        for (i=0; i < NumConnections; i++)
        {
            pTempArgs = pCLinkTable->GetEntry(ppConnArgs[i]->SourceID);
            if (pTempArgs)
            {
                /* If we found an existing entry, then fail the Open call since the last
                 * connection hasn't been cleaned up
                 */
                Result = FALSE;
                break;
            }
        }

        /* Open the channels only if none of the new channels are open */
        if (Result)
        {
            for (i=0; i < NumConnections; i++)
            {
                /* If RNDZ is disabled, all communications will go through the RNDZ Port */
                if (ppConnArgs[i]->bRNDZDisabled == TRUE)
                {
                    ppConnArgs[i]->WACPPort = ppConnArgs[i]->RNDZPort;
                }

                /* Update Application CLinkTable. */
                Result  = pCLinkTable->AddEntry(ppConnArgs[i]);
                if (Result)
                {
                    /* Only update the ConObjSvr if this is a client connection */
                    ConObjSvr.AddConnection(ppConnArgs[i]->SourceID);
                }
            }

            if (Result && (bSvrFlag == FALSE))
            {
                /* Call Interchange to open the connections */
                Result = ConObjSvr.Connect(ppConnArgs, NumConnections);
            }
        }
    }
    else
    {
        Result = FALSE;
    }
    return(Result);
}

/*================================================================================
CWACPApp::Connect - Calls RNDZConnect to start the RNDZ authentication process.
================================================================================*/
boolean CWACPApp::Connect(char* pSourceID)
{
    boolean Result = FALSE;

    Result = RNDZConnect(pSourceID);

    return (Result);
}

/*================================================================================
CWACPApp::Close - Pass through function for closing the current connection 
    pointed to by pSourceID. 
================================================================================*/
boolean CWACPApp::Close(char* pSourceID)
{
    boolean Result = FALSE;

    if (bActive)
    {
		/* Turn off ByPass mode */
		SetByPass(pSourceID, NULL);

		/* Send out the Shutdown message and update Link Table */
        Result = ConObjSvr.Close(pSourceID);
    }
    return Result;
}

/*================================================================================
CWACPApp::SendMsg - Pass through function for sending a message.
================================================================================*/
boolean CWACPApp::SendMsg(char* pSourceID, uint16 Family, uint8 Genus, 
                          uint8 Species, uint8* pBuf, uint32 BufSize, 
                          boolean bEncryptionOn)
{
    boolean Result = FALSE;

    
    if (bActive)
    {
        /* Prevent multiple threads from sending at the same time */
        SendMutex.Lock(INFINITE);

        Result = FmRouter.SendMsg(pSourceID, Family, Genus, Species, pBuf, 
                                  BufSize, bEncryptionOn);
        SendMutex.Release();
    }
    return Result;
}

/*================================================================================
CWACPApp::SendTrap - Creates a FmTRAP message by serializing the passed in 
    arguments into a CErrEStd object byte stream and sends it out as a TRAP 
    message.
================================================================================*/
boolean CWACPApp::SendTrap(char* pSourceID, uint32 MsgClassID, uint32 ErrClassID, 
                           uint8 ErrorSpecies, uint16 ErrorCode,
                           uint8* pData, uint16 DataSize)
{
    boolean Result = FALSE;
    uint8*  pBuf = NULL;
    uint16  BufSize = 0;

    CErrEStd ErrObj;
    ErrObj.set_Species(ErrorSpecies);

    /* Set the ErrorCode */
    ErrObj.SetErrorCode(ErrorCode);

    /* Set the ClassID */
    ErrObj.SetErrClassID(ErrClassID);

    /* Set the ClassID */
    if (pData)
    {
        ErrObj.SetData(pData, DataSize);
    }

    /* Get the Size of the error object */
    BufSize = (uint16)ErrObj.size_all();

    /* Allocate the proper Buffer Size */
    pBuf = (uint8*) SYS_MALLOC(BufSize);
    if (pBuf)
    {
        /* Serialize the error object */
        ErrObj.serialize_all(pBuf);
        Result = TrapObjSvr.TxError(pSourceID, MsgClassID, ErrorSpecies, pBuf, BufSize);
    }

    SYS_FREE(pBuf);
    pBuf = NULL;

    return (Result);
}

/*================================================================================
CWACPApp::SendMsgProcTrap - Pass through function for sending a TRAP message
    from an error found while processing a WACP message. 
================================================================================*/
boolean CWACPApp::SendMsgProcTrap(char* pSourceID, uint32 ClassID, uint16 ErrorCode)
{
    boolean Result = FALSE;

    Result = SendTrap(pSourceID, ClassID, ClassID, CERR_SpERRMSGPROC, ErrorCode, 
             NULL, 0);

    return (Result);
}

/*================================================================================
CWACPApp::SetConnOptions - Interface for setting the Connection Object Server
    settings. 
================================================================================*/
boolean CWACPApp::SetConnOptions(char* pSourceID, CONN_CFG_SPECIES Options, 
                                 boolean bSyncFlag, time_t Interval, 
                                 time_t Timeout)
{
    boolean Result = FALSE;
    if (ConObjSvr.SetTxOptions(pSourceID, (uint8)Options, bSyncFlag))
    {
        if (ConObjSvr.SetKeepAliveInterval(pSourceID, Interval))
        {
            Result = ConObjSvr.SetTimeoutInterval(pSourceID, Timeout);
        }
    }
    return (Result);
}

/*================================================================================
CWACPApp::GetConnOptions - Interface for getting the Connection Object Server
    settings. 
================================================================================*/
boolean CWACPApp::GetConnOptions(char* pSourceID, CONN_CFG_SPECIES* pOptions, 
                                 time_t* pInterval, time_t* pTimeout)
{
    boolean Result = FALSE;

    if ((pSourceID) && (pOptions) && (pInterval) && (pTimeout))
    {
        *pOptions = (CONN_CFG_SPECIES)ConObjSvr.GetTxOptions(pSourceID);
        *pInterval = ConObjSvr.GetKeepAliveInterval(pSourceID);
        *pTimeout = ConObjSvr.GetTimeoutInterval(pSourceID);
        Result = TRUE;
    }

    return (Result);
}

/*================================================================================
CWACPApp::AppStart - Main function for starting up a CWACPApp application session.
    Calls AppInit to store the pointer to the MOMPBaseClass that will handle 
    all incoming messages.
================================================================================*/
boolean CWACPApp::AppStart()
{
    boolean Result = TRUE;

    bAppInit = TRUE;
    /* APP_INIT retrieves a void*  pointer to the calling process object 
    * and sets it to pCMOMPBaseClass
    */
    pCMOMPBaseClass = pCallBackObj->AppInitInt(this);

    /* Check for a valid object ptr */
    if (pCMOMPBaseClass)
        Result = TRUE;

    return(Result);
}

/*================================================================================
CWACPApp ProcessMsg: Processes a new message.
================================================================================*/
boolean CWACPApp::ProcessMsg(char* pSourceID, uint8* pBuf, uint32 BufSize)
{
    uint32 RetVal = 0;
    ConnArgs* pConnArgs;

    TOUCH(BufSize);

    /* Grab the connection information */
    pConnArgs = pCLinkTable->GetEntry(pSourceID);
    if (pConnArgs)
    {
        uint16  Family;

        /* Extract the Family */
        Family = GetFamily(pBuf);

        /* Turn off the keepalive messages when we receive a message */
        ConObjSvr.StopKeepAlive(pSourceID);

        /* Check for a Bypass message. All FmCONNECTION/FmRNDZ messages are
         * never sent via ByPass.
         */
        if ((pConnArgs->pByPassSrc == NULL) ||
            (Family == FmCONNECTION) || (Family == FmRENDEZVOUS))
        {
            FmRouter.RecvMsg(pSourceID, pBuf, BufSize);
        }
        else
        {
#if defined(TRACE_SERVER_ON)
            SYS_TRACE(TRACE_Type_STATUS, "CWACPApp", "ProcessMsg",
                    0, 0, 0, "Sending ByPass Message to %s", pConnArgs->pByPassSrc);
#endif

            /* Process the ByPass Message */
            pSendPacket(pLnkStruct->pSemName, pConnArgs->pByPassSrc, pBuf, BufSize);
        }
    }

    return RetVal ? TRUE: FALSE;
}

/*================================================================================
CWACPApp GetGUID: Returns the CRNDZDStd object pointer saved for the connection 
corresponding with the passed in pSourceID.
================================================================================*/
CRNDZDStd* CWACPApp::GetGUID(char* pSourceID)
{
    CRNDZDStd* pGUIDStd = NULL;
    ConnArgs* pConnArgs;

    if (pSourceID)
    {
        /* Grab the connection information */
        pConnArgs = pCLinkTable->GetEntry(pSourceID);
        if (pConnArgs)
        {
            pGUIDStd = (CRNDZDStd*)pConnArgs->pGUIDStd;
        }
    }
    return (pGUIDStd);
}

/*================================================================================
CWACPApp GetStackType: Returns the stack type for the connection corresponding 
    with the passed in pSourceID.
================================================================================*/
STACK_TYPE CWACPApp::GetStackType(char* pSourceID)
{
    STACK_TYPE StackType = STACK_UNDEFINED;
    ConnArgs* pConnArgs;

    if (pSourceID)
    {
        /* Grab the connection information */
        pConnArgs = pCLinkTable->GetEntry(pSourceID);
        if (pConnArgs)
        {
            StackType = pConnArgs->StackType;
        }
    }
    return (StackType);
}

/*================================================================================
CWACPApp SetByPass: WACP API for setting up a bypass channel to allow a 
    connection between two connected SourceIDs. Setting pByPassID to NULL will 
    disable the bypass channel.
================================================================================*/
boolean CWACPApp::SetByPass(char* pSourceID, char* pByPassID)
{
    ConnArgs*   pConnArgs = NULL;
    boolean     Result = FALSE;

    /* Look up the linktable for the SourceID. */
    pConnArgs = pCLinkTable->GetEntry(pSourceID);
    if (pConnArgs)
    {
        ConnArgs*   pByConnArgs = NULL;
    
        if (pByPassID)
        {
            /* Look up the linktable for the SourceID. */
            pByConnArgs = pCLinkTable->GetEntry(pByPassID);
            if (pByConnArgs)
            {
#if defined(TRACE_SERVER_ON)
                SYS_TRACE(TRACE_Type_STATUS, "CWACPApp", "ProcessMsg",
                        0, 0, 0, "Setting SourceID %s to bypass SourceID %s\n", pSourceID, pByPassID);
#endif
                /* Set the current connection to bypass to the new SourceID */
                pConnArgs->pByPassSrc = (char*)SYS_MALLOC(ADDR_BUF_SIZE);
                if (pConnArgs->pByPassSrc)
                {
                    memmove(pConnArgs->pByPassSrc, pByConnArgs->SourceID, ADDR_BUF_SIZE);
                }

                /* Set the SourceID of the bypass to the current connection */
                pByConnArgs->pByPassSrc = (char*)SYS_MALLOC(ADDR_BUF_SIZE);
                if (pByConnArgs->pByPassSrc)
                {
                    memmove(pByConnArgs->pByPassSrc, pConnArgs->SourceID, ADDR_BUF_SIZE);
                }
             
                Result = TRUE;
            }
            else
            {
#if defined(TRACE_SERVER_ON)
                SYS_TRACE(TRACE_Type_STATUS, "CWACPApp", "ProcessMsg",
                        0, 0, 0, "Invalid SourceID for Bypass");
#endif

            }
        }
        else
        {
            /* Look up the linktable for the SourceID. */
            pByConnArgs = pCLinkTable->GetEntry((char*)pConnArgs->pByPassSrc);
            if (pByConnArgs)
            {
                /* Free up the ByPassSrc */
                SYS_FREE(pByConnArgs->pByPassSrc);

                /* Reset the ByPassSrc */
                pByConnArgs->pByPassSrc = NULL;

                Result = TRUE;
            }

            /* Free up the ByPassSrc */
            SYS_FREE(pConnArgs->pByPassSrc);

            /* Reset the ByPassSrc */
            pConnArgs->pByPassSrc = NULL;
        }
    }

    return Result;
}

/* End of File */


