/*-D************************************************************************* 
'*   This program source file is confidential and proprietary property of
'*    Welch Allyn, Inc. Unauthorized use or reproduction is prohibited.
'****************************************************************************
'****************************************************************************
'**
'****************************************************************************
'**  ////////////////////////////////////////////////////////////////////////
'**  '// CSerialCommSvr.cpp:                                               //
'**  '// Language       :   C++                                            //
'**  '// Application    :   WACP                                           //
'**  '// Sources        :   MOMP Standard Library                          //
'**  '///////////////////////////////////////////////////////////////////////
'*-E************************************************************************/
#include <assert.h>
#include "CSerialCommSvr.h"
#include "CCommSvr.h"
#include "SysUtil.h"

/*********************************************************************************
  Default constructor
*********************************************************************************/
CSerialCommSvr::CSerialCommSvr()
{
    // Initial the member vars
    m_hPort = NULL;
    m_bDCB = false;
    m_bCTOS = false;
    m_bSetConfig = false;
    m_bAutoDiscovery = false;
    m_NumOverlappedReads = 0;
    m_NumOverlappedWrites = 0;
    memset(m_CommPort, 0, sizeof(m_CommPort));
    InitializeCriticalSection(&m_CS);

    this->StackType = SERIAL;
}

/*-D*************************************************************************
*
*  CSerialCommSvr destructor
*
*  PURPOSE: Provide ordered shutdown and thorough cleanup of transport instance
*
*  ARGUMENTS: none
*
*  RETURNS: none
*
**-E************************************************************************/
CSerialCommSvr::~CSerialCommSvr()
{
    //Let outstanding writes complete
    while (m_NumOverlappedWrites)
    {
        Sleep(0);
    }
    if (m_hPort)
    {
        //Restore previous settings
        if (m_bDCB)
        {
            SetCommState(m_hPort, &m_DCB);
        }
        if (m_bCTOS)
        {
            SetCommTimeouts(m_hPort, &m_CTOS);
        }
        //Close the communication Port
        CloseHandle(m_hPort);
        m_hPort = NULL;
    }
    //Wait for any outstanding reads to complete:
    // They will acceptably fail after closing the handle with Error 995:
    // The I/O operation has been aborted because of either a thread exit or an application request.
    while (m_NumOverlappedReads)
    {
        Sleep(0);
    }
    DeleteCriticalSection(&m_CS);
}

/*-D*************************************************************************
*
*  CSerialCommSvr::SetConfig
*
*  PURPOSE: Updates addressing information
*
*  ARGUMENTS:
*      [IN] pAddr = A SerialAddr structure with extra information.
*           Client instances may specify NULL for this parameter.
*
*  RETURNS: none
*
*  NOTES:  As this function cannot return a value, an error is stored in
*          the instance and later checked in Connect function.

*          The format of the SourceID is: "COMn[:BBBB,P,D,S]"
*          COMn = Comm port (must be a string, not a number)
*          =====  Following values individually optional =====
*          BBBB = Baud rate - a numeric value (default DEFAULT_BAUD_RATE)
*          P    = Parity - 'N'|'E'|'M'|'O'|'S', (default no parity)
*          D    = Data bits (default 8)
*          S    = Stop bits (default 1)
**-E************************************************************************/
void CSerialCommSvr::SetConfig(void* pAddr)
{
    SerialAddr* pConfig = (SerialAddr*)pAddr;
    char ConnectionString[ADDR_BUF_SIZE];
    char* pTmp;

    //Initialize default parameters
    m_DCB2.BaudRate = DEFAULT_BAUD_RATE;
    m_DCB2.ByteSize = 8;
    m_DCB2.Parity = NOPARITY;
    m_DCB2.StopBits = ONESTOPBIT;

    if (this->SourceID)
    {
        strcpy(ConnectionString, this->SourceID);
        pTmp = strtok(ConnectionString, ":"); //get first delim (comm port)
        if (pTmp)
        {
            strcpy(m_CommPort, pTmp);
            //Get Baud Rate
            pTmp = strtok(NULL, ",");
            if (pTmp)
            {
                m_DCB2.BaudRate = atoi(pTmp);
                if (0 == m_DCB2.BaudRate)
                {
                    m_bSetConfig = false;
                    return;
                }
                //Get Parity
                pTmp = strtok(NULL, ",");
                if (pTmp)
                {
                    if (0 == _stricmp(pTmp, "N"))
                    {
                        m_DCB2.Parity = NOPARITY;
                    }
                    else if (0 == _stricmp(pTmp, "E"))
                    {
                        m_DCB2.Parity = EVENPARITY;
                    }
                    else if (0 == _stricmp(pTmp, "M"))
                    {
                        m_DCB2.Parity = MARKPARITY;
                    }
                    else if (0 == _stricmp(pTmp, "O"))
                    {
                        m_DCB2.Parity = ODDPARITY;
                    }
                    else if (0 == _stricmp(pTmp, "S"))
                    {
                        m_DCB2.Parity = SPACEPARITY;
                    }
                    else
                    {
                        m_bSetConfig = false;
                        return;
                    }
                    //Get Data bits
                    pTmp = strtok(NULL, ",");
                    if (pTmp)
                    {
                        m_DCB2.ByteSize = (BYTE)atoi(pTmp);
                        if (0 == m_DCB2.ByteSize)
                        {
                            m_bSetConfig = false;
                            return;
                        }
                        //Get Stop bits
                        pTmp = strtok(NULL, ",");
                        if (pTmp)
                        {
                            if (0 == strcmp(pTmp, "1"))
                            {
                                m_DCB2.StopBits = ONESTOPBIT;
                            }
                            else if (0 == strcmp(pTmp, "1.5"))
                            {
                                m_DCB2.StopBits = ONE5STOPBITS;
                            }
                            else if (0 == strcmp(pTmp, "2"))
                            {
                                m_DCB2.StopBits = TWOSTOPBITS;
                            }
                            else
                            {
                                m_bSetConfig = false;
                                return;
                            }
                        }
                    }
                }
            }
        }
    }
    if (pConfig)
    {
        m_bAutoDiscovery = pConfig->bAutoDiscovery ? true : false;
    }
    m_bSetConfig = true;
}

/*-D*************************************************************************
*
*  CSerialCommSvr::SendMsg
*
*  PURPOSE: Sends data from the current transport instance to the remote end
*
*  ARGUMENTS:
*      [IN] pBuf = Pointer to a buffer that contains the data to be sent.
*      [IN] BufSize = Size of the buffer pointed to by pBuf.
*
*  RETURNS: TRUE if successful, otherwise FALSE.
*
**-E************************************************************************/
boolean CSerialCommSvr::SendMsg(uint8* pBuf, uint32 BufSize)
{
    BOOL bResult = FALSE;
    DWORD dwBytes;
    PSERIAL_PACKET pPacket = NULL;

    if (NULL == pBuf) //validate params
    {
        return FALSE;
    }
    pPacket = (PSERIAL_PACKET)HeapAlloc(g_hHeap, HEAP_ZERO_MEMORY, sizeof(SERIAL_PACKET) - SERIAL_RCVBUFSIZE + BufSize);
    if (NULL == pPacket)
    {
        return FALSE;
    }
    memcpy(pPacket->Buffer, pBuf, BufSize);
    pPacket->State = OVERLAPPED_WRITE;
    InterlockedIncrement(&m_NumOverlappedWrites);
    bResult = WriteFile(m_hPort, pPacket->Buffer, BufSize, &dwBytes, &pPacket->Ol);
    if (!bResult)
    {
        switch (GetLastError())
        {
            case ERROR_IO_PENDING:
            {
                bResult = TRUE; //reset return boolean (this is expected).
                break;
            }
            default:
            {
                HeapFree(g_hHeap, 0, pPacket);
                InterlockedDecrement(&m_NumOverlappedWrites);
                break;
            }
        }
    }
    return (boolean)bResult;
}

/*-D*************************************************************************
*
*  CSerialCommSvr::Connect
*
*  PURPOSE: Opens a serial communications port based on the pSourceID.
*           Updates the WACP Port and SourceID once a connection is created.
*
*  ARGUMENTS:
*      [IN] pSourceID = String identifier(COM port) of the connection.
*      [IN] PortNum = Port value for the connection.
*
*  RETURNS: TRUE if successful, otherwise FALSE.
*
**-E************************************************************************/
boolean CSerialCommSvr::Connect(uint8* pSourceID, uint16 PortNum)
{
#if defined(_UNICODE) || defined(UNICODE)
    #define FORMAT_SPEC "%S"
#else
    #define FORMAT_SPEC "%s"
#endif
    PSERIAL_PACKET pPacket = NULL;
    TCHAR ComPort[MAX_PATH];

    boolean Result = FALSE;

    DCB PortDCB; //Port Control Settings Structure
    COMMTIMEOUTS CommTimeouts; //Comm Port timeout structure

    //Do not allow reentrancy
    if (m_hPort || !m_bSetConfig)
    {
        return FALSE;
    }
    // Per MSDN, prepend COM port (required for COM ports above 9)
    ComPort[0] = _T('\0'); //NULL prefix
    if (m_CommPort[0] != '\\')
    {
        _tcscpy(ComPort, _T("\\\\.\\"));
    }
    _stprintf(ComPort+_tcslen(ComPort), _T(FORMAT_SPEC), m_CommPort);
    // Open COM port
    m_hPort = CreateFile(ComPort, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH | FILE_FLAG_OVERLAPPED, NULL);
    if (INVALID_HANDLE_VALUE == m_hPort)
    {
        m_hPort = NULL;
        return FALSE;
    }
    if (!GetCommState(m_hPort, &PortDCB))
    {
        CloseHandle(m_hPort);
        m_hPort = NULL;
        return FALSE;
    }
    //backup initial settings
    memcpy(&m_DCB, &PortDCB, sizeof(DCB));
    m_bDCB = true;
    // Change the DCB structure settings.
    PortDCB.BaudRate = m_DCB2.BaudRate;         // Current baud
    PortDCB.fBinary = TRUE;                     // Binary mode; no EOF check
    PortDCB.fParity = TRUE;                     // Enable parity checking
    PortDCB.fOutxCtsFlow = FALSE;               // CTS Output flow control
    PortDCB.fOutxDsrFlow = FALSE;               // DSR Output flow control
    PortDCB.fDtrControl = DTR_CONTROL_DISABLE;  // DTR flow control type
    PortDCB.fDsrSensitivity = FALSE;            // DSR sensitivity
    PortDCB.fTXContinueOnXoff = FALSE;          // XOFF continues Tx
    PortDCB.fOutX = FALSE;                      // No XON/XOFF out flow control
    PortDCB.fInX = FALSE;                       // No XON/XOFF in flow control
    PortDCB.fErrorChar = FALSE;                 // Disable error replacement
    PortDCB.fNull = FALSE;                      // Disable null stripping
    PortDCB.fRtsControl = RTS_CONTROL_DISABLE;  // RTS flow control
    PortDCB.fAbortOnError = FALSE;              // Do not abort reads/writes on error
    PortDCB.ByteSize = m_DCB2.ByteSize;         // Number of bits/byte, 4-8
    PortDCB.Parity = m_DCB2.Parity;             // 0-4=no,odd,even,mark,space
    PortDCB.StopBits = m_DCB2.StopBits;         // 0,1,2 = 1, 1.5, 2

    // Device Communications Timeout parameters
    if (!GetCommTimeouts(m_hPort, &CommTimeouts))
    {
        CloseHandle(m_hPort);
        m_hPort = NULL;
        return FALSE;
    }
    //backup initial settings
    memcpy(&m_CTOS, &CommTimeouts, sizeof(COMMTIMEOUTS));
    m_bCTOS = true;

    // Change the COMMTIMEOUTS structure settings
    CommTimeouts.ReadIntervalTimeout = 10;
    CommTimeouts.ReadTotalTimeoutMultiplier = 0;
    CommTimeouts.ReadTotalTimeoutConstant = 0;
    CommTimeouts.WriteTotalTimeoutMultiplier = 0;
    CommTimeouts.WriteTotalTimeoutConstant = 0;
    if (!SetCommTimeouts(m_hPort, &CommTimeouts))
    {
        CloseHandle(m_hPort);
        m_hPort = NULL;
        return FALSE;
    }
    PurgeComm(m_hPort, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR); //ignore return code

    if (!SetCommState(m_hPort, &PortDCB))
    {
        CloseHandle(m_hPort);
        m_hPort = NULL;
        return FALSE;
    }

    // assert DTR and RTS
    EscapeCommFunction(m_hPort, SETDTR); //ignore return code
    EscapeCommFunction(m_hPort, SETRTS); //ignore return code

    if (ERROR_SUCCESS != this->Assign(m_hPort, (ULONG_PTR)this))
    {
        CloseHandle(m_hPort);
        m_hPort = NULL;
        return FALSE;
    }
    if (m_bAutoDiscovery)
    {
        // Specify a set of events to be monitored for the Port
        if (!SetCommMask(m_hPort, EV_DSR))
        {
            CloseHandle(m_hPort);
            m_hPort = NULL;
            return FALSE;
        }
        pPacket = (PSERIAL_PACKET)HeapAlloc(g_hHeap, HEAP_ZERO_MEMORY, sizeof(SERIAL_PACKET) - SERIAL_RCVBUFSIZE);
        if (NULL == pPacket)
        {
            CloseHandle(m_hPort);
            m_hPort = NULL;
            return FALSE;
        }
        pPacket->State = OVERLAPPED_INIT;
        InterlockedIncrement(&m_NumOverlappedReads);
        if (!WaitCommEvent(m_hPort, &pPacket->EventMask, &pPacket->Ol))
        {
            switch (GetLastError())
            {
                case ERROR_IO_PENDING:
                {
                    break;
                }
                default:
                {
                    HeapFree(g_hHeap, 0, pPacket);
                    InterlockedDecrement(&m_NumOverlappedReads);
                    CloseHandle(m_hPort);
                    m_hPort = NULL;
                    return FALSE;
                    break;
                }
            }
        }
    }
    if (ERROR_SUCCESS != PostReceive(NULL, SERIAL_MAX_OVERLAPPED_READS))
    {
        CloseHandle(m_hPort);
        m_hPort = NULL;
        return FALSE;
    }
    Result = TRUE;

    return Result;
    UNREFERENCED_PARAMETER(PortNum);
    UNREFERENCED_PARAMETER(pSourceID);
}

/*-D*************************************************************************
*
*  CSerialCommSvr::ProcessCP
*
*  PURPOSE: Callback to process queued I/O from completion port
*
*  ARGUMENTS:
*      [IN] dwBytes = number of bytes processed from the overlapped handle.
*      [IN] lpOl = Pointer to overlapped plus structure for network data.
*
*  RETURNS: none
*
**-E************************************************************************/
void CSerialCommSvr::ProcessCP(DWORD dwBytes, LPOVERLAPPED lpOl)
{
    PSERIAL_PACKET pPacket = (PSERIAL_PACKET)lpOl;
    DWORD dwModemStat;

    if (pPacket)
    {
        switch (pPacket->State)
        {
            case OVERLAPPED_INIT: //change in
            {
                if (m_bAutoDiscovery && GetCommModemStatus(m_hPort, &dwModemStat))
                {
                    if (!(dwModemStat & MS_DSR_ON)) //DTR low state (device unplugged)
                    {
                        this->Removal();
                    }
                }
                HeapFree(g_hHeap, 0, pPacket);
                InterlockedDecrement(&m_NumOverlappedReads);
                break;
            }
            case OVERLAPPED_WRITE:
            {
                HeapFree(g_hHeap, 0, pPacket);
                InterlockedDecrement(&m_NumOverlappedWrites);
                break;
            }
            case OVERLAPPED_READ:
            {
                EnterCriticalSection(&m_CS);
                this->WACPParserEx((uint8*)pPacket->Buffer, dwBytes, NULL);
                LeaveCriticalSection(&m_CS);
                PostReceive(pPacket);
                InterlockedDecrement(&m_NumOverlappedReads);
                break;
            }
            default:
            {
                assert(0);
                HeapFree(g_hHeap, 0, pPacket);
                break;
            }
        }
    }
}

/*-D*************************************************************************
*
*  CSerialCommSvr::RecvCPError
*
*  PURPOSE: Receive error callback from I/O completion port
*
*  ARGUMENTS:
*      [IN] LastError = standard WIN32 error of reason for overlapped failure.
*      [IN] lpOl = Pointer to overlapped plus structure for network data.
*
*  RETURNS: none
*
**-E************************************************************************/
void CSerialCommSvr::RecvCPError(DWORD LastError, LPOVERLAPPED lpOl)
{
    PSERIAL_PACKET pPacket = (PSERIAL_PACKET)lpOl;
    int state;

    if (pPacket)
    {
        state = pPacket->State;
        HeapFree(g_hHeap, 0, pPacket);
        switch (state)
        {
            case OVERLAPPED_WRITE:
            {
                InterlockedDecrement(&m_NumOverlappedWrites);
                break;
            }
            case OVERLAPPED_READ:
            case OVERLAPPED_INIT:
            {
                InterlockedDecrement(&m_NumOverlappedReads);
                break;
            }
            default:
            {
                assert(0);
                break;
            }
        }
    }
    return;
    UNREFERENCED_PARAMETER(LastError);
}

/*-D*************************************************************************
*
*  CSerialCommSvr::PostReceive
*
*  PURPOSE: Helper function for posting receives for network client that will
*           complete when incoming data is available.
*
*  ARGUMENTS:
*      [IN] pPacket = Pointer to overlapped plus structure for network data.
*      [IN] Count = number of receives to post.
*
*  RETURNS: ERROR_SUCCESS, or standard Win32 API error.
*
**-E************************************************************************/
DWORD CSerialCommSvr::PostReceive(PSERIAL_PACKET pPacket, UINT Count)
{
    DWORD Result = ERROR_SUCCESS;
    PSERIAL_PACKET pTempPacket = pPacket;
    DWORD dwBytes;

    if (pPacket && Count > 1)
    {
        return ERROR_INVALID_FUNCTION;
    }
    for (UINT i=0; i<Count; i++)
    {
        if (NULL == pTempPacket)
        {
            pPacket = (PSERIAL_PACKET)HeapAlloc(g_hHeap, HEAP_ZERO_MEMORY, sizeof(SERIAL_PACKET));
            if (NULL == pPacket)
            {
                Result = ERROR_OUTOFMEMORY;
                break;
            }
        }
        pPacket->State = OVERLAPPED_READ;
        InterlockedIncrement(&m_NumOverlappedReads);
        if (!ReadFile(m_hPort, pPacket->Buffer, SERIAL_RCVBUFSIZE, &dwBytes, &pPacket->Ol))
        {
            switch (Result = GetLastError())
            {
                case ERROR_IO_PENDING:
                {
                    Result = ERROR_SUCCESS;
                    break;
                }
                default:
                {
#if defined(TRACE_SERVER_ON)
                    SYS_PRINTF("ReadFile Error: %u\n", Result);
#endif
                    HeapFree(g_hHeap, 0, pPacket);
                    InterlockedDecrement(&m_NumOverlappedReads);
                    break;
                }
            }
        }
    }
    pPacket = pTempPacket;

    return Result;
}
