/*-D************************************************************************* 
'*   This program source file is confidential and proprietary property of
'*    Welch Allyn, Inc. Unauthorized use or reproduction is prohibited.
'****************************************************************************
'****************************************************************************
'**
'****************************************************************************
'**  ////////////////////////////////////////////////////////////////////////
'**  '// CUSBCommSvr.cpp:                                                  //
'**  '// Language       :   C++                                            //
'**  '// Application    :   WACP                                           //
'**  '// Sources        :   MOMP Standard Library                          //
'**  '///////////////////////////////////////////////////////////////////////
'*-E************************************************************************/

#include "CUSBCommSvr.h"
#include "CCommSvr.h"
#include "CWACPEvent.h"
#include "Mem.h"
#include "SysUtil.h"

#define USB_THREADPRIORITY THREAD_PRIORITY_NORMAL

/* WAUSBDLL requires 0 to be passed back on success for callback functions */
#if !defined(SUCCESS)
    #define SUCCESS 0
#endif

/*********************************************************************************
  Default constructor
*********************************************************************************/
CUSBCommSvr::CUSBCommSvr()
                  
{
    /* Initialize the lock mutexs for destroying the CommSvr */
    pClientExitMutex = new CWACPMutex(NULL);
    pDeviceMutex = new CWACPMutex(NULL);
    pDisconnectMutex = new CWACPMutex(NULL);
    pThreadEvent = new CWACPEvent(NULL, TRUE);
    
    /* Initialize the semaphore for handling data requests */
    pReadSem = new CWACPSemaphore(NULL, 0, MAX_SEM_COUNT);

    /* Creation of this class instance means the USB connection has been established.
     * Set the active state to TRUE
     */
    bActive = TRUE;

    INEndPointNum = 0;
    OUTEndPoint = 0;

    /* Initialize the lock mutexs for accessing the cache */
    pCacheMutex = new CWACPMutex(NULL);

    /* Create the queue managers for the IN/OUT and WACP queues */
    pCacheQMgr = new uint8QManager(Cache, USB_CACHE_SIZE);

    /* Initialize the state vars */
    bConnected = FALSE;
    bServer = FALSE;

    StackType = USB;
}

/*********************************************************************************
Default Destructor. Closes and resets the USB driver handles. Waits on the 
    pThreadEvent, pDeviceMutex, and ClientExitMutex to insure we do not destroy 
    the object or memory allocated for the driver before the StartClient
    thread has terminated. 
*********************************************************************************/
CUSBCommSvr::~CUSBCommSvr()
{   
    /* Wait until the Disconnect mutex can be locked or grab the lock to */
    if (pDisconnectMutex)
    {
        pDisconnectMutex->Lock(INFINITE);
    }

    /* Reset the bActive flag to kill the StartClient Thread */
    bActive = FALSE;

    /* Wait until the StartClient Thread has started. */
    if (pThreadEvent != NULL)
    {
        pThreadEvent->Wait(INFINITE);
        pThreadEvent->Reset();
        delete(pThreadEvent);
        pThreadEvent = NULL;
    }

    /* Check for a valid Read semaphore */
    if (pReadSem != NULL)
    {
        if (bConnected)
        {
            /* Disconnect the current open connection */
            WA_USB_Disconnect(SourceID); 

            /* Wait for the DeviceMutex to insure we are done with the Addr.pDevice memory */
            if (pDeviceMutex)
            {
                pDeviceMutex->Lock(INFINITE);
                pDeviceMutex->Release();
            }

            SYS_FREE(Addr.pDevice);
            Addr.pDevice = NULL;
        }

        /* Signal the reader thread to stop execution */
        pReadSem->Raise();
    }

    if (pDisconnectMutex != NULL)
    {
        pDisconnectMutex->Release();
    }

    if (pClientExitMutex != NULL)
    {
        /* Wait for the CUSBCommSvr threads to complete before completing the destructor */
        pClientExitMutex->Lock(INFINITE);
        pClientExitMutex->Release();

        /* Destroy the Mutexes */
        delete(pClientExitMutex);
        pClientExitMutex = NULL;
    }

    if (pDeviceMutex)
    {
        delete(pDeviceMutex);
        pDeviceMutex = NULL;
    }


    if (pDisconnectMutex != NULL)
    {
        /* Destroy the Mutexes */
        delete(pDisconnectMutex);
        pDisconnectMutex = NULL;
    }

    if (pReadSem != NULL)
    {
        /* Destroy the Semaphore */
        delete(pReadSem);
        pReadSem = NULL;
    }
    
    if (pCacheMutex != NULL)
    {
        delete(pCacheMutex);
        pCacheMutex = NULL;
    }

    if (pCacheQMgr != NULL)
    {
        delete(pCacheQMgr);
        pCacheQMgr = NULL;
    }

    if (bServer)
    {
        /* Close the USB Session. */
        WA_USB_CloseSession();
    }
}

/*********************************************************************************
CUSBCommSvr::SetConfig: Updates the addressing information of the USB
    connection by grabbing the first OUT Endpoint and collecting the number of 
    IN endpoints.

Parameters:
    pAddr - Pointer to a USBAddr structure that contains the DEVICE_DESC information
            for the connection.
*********************************************************************************/
void CUSBCommSvr::SetConfig(void* pAddr)
{
    USBAddr* pConfig = (USBAddr*)pAddr;
    int32 i;

    if (pConfig != NULL)
    {
        Addr.pDevice = pConfig->pDevice;

        /* Count the number of endpoints */
        for (i = Addr.pDevice->NumEndPoints - 1; i >= 0 ;i--)
        {
            if (Addr.pDevice->EndPoints[i].Direction == ENDPOINT_IN)
            {
                INEndPointNum++;
            }
        }
 
        /* Look for the first OUT endpoint */
        for (i = 0; i < Addr.pDevice->NumEndPoints;i++)
        {
            if (Addr.pDevice->EndPoints[i].Direction == ENDPOINT_OUT)
            {
                OUTEndPoint = Addr.pDevice->EndPoints[i].EndPointID;
                break;
            }
        }
    }
}

/*********************************************************************************
CUSBCommSvr::Connect: Calls WA_USB_Connect to start the connection. If successful,
    the USB Receive Timeout is set and the pThreadEvent is created. The SourceID 
    and the Port value is saved into the object.
    
Parameters:
    pSourceID - String identifier(USB_XXXXXX) of the connection 

    PortNum - Port value initially set to RNDZ port.

Returns: 
    TRUE, if WA_USB_Connect returns SUCCESS
    FALSE, if WA_USB_Connect returns FAILURE

*********************************************************************************/
boolean CUSBCommSvr::Connect(uint8* pSourceID, uint16 PortNum)
{
    boolean Result = FALSE;
    int32  RetVal;
    WAUSBCONFIG ConfigData;

    /* Set the SourceID */
    SetSourceID((char*)pSourceID);

    /* Set the Port Number */
    Port = PortNum;

    RetVal = WA_USB_Connect((char*)pSourceID, (RECVDATAPROC)this->RecvData, this);
    if (RetVal == 0)
    {
        bConnected = TRUE;

        ConfigData.ReceiveTimeOut = USB_RECEIVE_TIMEOUT;

        WA_USB_Configure((char*)pSourceID, &ConfigData);
        
        /* Start up the Reader Thread */
		Result = this->StartReader(StartClient, this, USB_THREADPRIORITY, (char*)pSourceID);
    }
    else
    {
        pThreadEvent->Raise();
    }

    return Result;
}

/*********************************************************************************
CUSBCommSvr::SendMsg: Calls WA_USB_Send to send a message out the OutEndPoint.
    
Parameters:
    pBuf - Pointer to a buffer that contains the data to be sent.

    BufSize - Size of the buffer pointed to by pBuf.

Returns: 
    TRUE, if WA_USB_Send returns SUCCESS
    FALSE, if WA_USB_Send returns FAILURE

*********************************************************************************/
boolean CUSBCommSvr::SendMsg(uint8* pBuf, uint32 BufSize)
{
    uint32 Result = 0;

    if (pBuf == NULL)
    {
        return FALSE;
    }

    Result = WA_USB_Send(SourceID, OUTEndPoint, pBuf, BufSize);

#if defined(DEBUG_PRINT)
    if (!Result)
    {
        for (uint32 i=0; i < BufSize;i++)
        {
            if ( i % 25 == 0)
                SYS_PRINTF("\n");
            SYS_PRINTF("%.2x ", (uint8)pBuf[i]);
        }
        SYS_PRINTF("\n\n");
    }
    else
    {
        SYS_PRINTF("\n\nERROR!!! Unable to send USB message\n");
    }
#endif  

    return (Result ? FALSE : TRUE);
}   

/*********************************************************************************
CUSBCommSvr::ReadData: Reads up to BytesToRead number of bytes of data from 
    the USB Cache. This will block until any data is ready or the connection is 
    closed.

Parameters:
    pBuf - Pointer to a buffer that will receive the data.

    BytesToRead - Number of bytes to read.

    pBytesRead - Pointer to a uint32 used to pass back the number of bytes read.

Returns: 
    TRUE, if any data was read from the cache.
    FALSE, if the cache is invalid, pBuf is invalid, or the connection is closed

*********************************************************************************/
boolean CUSBCommSvr::ReadData(char* pBuf, uint32 BytesToRead, uint32* pBytesRead)
{
    boolean Result = FALSE;
    uint32  DataSize = 0;
    if (pCacheQMgr && pBuf)
    {
        DataSize = pCacheQMgr->get_num_used();

#if defined(DEBUG_PRINT)
        SYS_PRINTF("CUSBCommSvr::ReadData DataSize=%d, BytesToRead=%d\n", DataSize,BytesToRead);
#endif
        /* Block until we get some data */
        if (DataSize == 0)
        {
            if (pReadSem->Wait(INFINITE) == WAIT_SUCCESS)
            {
                DataSize = pCacheQMgr->get_num_used();
            }
        }

        if (DataSize > 0)
        {
            /* Only read the requested number of bytes */
            if (DataSize > BytesToRead)
            {
                DataSize = BytesToRead;
            }

            pCacheMutex->Lock(INFINITE);
            Result = pCacheQMgr->dequeue_array((uint8*)pBuf, DataSize);
            pCacheMutex->Release();

            /* Update the number of bytes read */
            *pBytesRead = DataSize;

#if defined(DEBUG_PRINT)
            SYS_PRINTF("CUSBCommSvr::ReadData Read %d bytes of %d, Result = %d\n", BytesToRead, DataSize, Result);
#endif
        }
        else
        {
            *pBytesRead = 0;
#if defined(DEBUG_PRINT)
            SYS_PRINTF("CUSBCommSvr::Timeout ReadData Not enough data to read %d bytes of %d\n", BytesToRead, DataSize);
#endif
        }
    }
    return Result;
}

/*********************************************************************************
CUSBCommSvr::RecvData: Callback function for receiving data from the WAUSBDLL. 
    The data is copied into a Cache and another read request is queued up.

Parameters:
    Status - Contains the status of the receive operation. 
        ERROR_SUCCESS: The data was received successfully.
        ERROR_SEM_TIMEOUT: The requested timed out before completing.
        ERROR_OPERATION_ABORTED: The request was in progress when it was canceled 
            by the driver. This is usually the result of the application issuing a reset.
        ERROR_GEN_FAILURE: The device has failed the operation with an error.

    pDevName - Name of the device that received the data

    EndPointID - ID of the endpoint from which the data was received.

    pRecvData - Address of the buffer that contains the received data. This buffer
        was passed in the call to WA_USB_RecvAsync(). 

    RecvLength - Length of the valid data returned in the pRecvData buffer.

    pUserRecvData - User data that was registered by the call to WA_USB_Connect(). 

Returns: Always returns SUCCESS(0).
*********************************************************************************/
uint32 WAUSB_CALLBACK CUSBCommSvr::RecvData(uint32 Status, char* pDevName, int32 EndPointID, 
                           void* pRecvData, int32 RecvLength, void* pUserRecvData)
{
    CUSBCommSvr* pCommSvr = (CUSBCommSvr*)pUserRecvData;
    int     Result = FALSE;
	uint8* pBuffer = (uint8*)pRecvData;


	for (uint32 i=0; i < RecvLength;i++)
    {
            if ( i % 25 == 0)
                SYS_PRINTF("\n");
            SYS_PRINTF("%.2x ", pBuffer[i]);
    }

//#if defined(DEBUG_PRINT)
    SYS_PRINTF("\nCUSBCommSvr RecvData: pCommSvr %p, EndpointID %d, status = 0x%X, pDevName = %s\n", 
                pCommSvr, EndPointID, Status, pDevName);
//#endif

    TOUCH(pDevName);
    TOUCH(EndPointID);

    /* Check to see if the session is still active and that the status is SUCCESS or a Timeout.
     * If status is a TIMEOUT, then requeue the read request.
     */
    if((pCommSvr->bActive == TRUE) && ((Status == ERROR_SUCCESS) || (Status == ERROR_SEM_TIMEOUT)))
    {
        if ((Status == ERROR_SUCCESS) && 
            ((RecvLength < RCVBUFSIZE) && RecvLength))
        {      
            pCommSvr->pCacheMutex->Lock(INFINITE);

            /* Add the message to the CacheQMgr*/
            Result = pCommSvr->pCacheQMgr->enqueue_array((uint8*)pRecvData, RecvLength);

            pCommSvr->pCacheMutex->Release();

#if defined(DEBUG_PRINT)
            SYS_PRINTF("\nCUSBCommSvr RecvData: EndPointID %d, Adding %d bytes to the cache\n", 
                        EndPointID, RecvLength);
#endif
            if (Result)
            {
                /* Release the Read semaphore to request the next packet */
                pCommSvr->pReadSem->Raise();
            }
            else
            {
#if defined(DEBUG_PRINT_STATUS)
                SYS_PRINTF("\nCUSBCommSvr RecvData: EndPointID %d, Cache Write failed for %d, Free bytes %d\n", 
                    EndPointID, RecvLength, pCommSvr->pCacheQMgr->get_num_free());
#endif
            }

        }/* if ((RecvLength < RCVBUFSIZE) && RecvLength) */

        if(pCommSvr->bActive == TRUE)
        {
            /* Queue up another request */
            Result = WA_USB_RecvAsync(pCommSvr->SourceID, EndPointID, (uint8*)pRecvData, RCVBUFSIZE);
        }

#if defined(DEBUG_PRINT)
        SYS_PRINTF("CUSBCommSvr %p EndPointID %d, WA_USB_RecvAsync ret. Result = %d\n", 
            pCommSvr, EndPointID, Result);
#endif

    }/* if(pCommSvr->bActive == TRUE) */

    /* Return 0 for SUCCESS */
    return (SUCCESS);
}

/*********************************************************************************
CUSBCommSvr::DeviceRemoval: Callback function for device removals. Calls Removal()
    from the CBaseCommSvr class to terminate the connection.

Parameters:
    pDevName - Name of the device that has arrived

    pUserData - User data that was registered by the call to WA_USB_OpenSession().

Returns: Always returns SUCCESS(0).
*********************************************************************************/
uint32 WAUSB_CALLBACK CUSBCommSvr::DeviceRemoval(char* pDevName, void* pUserData)
{
    CUSBCommSvr* pUSBCommSvr = (CUSBCommSvr*)pUserData;

    if (pDevName)
    {
#if defined(TRACE_SERVER_ON)
        SYS_PRINTF("CUSBCommSvr DeviceRemoval: pDevName = %s\n", pDevName);
#endif

        if (pUSBCommSvr)
        {
            pUSBCommSvr->SetSourceID(pDevName);
            pUSBCommSvr->Removal();
        }
    }

    /* Return 0 for SUCCESS */
    return SUCCESS;
}

/*********************************************************************************
CUSBCommSvr::DeviceArrival: Callback function for accepting new device arrivals. 
    The DEVICE_DESC structure is copied to insure the DLL doesn't delete the data
    while the stack is still using it. A USBAddr structure is filled out and 
    Arrival() is called to start up a new USB connection.

Parameters:
    pDeviceDesc - Data that describes that device. This data includes the vendor 
        id, product id, number of endpoints, type of endpoint (bulk, interrupt, 
        isochronous) and the data direction of each endpoint. 

    pUserData - User data that was registered by the call to 
        WA_USB_OpenSession(). 

Returns: Always returns SUCCESS(0).
*********************************************************************************/
uint32 WAUSB_CALLBACK CUSBCommSvr::DeviceArrival(PDEVICE_DESC pDeviceDesc,void* pUserData)
{
    CUSBCommSvr*    pUSBCommSvr = (CUSBCommSvr*)pUserData;
    USBAddr         NewClntAddr;
    PDEVICE_DESC    pDevice = NULL;
        
    if (pDeviceDesc && pUSBCommSvr)
    {
        if (pUSBCommSvr->bActive == TRUE)
        {
//#if defined(TRACE_SERVER_ON_STATUS)
            SYS_PRINTF("CUSBCommSvr DeviceArrival: pDevName = %s\n", pDeviceDesc->szDevName);
            SYS_PRINTF("CUSBCommSvr::DeviceArrival Vendor ID = %d:\n", pDeviceDesc->VendorID);
            SYS_PRINTF("CUSBCommSvr::DeviceArrival ProductID ID = %d:\n", pDeviceDesc->ProductID);
            SYS_PRINTF("CUSBCommSvr::DeviceArrival ProductVersion = %d:\n", pDeviceDesc->ProductVersion);
            SYS_PRINTF("CUSBCommSvr::DeviceArrival DeviceSpeed = %d:\n", pDeviceDesc->DeviceSpeed);
            SYS_PRINTF("CUSBCommSvr::DeviceArrival NumEndPoints = %d:\n", pDeviceDesc->NumEndPoints);
            SYS_PRINTF("CUSBCommSvr::DeviceArrival EndPointType = %d:\n", pDeviceDesc->EndPoints[0].EndPointType);
            SYS_PRINTF("CUSBCommSvr::DeviceArrival EndPointID = %d:\n", pDeviceDesc->EndPoints[0].EndPointID);
            SYS_PRINTF("CUSBCommSvr::DeviceArrival Direction = %d:\n", pDeviceDesc->EndPoints[0].Direction);
//#endif

            /* Setup default parameters for new connection setup */

            /* Copy the device structure.. freed in USBCommSvr destrutor */
            pDevice = (PDEVICE_DESC)SYS_MALLOC(sizeof(DEVICE_DESC) 
                + (sizeof(ENDPOINT_DESC) * (pDeviceDesc->NumEndPoints - 1)));
            if (pDevice)
            {
                memcpy(pDevice, pDeviceDesc, sizeof(DEVICE_DESC));
                for (uint16 i=0; i < pDeviceDesc->NumEndPoints; i++)
                {
                    pDevice->EndPoints[i].EndPointID = pDeviceDesc->EndPoints[i].EndPointID;
                    pDevice->EndPoints[i].EndPointType = pDeviceDesc->EndPoints[i].EndPointType;
                    pDevice->EndPoints[i].Direction = pDeviceDesc->EndPoints[i].Direction;
                }

                NewClntAddr.pDevice = pDevice;

                pUSBCommSvr->SetSourceID(pDeviceDesc->szDevName);
                pUSBCommSvr->Arrival(pUSBCommSvr->SourceID, pUSBCommSvr->StackType, &NewClntAddr, sizeof(USBAddr));
            }
        }
    }/* if (pDeviceDesc && pUSBCommSvr) */

    /* Return 0 for SUCCESS */
    return SUCCESS;
}

/*********************************************************************************
CUSBCommSvr::DeviceNotify: Callback function for accepting device arrivals from 
    a device that has been freed up by another application. Calls DeviceArrival 
    to mimic a new device arrival.
    

Parameters:
    pDeviceDesc - Data that describes that device. This data includes the vendor 
        id, product id, number of endpoints, type of endpoint (bulk, interrupt, 
        isochronous) and the data direction of each endpoint. 

    pUserData - User data that was registered by the call to 
        WA_USB_OpenSession(). 

Returns: Always returns SUCCESS(0).
*********************************************************************************/
uint32 WAUSB_CALLBACK CUSBCommSvr::DeviceNotify(PDEVICE_DESC pDeviceDesc, void* pUserData)
{
    uint32 Result = SUCCESS;

    TOUCH(pDeviceDesc);
    TOUCH(pUserData);

#if defined(TRACE_SERVER_ON)
    if (pDeviceDesc)
    {
        SYS_PRINTF("CUSBCommSvr DeviceNotify: pDevName = %s\n", pDeviceDesc->szDevName);
        SYS_PRINTF("CUSBCommSvr::DeviceNotify Vendor ID = %d:\n", pDeviceDesc->VendorID);
        SYS_PRINTF("CUSBCommSvr::DeviceNotify ProductID ID = %d:\n", pDeviceDesc->ProductID);
        SYS_PRINTF("CUSBCommSvr::DeviceNotify ProductVersion = %d:\n", pDeviceDesc->ProductVersion);
        SYS_PRINTF("CUSBCommSvr::DeviceNotify DeviceSpeed = %d:\n", pDeviceDesc->DeviceSpeed);
        SYS_PRINTF("CUSBCommSvr::DeviceNotify NumEndPoints = %d:\n", pDeviceDesc->NumEndPoints);
        SYS_PRINTF("CUSBCommSvr::DeviceNotify EndPointType = %d:\n", pDeviceDesc->EndPoints[0].EndPointType);
        SYS_PRINTF("CUSBCommSvr::DeviceNotify EndPointID = %d:\n", pDeviceDesc->EndPoints[0].EndPointID);
        SYS_PRINTF("CUSBCommSvr::DeviceNotify Direction = %d:\n", pDeviceDesc->EndPoints[0].Direction);
    }
#endif

    if (pDeviceDesc)
    {
        Result = DeviceArrival(pDeviceDesc, pUserData);
    }

    return Result;
}

/*********************************************************************************
CUSBCommSvr::StartClient: Handles the start of the new client connection thread.
    Before any data is read, the method creates a list of asynchronous read 
    requests for each IN endpoint. The USB driver requires a distinct memory buffer 
    for each request so the buffers are allocated for each request. The number of 
    queued up requests are based on USB_MAX_REQUEST_QUEUE as defined in target.h.

    The function loops continuously waiting for messages from the USB driver. 
    Once a message is received, the message is sent via the callback function 
    to be processed by the WACP stack. The loop continues until the bActive flag is
    set to FALSE;
    
Parameters:
    pData - Pointer to the CUSBCommSvr structure that contains the addressing
            information for the USB connection. 

Returns: 
    Always returns TRUE.

*********************************************************************************/
uint32 WACPAPI CUSBCommSvr::StartClient(void* pData)
{
    CUSBCommSvr* pCommSvr = (CUSBCommSvr*)pData;
    char    Buffer[RCVBUFSIZE];
    uint32  MsgSize, PacketLen;
    uint32  BytesProcessed, Index = 0;

#if defined(DEBUG_PRINT)
    SYS_PRINTF("CUSBCommSvr StartClient: Thread starting %p %s\n", pCommSvr, pCommSvr->GetSourceID());
#endif
    if (pCommSvr)
    {
        if (pCommSvr->bActive == TRUE)
        {
            uint32 Result = TRUE;
            uint32 i,j;
            uint32 BufCount = 0;
            boolean bError = FALSE;
            AsyncRequest** pRequestList;

            /* Lock this mutex until this thread terminates */
            pCommSvr->pClientExitMutex->Lock(INFINITE);
            pCommSvr->pDeviceMutex->Lock(INFINITE);

            /* Create separate data buffers for each endpoint read request.
            * More requests will facilitate data streaming, but at the cost
            * of more data buffers being allocated. TODO: Consider making this a 
            * configurable parameter via target.h
            */
            pRequestList = (AsyncRequest**)SYS_MALLOC(pCommSvr->INEndPointNum * sizeof(AsyncRequest*) * USB_MAX_REQUEST_QUEUE);
            if (pRequestList && pCommSvr->Addr.pDevice)
            {
                /* Loop through all IN endpoints */
                for (i = 0; i < (uint32)pCommSvr->Addr.pDevice->NumEndPoints; i++)
                {
                    /* Check for IN endpoints and any errors along the way */
                    if (pCommSvr->Addr.pDevice->EndPoints[i].Direction == ENDPOINT_IN)
                    {
                        /* create a separate buffer for each endpoint request */
                        for (j = 0; j < USB_MAX_REQUEST_QUEUE; j++)
                        {
                            pRequestList[BufCount] = (AsyncRequest*)SYS_MALLOC(sizeof(AsyncRequest));
                            if (pRequestList[BufCount])
                            {
                                /* Store the EndPoint ID */
                                pRequestList[BufCount]->EndpointID = pCommSvr->Addr.pDevice->EndPoints[i].EndPointID;

                                /* Create a data buffer for this request */
                                pRequestList[BufCount]->pDataBuf = (uint8*)SYS_MALLOC(RCVBUFSIZE);
                                if (pRequestList[BufCount])
                                {
                                    BufCount++;
                                }
                                else
                                {
                                    /* A single out of memory error is enough to terminate the thread.
                                    * Set Result to TRUE and exit the thread. 
                                    */
                                    bError = TRUE;
                                }
                            }
                            else
                            {
                                /* A single out of memory error is enough to terminate the thread.
                                * Set Result to TRUE and exit the thread. 
                                */
                                bError = TRUE;
                            }
                        }/* for (j = 0; j < MAX_USB_REQUEST_QUEUE; j++) */

                    }/* if ((pCommSvr->Addr.pDevice->EndPoints[i].Direction == ENDPOINT_IN) && (Result)) */

                    /* If bError is TRUE then we found a memory allocation error */
                    if (bError)
                        break;

                }/* for (i = 0; i < pCommSvr->Addr.pDevice->NumEndPoints; i++) */

            }/* if (pRequestList) */

            pCommSvr->pDeviceMutex->Release();

            if (bError == FALSE)
            {
                for (i = 0; i < BufCount && (pCommSvr->bActive == TRUE); i++)
                {
                    Result = WA_USB_RecvAsync(pCommSvr->SourceID, pRequestList[i]->EndpointID, 
                                            pRequestList[i]->pDataBuf, RCVBUFSIZE);
                    if (Result != ERROR_IO_PENDING)
                    {
#if defined(DEBUG_PRINT)
                        SYS_PRINTF("WA_USB_RecvAsync Result = %d\n", Result);
#endif
                        break;
                    }
                }

                /* Allow the Destructor to be called */
                pCommSvr->pThreadEvent->Raise();

                /* Wait on the Recv callback function to signal when to request more data reads */
                pCommSvr->pReadSem->Wait(INFINITE);
                while(pCommSvr->bActive == TRUE)
                {
                    /* Receive the next message from client */
                    Result = pCommSvr->ReadData((char*)&Buffer[Index], RCVBUFSIZE - Index, &MsgSize);
                    if ((Result) && (MsgSize > 0))
                    {
                        /* Pass the data to the WACP Parser */
                        BytesProcessed = pCommSvr->WACPParser((uint8*)Buffer, MsgSize + Index, &PacketLen);

                        if(PacketLen > RCVBUFSIZE)
                        {
                            /* Invalid preamble/size combination found.  Throw away the preamble and
                               the size.
                             */
                            Index = MsgSize + Index - BytesProcessed - MSG_HEADER;
                            memcpy(Buffer, &Buffer[BytesProcessed + MSG_HEADER], Index);
                        }
                        else if (BytesProcessed == 0)                    
                        {
                            /* Partial packet, increment the Index and continue to read data */
                            Index += MsgSize;
                        }
                        else if (BytesProcessed == (uint32)MsgSize + Index)
                        {
                            /* Full packet received, reset Index to 0 */
                            Index = 0;
                        }
                        else if (BytesProcessed < (uint32)MsgSize + Index)
                        {
                            /* Multiple packets, with partial packet found, copy over processed packet 
                            * and set Index to the first unprocessed packet byte
                            */
                            Index = MsgSize + Index - BytesProcessed;
#if defined(DEBUG_PRINT)
                            SYS_PRINTF("Index = %d ,MsgSize = %d, BytesProcessed = %d\n", Index, MsgSize, BytesProcessed);
#endif
                            memmove(Buffer, &Buffer[BytesProcessed], Index);
                        }
                    }/* if (Result) */
                }/* while(pCommSvr->bActive == TRUE) */
            }/* if if (bError == FALSE) */
            else
            {
                /* Allow the Destructor to be called */
                pCommSvr->pThreadEvent->Raise();

                pCommSvr->pDeviceMutex->Release();
            }

            pCommSvr->pDisconnectMutex->Lock(INFINITE);
            if (bError == FALSE) 
            {
                /* Free up the data buffers and the AsyncRequest structures */
                for (i = 0; i < BufCount; i++)
                {
                    SYS_FREE(pRequestList[i]->pDataBuf);
                    SYS_FREE(pRequestList[i]);
                }

                SYS_FREE(pRequestList);
            }
            pCommSvr->pDisconnectMutex->Release();

            /* Release the exit mutex */
            pCommSvr->pClientExitMutex->Release();
        }
        else
        {
            /* Release the exit mutex */
            pCommSvr->pThreadEvent->Raise();
        }
    }

#if defined(DEBUG_PRINT)
    SYS_PRINTF("CUSBCommSvr StartClient: Thread exiting %p\n", pCommSvr);
#endif

    return TRUE;
}

/*********************************************************************************
CUSBCommSvr::StartServer: Opens a new USB session to receive device arrival/removal
    notification. 
    
Returns: Result from WA_USB_OpenSession.

*********************************************************************************/
uint32 CUSBCommSvr::StartServer()
{
    uint32  Result;

    bServer = TRUE;

    pThreadEvent->Raise();

    /* Open the USB session */
    Result = WA_USB_OpenSession((DEVICEARRIVALPROC)DeviceArrival, this, 
                                (DEVICEREMOVALPROC)DeviceRemoval, this,
                                (DEVICEFREEPROC)DeviceNotify, this); 

    return Result;
}

/* End of File */





